#include "messenger/Client.h"
#include <sstream>

#define CLILOG(lvl, msg) SMLOG(m_pLogListener, lvl, msg)

namespace Mqtt {

CAsyncClient::CAsyncClient(const std::string& _szClientId,
                           const std::string& _szServerUri,
                           IDeliveryListener* _pDeliveryListener,
                           Messenger::ILogListener* _pLogListener):
    Mqtt::CClientImpl(_szServerUri, _szClientId),
    m_pDeliveryListener(_pDeliveryListener),
    m_pConnectionListener(NULL),
    m_pLogListener(_pLogListener),
    m_pSslConfig(NULL),
    m_pWillConfig(NULL),
    m_NetworkProxy(*this) {
    RegisterListener(&m_NetworkProxy);
    m_pConnConfig = new CConnectConfig();
    m_pConnConfig->SetServerURI(_szServerUri);
    m_pConnConfig->SetCleanSession(false);
}

CAsyncClient::~CAsyncClient() {
}

void CAsyncClient::ActivateThreadPool(int _nMinThreads, int _nMaxThreads) {
    m_NetworkProxy.ActivateThreadPool(_nMinThreads, _nMaxThreads);
}

void CAsyncClient::DeactivateThreadPool() {
    m_NetworkProxy.DeactivateThreadPool();
}

void CAsyncClient::SetServerURI(const std::string& _szServerURI) {
    std::string curr = m_pConnConfig->GetServerURI();

    if (_szServerURI != curr) {
        m_pConnConfig->SetServerURI(_szServerURI);
        CLILOG(Information, "Server URI changed (n=" << _szServerURI << ", o=" << curr << ")");
    }
}

std::string CAsyncClient::GetServerUri() const {
    return m_pConnConfig->GetServerURI();
}

void CAsyncClient::SetUsername(const std::string& _szUsername) {
    m_pConnConfig->SetUsername(_szUsername);
}

void CAsyncClient::SetPassword(const std::string& _szPassword) {
    m_pConnConfig->SetPassword(_szPassword);
}

void CAsyncClient::SetCleanSession(bool _bCleanSession) {
    m_pConnConfig->SetCleanSession(_bCleanSession);
}

bool CAsyncClient::IsCleanSession() const {
    return m_pConnConfig->GetCleanSession();
}

void CAsyncClient::SetKeepAliveInterval(int _nInterval) {
    m_pConnConfig->SetKeepAliveInterval(_nInterval);
}

void CAsyncClient::SetConnectionTimeoutInterval(int _nInterval) {
    m_pConnConfig->SetTimeout(_nInterval);
}

void CAsyncClient::SetWillConfiguration(const std::string& _szTopic,
                                        const CMessage::Ptr _pMsg) {
    if (m_pWillConfig.isNull()) {
        m_pWillConfig = new CWillConfig();
    }

    m_pWillConfig->SetTopic(_szTopic);
    m_pWillConfig->SetMessage(_pMsg);
    m_pConnConfig->SetWill(m_pWillConfig);
}

void CAsyncClient::DisableLastWillMessage() {
    m_pWillConfig = NULL;
}

#ifdef USE_SECURE_CONNECTION
void CAsyncClient::SetSecureConfiguration(const std::string& _szTrustStore,
        const std::string& _szKeyStore,
        const std::string& _szPrivateKey,
        const std::string& _szPrivateKeyPassword) {
    if (m_pSslConfig.isNull()) {
        m_pSslConfig = new CSslConfig();
    }

    m_pSslConfig->SetTrustStore(_szTrustStore);
    m_pSslConfig->SetKeyStore(_szKeyStore);
    m_pSslConfig->SetPrivateKey(_szPrivateKey);
    m_pSslConfig->SetPrivateKeyPassword(_szPrivateKeyPassword);
    m_pConnConfig->SetSsl(m_pSslConfig);
}
#endif

void CAsyncClient::Connect() {
    CLILOG(Debug, "Connecting "
           << "(b=" << m_pConnConfig->GetServerURI()
           << ", id=" << GetClientId()
           << ", c=" << m_pConnConfig->GetCleanSession()
           << ", k=" << m_pConnConfig->GetKeepAliveInterval()
           << ")");
    Mqtt::CClientImpl::Connect(m_pConnConfig);
}

void CAsyncClient::Disconnect() {
    if (IsConnected()) {
        CLILOG(Trace, "Disconnecting...");

        try {
            // If clean session is set to true, there is no need to unsubscribe from the response topic.
            Mqtt::CClientImpl::Disconnect();
        } catch (const Poco::RuntimeException& exc) {
            std::stringstream ss;
            ss << "Error while disconnection: " << exc.message();
            throw Poco::RuntimeException(ss.str());
        }

        CLILOG(Information, "Disconnected.");
    }
}

void CAsyncClient::Publish(const std::string& _szTopic,
                           const CMessage::Ptr _pMsg) {
    m_NetworkProxy.Publish(_szTopic, _pMsg);
}

void CAsyncClient::RegisterConnectionListener(Messenger::IConnectionListener*
        _pConnectionListener) {
    m_pConnectionListener = _pConnectionListener;
}

void CAsyncClient::UnregisterConnectionListener() {
    m_pConnectionListener = NULL;
}

void CAsyncClient::OnMessageArrived(const std::string& _szTopic,
                                    const CMessage::Ptr _pMsg) {
    m_pDeliveryListener ? m_pDeliveryListener->OnMessageArrived(_szTopic,
            _pMsg) : void();
}

void CAsyncClient::OnPublishFailed(int _nCode, const std::string& _szTopic,
                                   const CMessage::Ptr _pMsg) {
    CLILOG(Error, "Publish failed (c=" << _nCode << ", t=" << _szTopic << ").");
    m_pDeliveryListener ? m_pDeliveryListener->OnPublishFailed(_nCode, _szTopic,
            _pMsg) : void();
}

void CAsyncClient::OnConnectionLost(const std::string& _szCause) {
    if (m_pConnectionListener != NULL) {
        m_pConnectionListener->OnConnectionLost(_szCause);
    }
}

void CAsyncClient::OnPublishRequired(const std::string& _szTopic,
                                     const CMessage::Ptr _pMsg) {
    PublishSync(_szTopic, _pMsg);
}

CBuffClient::CBuffClient(const std::string& _szClientId,
                         const std::string& _szServerUri,
                         IDeliveryListener& _DeliveryListener,
                         Messenger::ILogListener* _pLogListener,
                         unsigned int _nMaxQueueSize):
    CAsyncClient(_szClientId,
                 _szServerUri, &_DeliveryListener, _pLogListener),
    m_nMaxQueueSize(_nMaxQueueSize),
    m_bIgnoreSessionPresent(true) {
}

void CBuffClient::Connect() {
    Mqtt::CAsyncClient::Connect();

    if (m_bIgnoreSessionPresent || !IsSessionPresent()) {
        CLILOG(Information, "Session will be updated with new subscriptions (s="
               << m_Subscriptions.size() << ", u=" << m_Unsubscriptions.size() << ")");
        SubscribeStore();
        UnsubscribeStore();
    } else {
        CLILOG(Information, "Session is present, skipping the subscription.");
    }

    PublishStore();
}

void CBuffClient::ConnectBasic() {
    Mqtt::CAsyncClient::Connect();
}

void CBuffClient::Subscribe(const std::string& _szTopic, int _nQos) {
    Mqtt::CClientImpl::TopicCollection topic;
    topic.push_back(_szTopic);
    Mqtt::CClientImpl::QosCollection qos;
    qos.push_back(_nQos);
    Subscribe(topic, qos);
}

void CBuffClient::Subscribe(const Mqtt::CClientImpl::TopicCollection& _Topics,
                            const Mqtt::CClientImpl::QosCollection& _QoS) {
    try {
        Mqtt::CClientImpl::Subscribe(_Topics, _QoS);
    } catch (const Poco::RuntimeException& e) {
        CLILOG(Trace, "Failed to subscribe (c=" << e.code() << ", " << e.message() <<
               ")");
    }

    // Will be used on next connect.
    StoreSubscription(_Topics, _QoS);
    // Remove if topic was added to buffer
    RemoveUnsubscription(_Topics);
}

void CBuffClient::Unsubscribe(const std::string& _szTopic) {
    Mqtt::CClientImpl::TopicCollection topic;
    topic.push_back(_szTopic);
    Unsubscribe(topic);
}

void CBuffClient::Unsubscribe(const Mqtt::CClientImpl::TopicCollection&
                              _Topics) {
    try {
        Mqtt::CClientImpl::Unsubscribe(_Topics);
    } catch (const Poco::RuntimeException& e) {
        CLILOG(Debug, "Failed to unsubscribe (" << e.code() << ")");
    }

    // Will be used on next connect.
    StoreUnsubscription(_Topics);
    // Remove if topic was added to buffer
    RemoveSubscription(_Topics);
}

void CBuffClient::IgnoreSessionPresent(bool _bIgnore) {
    m_bIgnoreSessionPresent = _bIgnore;
}

void CBuffClient::OnPublishFailed(int _nCode, const std::string& _szTopic,
                                  const CMessage::Ptr _pMsg) {
    if ((_nCode == MQTTCLIENT_FAILURE || _nCode == MQTTCLIENT_DISCONNECTED) &&
            _pMsg->GetQos() > Mqtt::CMessage::QOS0) {
        // Store message that failes because client is not connected.
        CPublishHolder::Ptr pub = new CPublishHolder(_szTopic, _pMsg);
        m_PublishMutex.lock();

        if (m_Publications.size() >= m_nMaxQueueSize) {
            m_Publications.pop();
        }

        m_Publications.push(pub);
        m_PublishMutex.unlock();
        CLILOG(Trace, "Message published to <" << _szTopic <<
               "> is stored and will be published on next connect.");
    } else {
        CAsyncClient::OnPublishFailed(_nCode, _szTopic, _pMsg);
    }
}

void CBuffClient::StoreSubscription(const CClientImpl::TopicCollection& _Topics,
                                    const CClientImpl::QosCollection& _QoS) {
    m_SubscribeMutex.lock();
    Mqtt::CClientImpl::TopicCollection::const_iterator itr;
    Mqtt::CClientImpl::QosCollection::const_iterator itrQos;

    for (itr = _Topics.begin(), itrQos = _QoS.begin(); itr != _Topics.end();
            ++itr, ++itrQos) {
        SubElement sub(*itr, *itrQos);
        m_Subscriptions.push_back(sub);
        CLILOG(Debug, "Subscription stored: " << sub.first);
    }

    m_SubscribeMutex.unlock();
}

void CBuffClient::StoreUnsubscription(const CClientImpl::TopicCollection&
                                      _Topics) {
    m_UnsubscribeMutex.lock();
    Mqtt::CClientImpl::TopicCollection::const_iterator itr;

    for (itr = _Topics.begin(); itr != _Topics.end(); ++itr) {
        SubElement sub(*itr, 0); // QoS is unused, so it is set to 0.
        m_Unsubscriptions.push_back(sub);
        CLILOG(Debug, "Unsubscription stored: " << sub.first);
    }

    m_UnsubscribeMutex.unlock();
}

void CBuffClient::RemoveSubscription(const CClientImpl::TopicCollection&
                                     _Topics) {
    m_SubscribeMutex.lock();
    Remove(m_Subscriptions, _Topics);
    m_SubscribeMutex.unlock();
}

void CBuffClient::RemoveUnsubscription(const CClientImpl::TopicCollection&
                                       _Topics) {
    m_UnsubscribeMutex.lock();
    Remove(m_Unsubscriptions, _Topics);
    m_UnsubscribeMutex.unlock();
}

void CBuffClient::Remove(CBuffClient::SubscriptionBuffer& _Buffer,
                         const CClientImpl::TopicCollection& _Topics) {
    SubscriptionBuffer::iterator itr;
    std::string topic;

    for (itr = _Buffer.begin(); itr != _Buffer.end();) {
        topic = itr->first;

        if (std::find(_Topics.begin(), _Topics.end(), topic) != _Topics.end()) {
            itr = _Buffer.erase(itr);
        } else {
            ++itr;
        }
    }
}

void CBuffClient::PublishStore() {
    m_PublishMutex.lock();
    size_t pubsize = m_Publications.size();

    while (!m_Publications.empty()) {
        CPublishHolder::Ptr elm = m_Publications.front();
        m_Publications.pop();
        const std::string topic = elm->GetTopic();
        const Mqtt::CMessage::Ptr pMsg = elm->GetMessage();
        Publish(topic, pMsg);
    }

    m_PublishMutex.unlock();

    if (pubsize != 0) {
        CLILOG(Information, "Store published " << pubsize << " messages");
    }
}

void CBuffClient::SubscribeStore() {
    Mqtt::CClientImpl::TopicCollection topics;
    Mqtt::CClientImpl::QosCollection qoslevels;
    m_SubscribeMutex.lock();

    while (!m_Subscriptions.empty()) {
        SubElement elm = m_Subscriptions.front();
        m_Subscriptions.pop_front();
        const std::string topic = elm.first;
        const int qos = elm.second;
        topics.push_back(topic);
        qoslevels.push_back(qos);
    }

    m_SubscribeMutex.unlock();
    Subscribe(topics, qoslevels);

    if (topics.size() != 0) {
        CLILOG(Debug, "Subscribed on " << topics.size() << " topics");
    }
}

void CBuffClient::UnsubscribeStore() {
    Mqtt::CClientImpl::TopicCollection topics;
    m_UnsubscribeMutex.lock();

    while (!m_Unsubscriptions.empty()) {
        SubElement elm = m_Unsubscriptions.front();
        m_Unsubscriptions.pop_front();
        const std::string topic = elm.first;
        topics.push_back(topic);
    }

    m_UnsubscribeMutex.unlock();
    Unsubscribe(topics);

    if (topics.size() != 0) {
        CLILOG(Debug, "Unsubscribed from " << topics.size() << " topics");
    }
}

}
