/****************************************************************************
 *  Copyright (c) 2015 RT-RK d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.rt-rk.com
 *  -----------------------------------------------------
 *
 * \project Oblo Home Manager - messaging library
 * \file Messages.cpp
 * \brief
 * Classes for different types of messages arriving from the network.
 *
 * @Author Milan Tucic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#include "messenger/Messages.h"
#include <limits.h>
#include <Poco/Timestamp.h>
#include <Poco/LocalDateTime.h>
#include <Poco/DateTimeFormat.h>

namespace Messenger {

static const std::string UID = "uid";
static const std::string PARAMS = "params";
static const std::string SENDER = "sender";
static const std::string TYPE = "type";
static const std::string NAME = "name";
static const std::string TIMESTAMP = "timestamp";
static const std::string TS = "ts";
static const std::string ETS = "ets";
static const std::string ONLINE = "online";
static const std::string CODE = "code";
static const std::string MESSAGE = "message";

Poco::Random CMessageImpl::RANDOM_GENERATOR;
const Poco::Timestamp::TimeVal CMessageImpl::EXPIRATION_NOT_SET = 0;

CMessageImpl::CMessageImpl():
    m_nId(RANDOM_GENERATOR.next()),
    m_tExpirationTime(EXPIRATION_NOT_SET) {
    m_Message = Poco::DynamicStruct();
}

CMessageImpl::CMessageImpl(const CMessageImpl& _Message):
    m_nId(_Message.m_nId),
    m_pParams(_Message.m_pParams),
    m_tTimestamp(_Message.m_tTimestamp),
    m_tExpirationTime(_Message.m_tExpirationTime),
    m_Message(_Message.m_Message) {
}

CMessageImpl& CMessageImpl::operator =(const CMessageImpl& _Msg) {
    if (this != &_Msg) {
        m_nId = _Msg.m_nId;
        m_pParams = _Msg.m_pParams;
        m_tTimestamp = _Msg.m_tTimestamp;
        m_tExpirationTime = _Msg.m_tExpirationTime;
        m_Message = _Msg.m_Message;
    }

    return *this;
}

unsigned int CMessageImpl::GetId() const {
    return m_nId;
}

const Poco::DynamicStruct::Ptr CMessageImpl::GetParams() const {
    return m_pParams;
}

Poco::DynamicStruct& CMessageImpl::GetParamsHolder() {
    Poco::DynamicStruct& mds = const_cast<Poco::DynamicStruct&>(m_Message.extract<Poco::DynamicStruct>());
    mds[PARAMS] = Poco::DynamicStruct();
    return const_cast<Poco::DynamicStruct&>(mds[PARAMS].extract<Poco::DynamicStruct>());
}

void CMessageImpl::SetParams(Poco::DynamicStruct::Ptr _pParams) {
    m_pParams = _pParams;
}

Poco::Timestamp CMessageImpl::GetTimestamp() const {
    return m_tTimestamp;
}

void CMessageImpl::SetTimestamp(const Poco::Timestamp& _Timestamp) {
    m_tTimestamp = _Timestamp;
}

Poco::Timestamp CMessageImpl::GetExpirationTime() const {
    return m_tExpirationTime;
}

bool CMessageImpl::IsExpirationSet() const {
    return m_tExpirationTime != EXPIRATION_NOT_SET;
}

void CMessageImpl::SetTimestamp(const std::string& _szTimestamp) {
    int diff;
    Poco::DateTime time;
    Poco::DateTimeParser::parse(Poco::DateTimeFormat::ISO8601_FORMAT, _szTimestamp, time, diff);
    m_tTimestamp = time.timestamp();
}

void CMessageImpl::SetId(unsigned int _nId) {
    m_nId = _nId;
}

std::string CMessageImpl::ToJson(ProtocolVersion _eProtocol) {
    std::string output;
    ToJson(_eProtocol, output);
    return output;
}

void CMessageImpl::ToJson(Messenger::ProtocolVersion _eProtocol, std::string &_szOutput) {
    Poco::DynamicStruct& mds = const_cast<Poco::DynamicStruct&>(m_Message.extract<Poco::DynamicStruct>());
    mds.insert(UID, m_nId);

    if (!m_pParams.isNull()) {
        mds.insert(PARAMS, *m_pParams);
    }

    if (_eProtocol == PROTOCOL_WISE_1) {
        Poco::Timestamp::TimeVal ts = m_tTimestamp.epochMicroseconds() / 1000;
        mds.insert(TS, ts);

        if (IsExpirationSet()) {
            Poco::Timestamp::TimeVal ets = m_tExpirationTime.epochMicroseconds() / 1000;
            mds.insert(ETS, ets);
        }
    }

    InflateJson(mds, _eProtocol);

    m_Message.convert<std::string>(_szOutput);
}

void CMessageImpl::Parse(CMessageImpl& _MessageToFill, const Poco::DynamicStruct& _MessageToParse, ProtocolVersion _eProtocol) {
    Poco::DynamicStruct::Ptr ds = NULL;
    _MessageToFill.SetId(_MessageToParse[UID].convert<int>());

    if (_MessageToParse.contains(PARAMS)) {
        ds = new Poco::DynamicStruct(_MessageToParse[PARAMS].extract<Poco::DynamicStruct>());
    } else {
        ds = new Poco::DynamicStruct();
    }

    _MessageToFill.SetParams(ds);

    if (_eProtocol == PROTOCOL_WISE_1) {
        Poco::Timestamp::TimeVal ts = _MessageToParse[TS].convert<Poco::Timestamp::TimeVal>();
        _MessageToFill.m_tTimestamp = FromEpochMilliseconds(ts);
        Poco::Timestamp::TimeVal time = EXPIRATION_NOT_SET;

        if (_MessageToParse.find(ETS) != _MessageToParse.end()) {
            time = _MessageToParse[ETS].convert<Poco::Timestamp::TimeVal>();
        }

        _MessageToFill.m_tExpirationTime = FromEpochMilliseconds(time);
    }
}

Poco::Timestamp CMessageImpl::FromEpochMilliseconds(Poco::Timestamp::TimeVal _nMilliseconds) {
    Poco::Timestamp::TimeVal seconds = _nMilliseconds / 1000;
    Poco::Timestamp::TimeVal millis = _nMilliseconds % 1000;
    millis *= 1000; // to microseconds
    Poco::Timestamp ts = Poco::Timestamp::fromEpochTime(seconds);
    ts += millis;
    return ts;
}

CEventMessage::CEventMessage() {
}

CEventMessage::CEventMessage(const CEventMessage& _Event):
    CMessageImpl(_Event),
    m_szName(_Event.m_szName) {
}

CEventMessage& CEventMessage::operator =(const CEventMessage& _Event) {
    if (this != &_Event) {
        CMessageImpl::operator=(_Event);
        m_szName = _Event.m_szName;
    }

    return *this;
}

std::string CEventMessage::GetName() const {
    return m_szName;
}

void CEventMessage::SetName(const std::string& _szName) {
    m_szName = _szName;
}

std::string CEventMessage::GetTimestampString() const {
    std::string time = Poco::DateTimeFormatter::format(Poco::LocalDateTime(GetTimestamp()), Poco::DateTimeFormat::ISO8601_FORMAT);
    return time;
}

CEventMessage::Ptr CEventMessage::Parse(const Poco::DynamicStruct& _Message, ProtocolVersion _eProtocol) {
    CEventMessage::Ptr msg(new CEventMessage());

    try {
        CMessageImpl::Parse(*msg, _Message, _eProtocol);
        msg->m_szName = _Message[NAME].extract<std::string>();

        if (_eProtocol == PROTOCOL_OBLO_1) {
            if (_Message.find(TIMESTAMP) != _Message.end()) {
                msg->SetTimestamp(_Message[TIMESTAMP].extract<std::string>());
            }
        }
    } catch (const Poco::Exception& e) {
        throw Poco::DataFormatException("Error while parsing event: " + e.displayText());
    }

    return msg;
}

void CEventMessage::InflateJson(Poco::DynamicStruct& _Obj, ProtocolVersion _eProtocol) const {
    _Obj.insert(NAME, m_szName);

    if (_eProtocol == PROTOCOL_OBLO_1) {
        _Obj.insert(TIMESTAMP, GetTimestampString());
    }
}

CLastwillMessage::CLastwillMessage():
    m_bOnline(false) {
}

CLastwillMessage::CLastwillMessage(const CLastwillMessage& _Obj):
    CMessageImpl(_Obj),
    m_bOnline(_Obj.m_bOnline) {
}

CLastwillMessage& CLastwillMessage::operator =(const CLastwillMessage& _Obj) {
    if (this != &_Obj) {
        CMessageImpl::operator=(_Obj);
        m_bOnline = _Obj.m_bOnline;
    }

    return *this;
}

bool CLastwillMessage::IsOnline() const {
    return m_bOnline;
}

void CLastwillMessage::SetStatus(bool _bOnline) {
    m_bOnline = _bOnline;
}

CLastwillMessage::Ptr CLastwillMessage::Parse(const Poco::DynamicStruct& _Message, ProtocolVersion _eProtocol) {
    CLastwillMessage::Ptr msg(new CLastwillMessage());

    try {
        CMessageImpl::Parse(*msg, _Message, _eProtocol);
        msg->SetStatus(_Message[ONLINE].extract<bool>());
    } catch (const Poco::Exception& e) {
        throw Poco::DataFormatException("Error while parsing lwm: " + e.displayText());
    }

    return msg;
}

void CLastwillMessage::InflateJson(Poco::DynamicStruct& _Obj, ProtocolVersion _eProtocol) const {
    _Obj.insert(ONLINE, m_bOnline);
}

CRequestMessage::CRequestMessage() {
    m_pCompletitionListener = NULL;
}

CRequestMessage::CRequestMessage(const CRequestMessage& _Request) :
    CMessageImpl(_Request),
    m_szSender(_Request.m_szSender),
    m_szType(_Request.m_szType),
    m_szName(_Request.m_szName),
    m_pCompletitionListener(_Request.m_pCompletitionListener) {
}

CRequestMessage& CRequestMessage::operator =(const CRequestMessage& _Request) {
    if (this != &_Request) {
        CMessageImpl::operator=(_Request);
        m_szSender = _Request.m_szSender;
        m_szType = _Request.m_szType;
        m_szName = _Request.m_szName;
        m_pCompletitionListener = _Request.m_pCompletitionListener;
    }

    return *this;
}

std::string CRequestMessage::GetSender() const {
    return m_szSender;
}

std::string CRequestMessage::GetType() const {
    return m_szType;
}

void CRequestMessage::SetType(const std::string& _szType) {
    m_szType = _szType;
}

std::string CRequestMessage::GetName() const {
    return m_szName;
}

void CRequestMessage::SetName(const std::string& _szName) {
    m_szName = _szName;
}

CRequestMessage::Ptr CRequestMessage::Parse(const Poco::DynamicStruct& _Message, ProtocolVersion _eProtocol) {
    CRequestMessage::Ptr msg(new CRequestMessage());

    try {
        CMessageImpl::Parse(*msg, _Message, _eProtocol);
        msg->SetSender(_Message[SENDER].extract<std::string>());
        msg->SetName(_Message[NAME].extract<std::string>());

        if (_eProtocol == PROTOCOL_OBLO_1) {
            msg->SetType(_Message[TYPE].extract<std::string>());
        }
    } catch (const Poco::Exception& e) {
        throw Poco::DataFormatException("Error while parsing request: " + e.displayText());
    }

    return msg;
}

void CRequestMessage::SetSender(const std::string& _szSender) {
    m_szSender = _szSender;
}

void CRequestMessage::InflateJson(Poco::DynamicStruct& _Obj, ProtocolVersion _eProtocol) const {
    _Obj.insert(SENDER, m_szSender);
    _Obj.insert(NAME, m_szName);

    if (_eProtocol == PROTOCOL_OBLO_1) {
        _Obj.insert(TYPE, m_szType);
    }
}

void CRequestMessage::RegisterCompletionListener(ICompletionListener* _pCompletionListener) {
    m_pCompletitionListener = _pCompletionListener;
}

void CRequestMessage::ReportResponse(const CResponseMessage::Ptr _pResponse) {
    if (m_pCompletitionListener) {
        m_pCompletitionListener->OnRequestCompleted(GetId());
    }
}

void CRequestMessage::ReportFailure() {
    if (m_pCompletitionListener) {
        m_pCompletitionListener->OnRequestCompleted(GetId());
    }
}

CResponseMessage::CResponseMessage(): m_nCode(0) {
}

CResponseMessage::CResponseMessage(const CResponseMessage& _Response):
    CMessageImpl(_Response),
    m_nCode(_Response.m_nCode),
    m_szMessage(_Response.m_szMessage) {
}

CResponseMessage& CResponseMessage::operator =(const CResponseMessage& _Response) {
    if (this != &_Response) {
        CMessageImpl::operator=(_Response);
        m_nCode = _Response.m_nCode;
        m_szMessage = _Response.m_szMessage;
    }

    return *this;
}

CResponseMessage::Ptr CResponseMessage::Parse(const Poco::DynamicStruct& _Message, ProtocolVersion _eProtocol) {
    CResponseMessage::Ptr msg(new CResponseMessage());

    try {
        CMessageImpl::Parse(*msg, _Message, _eProtocol);
        msg->SetCode(_Message[CODE].convert<int>());

        // Message is an optional field.
        if (_Message.contains(MESSAGE)) {
            msg->SetMessage(_Message[MESSAGE].extract<std::string>());
        }
    } catch (const Poco::Exception& e) {
        throw Poco::DataFormatException("Error while parsing response: " + e.displayText());
    }

    return msg;
}

int CResponseMessage::GetCode() const {
    return m_nCode;
}

void CResponseMessage::SetCode(int _nCode) {
    m_nCode = _nCode;
}

std::string CResponseMessage::GetMessage() const {
    return m_szMessage;
}

void CResponseMessage::SetMessage(const std::string& _szMessage) {
    m_szMessage = _szMessage;
}

void CResponseMessage::InflateJson(Poco::DynamicStruct& _Obj, ProtocolVersion _eProtocol) const {
    _Obj.insert(CODE, m_nCode);

    if (!m_szMessage.empty()) {
        _Obj.insert(MESSAGE, m_szMessage);
    }
}

CSyncRequest::CSyncRequest(const CRequestMessage& _Request) :
    CRequestMessage(_Request), m_pResponse(NULL) {
}

const CResponseMessage::Ptr CSyncRequest::GetResponse() const {
    return m_pResponse;
}

void CSyncRequest::Wait(unsigned int _nTimeout) const {
    m_SyncObj.wait(_nTimeout);
}

void CSyncRequest::ReportResponse(const CResponseMessage::Ptr _pResponse) {
    m_pResponse = _pResponse;
    m_SyncObj.notify();
    CRequestMessage::ReportResponse(_pResponse);
}

void CSyncRequest::ReportFailure() {
    m_SyncObj.notify();
    CRequestMessage::ReportFailure();
}

CAsyncRequest::CAsyncRequest(const CRequestMessage& _Request, const CRequestDelegate::Ptr _pDelegate) :
    CRequestMessage(_Request),
    m_pRequestDelegate(_pDelegate) {
}

void CAsyncRequest::ReportResponse(const CResponseMessage::Ptr _pResponse) {
    if (!_pResponse.isNull() && !m_pRequestDelegate.isNull()) {
        m_pRequestDelegate->ReportSuccess(GetId(), *_pResponse);
    }

    CRequestMessage::ReportResponse(_pResponse);
}

void CAsyncRequest::ReportFailure() {
    if (!m_pRequestDelegate.isNull() && !m_pRequestDelegate.isNull()) {
        m_pRequestDelegate->ReportFailure(GetId());
    }

    CRequestMessage::ReportFailure();
}

CRequestDelegate::CRequestDelegate() :
    m_pObject(NULL),
    m_pStubSucc(NULL),
    m_pStubFail(NULL) {
}

void CRequestDelegate::ReportSuccess(unsigned int _nRequestId, const CResponseMessage& _Response) const {
    (*m_pStubSucc)(m_pObject, _nRequestId, _Response, m_bAutoDelete);
}

void CRequestDelegate::ReportFailure(unsigned int _nRequestId) const {
    (*m_pStubFail)(m_pObject, _nRequestId, m_bAutoDelete);
}

}
