/****************************************************************************
 *  Copyright (c) 2015 RT-RK d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.rt-rk.com
 *  -----------------------------------------------------
 *
 * \project Oblo Home Manager - messaging library
 * \file Messenger.cpp
 * \brief
 * Implementation of messagining inside of OBLO system.
 *
 * @Author Milan Tucic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#include "messenger/Messenger.h"

#include <sstream>
using namespace std;

#include "Poco/JSON/Parser.h"
#include "Poco/Dynamic/Var.h"
#include "Poco/StringTokenizer.h"
#include "Poco/String.h"

namespace Messenger {

const std::string ERR_MSGR_NOT_CONNECTED = "Messenger not connected.";
const std::string TOPIC_SEPARATOR = "/";
const std::string EMPTY_STRING = "";

#define MLOG(lvl, msg) { \
    std::stringstream ss; \
    ss << msg; \
    On##lvl(ss.str());\
}

CLogger::CLogger(const string& _szIdentifier):
    m_pLogListener(NULL),
    m_szId(_szIdentifier) {
}

void CLogger::RegisterLogListener(ILogListener* _pLogListener) {
    m_pLogListener = _pLogListener;
}

void CLogger::UnregisterLogListener() {
    m_pLogListener = NULL;
}

void CLogger::OnTrace(const std::string& _Msg) {
    m_pLogListener ? m_pLogListener->OnTrace(_Msg) : void();
}

void CLogger::OnDebug(const std::string& _Msg) {
    m_pLogListener ? m_pLogListener->OnDebug(_Msg) : void();
}

void CLogger::OnInformation(const std::string& _Msg) {
    m_pLogListener ? m_pLogListener->OnInformation(_Msg) : void();
}

void CLogger::OnWarning(const std::string& _Msg) {
    m_pLogListener ? m_pLogListener->OnWarning(_Msg) : void();
}

void CLogger::OnError(const std::string& _Msg) {
    m_pLogListener ? m_pLogListener->OnError(_Msg) : void();
}

CCleaner::CCleanupElement::CCleanupElement(unsigned int _nId, int _nTimeout) {
    m_nId = _nId;
    m_nTimeout = _nTimeout;
}

CCleaner::CCleanupElement::CCleanupElement(const CCleaner::CCleanupElement& _Elm) {
    m_nId = _Elm.m_nId;
    m_nTimeout = _Elm.m_nTimeout;
}

CCleaner::CCleanupElement& CCleaner::CCleanupElement::operator =(const CCleaner::CCleanupElement& _Elm) {
    if (this != &_Elm) {
        m_nId = _Elm.m_nId;
        m_nTimeout = _Elm.m_nTimeout;
    }

    return *this;
}

CCleaner::CCleaner():
    m_CleanupTimer(1000, 1000) {
    m_CleanupTimer.start(Poco::TimerCallback<CCleaner>(*this, &CCleaner::OnCleanup));
}

CCleaner::CCleaner(const CCleaner& _Cleaner):
    m_CleanupList(_Cleaner.m_CleanupList),
    m_CleanupTimer(1000, 1000) {
    m_CleanupTimer.start(Poco::TimerCallback<CCleaner>(*this, &CCleaner::OnCleanup));
}

void CCleaner::StopCleaning() {
    m_CleanupTimer.stop();
    CleanupAll();
}

void CCleaner::ScheduleCleanup(unsigned int _nId, unsigned int _nTime) {
    Poco::Mutex::ScopedLock lock(m_CleanupMutex);
    CleanUpList::iterator itr;
    CCleanupElement::Ptr tmpElm;
    int diff = _nTime;
    int prevDiff = diff;

    for (itr = m_CleanupList.begin(); itr != m_CleanupList.end(); ++itr) {
        tmpElm = *itr;
        diff -= tmpElm->m_nTimeout;

        if (diff <= 0) {
            tmpElm->m_nTimeout = abs(diff);
            m_CleanupList.insert(itr, new CCleanupElement(_nId, prevDiff));
            break;
        }

        prevDiff = diff;
    }

    // If list is empty, or element is not insert inside the loop it will be processed here.
    if (itr == m_CleanupList.end()) {
        m_CleanupList.push_back(new CCleanupElement(_nId, prevDiff));
    }
}

void CCleaner::OnCleanup(Poco::Timer& timer) {
    Poco::Mutex::ScopedLock lock(m_CleanupMutex);
    CleanUpList::iterator itr;
    CCleanupElement::Ptr elm;
    int diff = 1 + timer.skipped();

    for (itr = m_CleanupList.begin(); itr != m_CleanupList.end(); ++itr) {
        elm = *itr;
        elm->m_nTimeout = elm->m_nTimeout - diff;

        if (elm->m_nTimeout < 0) {
            CleanRequest(elm->m_nId);
        } else {
            break;
        }
    }

    m_CleanupList.erase(m_CleanupList.begin(), itr);
}

void CCleaner::CleanupAll() {
    Poco::Mutex::ScopedLock lock(m_CleanupMutex);
    CleanUpList::iterator itr;
    CCleanupElement::Ptr elm;

    for (itr = m_CleanupList.begin(); itr != m_CleanupList.end(); ++itr) {
        elm = *itr;
        CleanRequest(elm->m_nId);
    }
}

CMessenger::CMessenger(const std::string& _szTag,
                       const std::string& _szUsername,
                       const std::string& _szPassword,
                       const std::string& _szUri,
                       ProtocolVersion _eProtocol) :
    CLogger(_szTag),
    m_szUserTag(_szTag),
    m_ObloClient(_szUsername, _szUri, *this, this),
    m_bInitialized(false),
    m_pResponder(NULL),
    m_pDefaultListener(NULL),
    m_szRequestTopic("/req"),
    m_szResponseTopic("/rsp"),
    m_szEventTopic("/evt"),
    m_szStatusTopic("/lwm"),
    m_eProtocolVersion(_eProtocol) {
    if (m_eProtocolVersion == PROTOCOL_WISE_1) {
        m_szStatusTopic = "/sts";
    }

    m_ObloClient.SetUsername(_szUsername);
    m_ObloClient.SetPassword(_szPassword);
}

CMessenger::~CMessenger() {
    MLOG(Trace, "Deinitialization is started.");
    // Wait on processing to complete
    Poco::RWLock::ScopedWriteLock lock(m_ProcessingSync);
    StopCleaning();

    try {
        Disconnect();
    } catch (const Poco::RuntimeException& e) {
    }

    CleanRequests();
    MLOG(Information, "Deinitialized.");
}

void CMessenger::ActivateThreadPool(int _nMinThreads, int _nMaxThreads) {
    m_ObloClient.ActivateThreadPool(_nMinThreads, _nMaxThreads);
}

void CMessenger::DeactivateThreadPool() {
    m_ObloClient.DeactivateThreadPool();
}

void CMessenger::SetSecureConfiguration(const std::string& _szTrustStore,
                                        const std::string& _szKeyStore,
                                        const std::string& _szPrivateKey,
                                        const std::string& _szPrivateKeyPassword) {
    m_ObloClient.SetSecureConfiguration(_szTrustStore, _szKeyStore, _szPrivateKey, _szPrivateKeyPassword);
}

void CMessenger::Connect() {
    Initialize();
    MLOG(Debug, "Connecting to " << m_ObloClient.GetServerUri()
         << " (id=" << m_ObloClient.GetClientId()
         << ", c=" << m_ObloClient.IsCleanSession() << ")");

    try {
        m_ObloClient.Connect();
        MLOG(Debug, "Connected.");
        // Lastwill with status ONLINE
        Mqtt::CMessage::Ptr status = GenerateStatusMessage(true);
        m_ObloClient.Publish(m_szUserTag + m_szStatusTopic, status);
        MLOG(Debug, "Status message sent.");
    } catch (const Poco::RuntimeException& exc) {
        std::stringstream ss;
        ss << "Initialization error: " << exc.message();
        throw Poco::RuntimeException(ss.str());
    }

    MLOG(Information, "Connection established.");
}

void CMessenger::Initialize() {
    if (!m_bInitialized) {
        m_bInitialized  = true;
        // Last will message
        Mqtt::CMessage::Ptr status = GenerateStatusMessage(false);
        m_ObloClient.SetWillConfiguration(m_szUserTag + m_szStatusTopic, status);
        // Response messages
        std::string topic = m_szUserTag + m_szResponseTopic;
        m_ObloClient.Subscribe(topic, Mqtt::CMessage::QOS1);
        MLOG(Debug, "Registered for response messages.");
    }
}

Mqtt::CMessage::Ptr CMessenger::GenerateStatusMessage(bool _bOnline) const {
    Mqtt::CMessage::Ptr mp = new Mqtt::CMessage();
    CLastwillMessage lwm;
    lwm.SetStatus(_bOnline);
    lwm.ToJson(m_eProtocolVersion, mp->GetPayloadHolder());
    mp->SetQos(Mqtt::CMessage::QOS1);
    return mp;
}

void CMessenger::Reconnect(bool _bCleanSession) {
    Disconnect();
    m_ObloClient.SetCleanSession(_bCleanSession);
    Connect();
    stringstream ss;
    ss << "Reconnected " << (_bCleanSession ? "with" : "without") << " clean session";
    MLOG(Debug, ss.str());
}

void CMessenger::Disconnect() {
    if (m_ObloClient.IsConnected()) {
        Mqtt::CMessage::Ptr status = GenerateStatusMessage(false);
        m_ObloClient.PublishSync(m_szUserTag + m_szStatusTopic, status);
        m_ObloClient.Disconnect();
    }
}

bool CMessenger::IsConnected() const {
    return m_ObloClient.IsConnected();
}

void CMessenger::IgnoreSessionPresent(bool _bIgnore) {
    m_ObloClient.IgnoreSessionPresent(_bIgnore);
}

void CMessenger::RegisterResponder(IResponder* _pResponder) {
    if (m_pResponder != NULL) {
        MLOG(Trace, "Re-writing responder.");
        m_pResponder = _pResponder;
    } else {
        m_pResponder = _pResponder;
        MLOG(Trace, "Registering responder...");
        std::string topic = m_szUserTag + m_szRequestTopic;

        try {
            m_ObloClient.Subscribe(topic, Mqtt::CMessage::QOS1);
        } catch (const Poco::RuntimeException& exc) {
            std::stringstream ss;
            ss << "Error while registering responder: " << exc.message();
            throw Poco::RuntimeException(ss.str());
        }
    }

    MLOG(Information, "Responder registered.");
}

void CMessenger::UnregisterResponder() {
    m_pResponder = NULL;
    MLOG(Trace, "Cleaning responder...");
    std::string topic = m_szUserTag + m_szRequestTopic;

    try {
        m_ObloClient.Unsubscribe(topic);
    } catch (const Poco::RuntimeException& exc) {
        std::stringstream ss;
        ss << "Error while cleaning responder: " << exc.message();
        throw Poco::RuntimeException(ss.str());
    }

    MLOG(Information, "Responder is cleaned.");
}

void CMessenger::ActivateFunctionality(const string& _szFunctionality) {
    MLOG(Trace, "Activating a functionality...");
    std::string topic = m_szUserTag + TOPIC_SEPARATOR + _szFunctionality + m_szRequestTopic;

    try {
        m_ObloClient.Subscribe(topic, Mqtt::CMessage::QOS1);
    } catch (const Poco::RuntimeException& exc) {
        std::stringstream ss;
        ss << "Error while activating a functionality: " << exc.message();
        throw Poco::RuntimeException(ss.str());
    }

    MLOG(Trace, "Functionality is activated.");
}

void CMessenger::DeactivateFunctionality(const string& _szFunctionality) {
    MLOG(Trace, "Deactivating a functionality...");
    std::string topic = m_szUserTag + TOPIC_SEPARATOR + _szFunctionality + m_szRequestTopic;

    try {
        m_ObloClient.Unsubscribe(topic);
    } catch (const Poco::RuntimeException& exc) {
        std::stringstream ss;
        ss << "Error while deactivating a functionality: " << exc.message();
        throw Poco::RuntimeException(ss.str());
    }

    MLOG(Information, "Functionality is deactivated.");
}

void CMessenger::SendEvent(CEventMessage& _Event, const std::string& _szFunctionality) {
    std::string topic;
    topic = _szFunctionality.empty() ? m_szUserTag + m_szEventTopic : m_szUserTag + TOPIC_SEPARATOR + _szFunctionality + m_szEventTopic;
    Mqtt::CMessage::Ptr pubmsg = new Mqtt::CMessage();
    std::string& payload = pubmsg->GetPayloadHolder();
    _Event.ToJson(m_eProtocolVersion, payload);
    pubmsg->SetQos(Mqtt::CMessage::QOS1);
    MLOG(Trace, "Sending event: " << payload << " to: " << topic << ".");

    try {
        m_ObloClient.Publish(topic, pubmsg);
    } catch (const Poco::RuntimeException& exc) {
        std::stringstream ss;
        ss << "Error while sending event: " << exc.message();
        throw Poco::RuntimeException(ss.str());
    }

    MLOG(Debug, "Event (" << _Event.GetName() << ") sent.");
}

void CMessenger::RegisterForEvents(IEventListener* _pEventListener, const string& _szTag, const string& _szFunctionality) {
    Poco::Mutex::ScopedLock lock(m_EvtLwmMutex);
    std::string topic;
    topic = _szFunctionality.empty() ? _szTag : _szTag + TOPIC_SEPARATOR + _szFunctionality;
    event_listeners_t listeners = m_EventListeners[topic];

    if (listeners.empty()) {
        MLOG(Trace, "Registering on events from: " << topic << ".");

        try {
            m_ObloClient.Subscribe(topic + m_szEventTopic, Mqtt::CMessage::QOS1);
        } catch (const Poco::RuntimeException& e) {
            std::stringstream ss;
            ss << "Error while registering on events: " << e.message();
            throw Poco::RuntimeException(ss.str());
        }

        MLOG(Debug, "Registered on events from: " << topic << ".");
    }

    listeners.push_back(_pEventListener);
    m_EventListeners[topic] = listeners;
    MLOG(Information, "Added events listener for: " << topic << ".");
}

void CMessenger::UnregisterFromEvents(IEventListener* _pEventListener, const string& _szTag, const string& _szFunctionality) {
    std::string tag;
    tag = _szFunctionality.empty() ? _szTag : _szTag + TOPIC_SEPARATOR + _szFunctionality;
    MLOG(Trace, "Cleaning up event listener for: " << tag << ".");
    event_listeners_t::iterator itr;
    Poco::Mutex::ScopedLock lock(m_EvtLwmMutex);
    event_listeners_t listeners = m_EventListeners[tag];
    itr = std::find(listeners.begin(), listeners.end(), _pEventListener);

    if (itr != listeners.end()) {
        listeners.erase(itr);
    }

    if (listeners.empty()) {
        MLOG(Debug, "Unregistering from events for: " << tag << ".");

        try {
            m_ObloClient.Unsubscribe(tag + m_szEventTopic);
        } catch (const Poco::RuntimeException& exc) {
            std::stringstream ss;
            ss << "Error while unregistering from events: " << exc.message();
            throw Poco::RuntimeException(ss.str());
        }

        MLOG(Debug, "Unregistered from events for: " << tag << ".");
    }

    m_EventListeners[tag] = listeners;
    MLOG(Information, "Event listener for: " << tag << " cleaned.");
}

void CMessenger::RegisterForLwm(IEventListener* _pLwmListener, const string& _szTag) {
    Poco::Mutex::ScopedLock lock(m_EvtLwmMutex);
    event_listeners_t listeners = m_LwmListeners[_szTag];

    if (listeners.empty()) {
        MLOG(Trace, "Registering on LWM from: " << _szTag << ".");

        try {
            m_ObloClient.Subscribe(_szTag + m_szStatusTopic, Mqtt::CMessage::QOS1);
        } catch (const Poco::RuntimeException& e) {
            std::stringstream ss;
            ss << "Error while registering on events: " << e.message();
            throw Poco::RuntimeException(ss.str());
        }

        MLOG(Debug, "Registered on LWM from: " << _szTag << ".");
    }

    listeners.push_back(_pLwmListener);
    m_LwmListeners[_szTag] = listeners;
    MLOG(Information, "Added LWM listener for: " << _szTag << ".");
}

void CMessenger::UnregisterFromLwm(IEventListener* _pLwmListener, const string& _szTag) {
    MLOG(Trace, "Cleaning up LWM listener for: " << _szTag << ".");
    event_listeners_t::iterator itr;
    Poco::Mutex::ScopedLock lock(m_EvtLwmMutex);
    event_listeners_t listeners = m_LwmListeners[_szTag];
    itr = std::find(listeners.begin(), listeners.end(), _pLwmListener);

    if (itr != listeners.end()) {
        listeners.erase(itr);
    }

    if (listeners.empty()) {
        MLOG(Debug, "Unregistering from LWM for: " << _szTag << ".");

        try {
            m_ObloClient.Unsubscribe(_szTag + m_szStatusTopic);
        } catch (const Poco::RuntimeException& exc) {
            std::stringstream ss;
            ss << "Error while unregistering from events: " << exc.message();
            throw Poco::RuntimeException(ss.str());
        }

        MLOG(Debug, "Unregistered from LWM for: " << _szTag << ".");
    }

    m_LwmListeners[_szTag] = listeners;
    MLOG(Information, "LWM listener for: " << _szTag << " cleaned.");
}

void CMessenger::Request(CRequestMessage::Ptr _pRequest, const std::string& _szTag, const std::string& _szFunctionality) {
    if (m_ObloClient.IsConnected()) {
        _pRequest->SetSender(m_szUserTag);
        std::string topic;
        topic = _szFunctionality.empty() ? _szTag : _szTag + TOPIC_SEPARATOR + _szFunctionality;
        topic += m_szRequestTopic;
        Mqtt::CMessage::Ptr pubmsg = new Mqtt::CMessage();
        std::string& payload = pubmsg->GetPayloadHolder();
        _pRequest->ToJson(m_eProtocolVersion, payload);
        pubmsg->SetQos(Mqtt::CMessage::QOS0);
        MLOG(Trace, "Sending request: " << payload << " to: " << topic << ".");
        Poco::Mutex::ScopedLock lock(m_RequestsMutex);

        try {
            m_ObloClient.Publish(topic, pubmsg);
        } catch (const Poco::RuntimeException& exc) {
            std::stringstream ss;
            ss << "Error while sending request: " << exc.message();
            throw Poco::RuntimeException(ss.str());
        }

        // Clean if there was a request with the same ID.
        CleanRequest(_pRequest->GetId());
        m_Requests[_pRequest->GetId()] = _pRequest;
        MLOG(Debug, "Request sent (d=" << _szTag << ", n=" << _pRequest->GetId() << ")");
    } else {
        throw Poco::IllegalStateException(ERR_MSGR_NOT_CONNECTED);
    }
}

void CMessenger::CleanRequest(unsigned int _nReqId) {
    requests_t::iterator itr;
    CRequestMessage::Ptr pRequest;
    Poco::Mutex::ScopedLock lock(m_RequestsMutex);
    itr = m_Requests.find(_nReqId);

    if (itr != m_Requests.end()) {
        MLOG(Trace, "Cleaning request: " << _nReqId);
        pRequest = itr->second;
        m_Requests.erase(itr);
        pRequest->ReportFailure();
        MLOG(Debug, "Request (" << _nReqId << ") cleaned.");
    }
}

void CMessenger::CleanRequests() {
    MLOG(Trace, "Cleaning all requests.");
    m_RequestsMutex.lock();
    m_Requests.clear();
    m_RequestsMutex.unlock();
    MLOG(Debug, "Requests are cleaned.");
}

CUser::Ptr CMessenger::GetUser(const std::string& _szTag) {
    CUser::Ptr obj = new CUser(_szTag);
    obj->SetMessenger(this);
    return obj;
}

void CMessenger::RegisterConnectionListener(Messenger::IConnectionListener* _pConnectionListener) {
    m_ObloClient.RegisterConnectionListener(_pConnectionListener);
}

void CMessenger::UnregisterConnectionListener() {
    m_ObloClient.RegisterConnectionListener(NULL);
}

void CMessenger::RegisterDeafultListener(IEventListener* _pListener) {
    m_pDefaultListener = _pListener;
}

void CMessenger::ProcessEvent(const Mqtt::CMessage& _Msg, const std::string& _szSenderTag, const Poco::DynamicStruct::Ptr _pPayload) {
    MLOG(Trace, "Processing an event message from: " << _szSenderTag << ".");
    event_listener_map_t::iterator itr;
    Poco::Mutex::ScopedLock lock(m_EvtLwmMutex);
    itr = m_EventListeners.find(_szSenderTag);
    event_listeners_t listeners;

    if (itr != m_EventListeners.end()) {
        listeners = itr->second;
    } else if (m_pDefaultListener) {
        listeners.push_back(m_pDefaultListener);
    }

    if (!listeners.empty()) {
        CEventMessage::Ptr event;
        event = CEventMessage::Parse(*_pPayload, m_eProtocolVersion);
        InspectRetainedFlag(_Msg.IsRetained(),  "evt:" + event->GetName(), _szSenderTag);
        InspectQoS(_Msg.GetQos(), Mqtt::CMessage::QOS1,  "evt:" + event->GetName(), _szSenderTag);
        MLOG(Trace, "Event (" << event->GetName() << ") arrived from: " << _szSenderTag << ".");
        event_listeners_t::iterator eventItr;

        for (eventItr = listeners.begin(); eventItr != listeners.end(); ++eventItr) {
            IEventListener* lsnr = *eventItr;
            lsnr = lsnr ? lsnr : m_pDefaultListener; // if null, set default listener.

            if (lsnr) {
                lsnr->OnEvent(_szSenderTag, event);
            }
        }

        MLOG(Debug, "Event processed (s=" << _szSenderTag << ", n='" << event->GetName() << "')");
    }
}

void CMessenger::ProcessLastwill(const Mqtt::CMessage& _Msg, const std::string& _szSenderTag, const Poco::DynamicStruct::Ptr _pPayload) {
    MLOG(Trace, "Processing a lastwill message from: " << _szSenderTag << ".");
    event_listener_map_t::iterator itr;
    Poco::Mutex::ScopedLock lock(m_EvtLwmMutex);
    itr = m_LwmListeners.find(_szSenderTag);
    event_listeners_t listeners;

    if (itr != m_LwmListeners.end()) {
        listeners = itr->second;
    } else if (m_pDefaultListener) {
        listeners.push_back(m_pDefaultListener);
    }

    if (!listeners.empty()) {
        CLastwillMessage::Ptr lastwill;
        lastwill = CLastwillMessage::Parse(*_pPayload, m_eProtocolVersion);
        InspectRetainedFlag(_Msg.IsRetained(), "LWM", _szSenderTag);
        InspectQoS(_Msg.GetQos(), Mqtt::CMessage::QOS1, "LWM", _szSenderTag);
        MLOG(Trace, "Lastwill (online status: " << lastwill->IsOnline() << ") arrived from: " << _szSenderTag << ".");
        event_listeners_t::iterator eventItr;

        for (eventItr = listeners.begin(); eventItr != listeners.end(); ++eventItr) {
            IEventListener* lsnr = *eventItr;
            lsnr = lsnr ? lsnr : m_pDefaultListener; // if null, set default listener.

            if (lsnr) {
                lsnr->OnLastwill(_szSenderTag, lastwill);
            }
        }

        MLOG(Debug, "Lastwill processed (s=" << _szSenderTag << ", s=" << lastwill->IsOnline() << ")");
    }
}

void CMessenger::ProcessRequest(const Mqtt::CMessage& _Msg, const Poco::DynamicStruct::Ptr _pPayload, const string& _szFunctionality) {
    MLOG(Trace, "Processing a request.");

    if (m_pResponder != NULL) {
        CRequestMessage::Ptr req;
        req = CRequestMessage::Parse(*_pPayload, m_eProtocolVersion);
        InspectRetainedFlag(_Msg.IsRetained(),  "req:" + req->GetName(), req->GetSender());
        InspectQoS(_Msg.GetQos(), Mqtt::CMessage::QOS0,  "req:" + req->GetName(), req->GetSender());
        MLOG(Trace, "Request arrived from: " << req->GetSender() << ".");
        CResponseMessage::Ptr response = NULL;

        if (_szFunctionality.empty()) {
            response = m_pResponder->Respond(req);
        } else {
            response = m_pResponder->FunctionalityRespond(_szFunctionality, req);
        }

        if (!response.isNull()) {
            response->SetId(req->GetId());
            SendResponse(req->GetSender(), response);
        }

        MLOG(Debug, "Request processed (s=" << req->GetSender() << ", n='" << req->GetName() << "', id=" << req->GetId() << ")");
    }
}

void CMessenger::ProcessFailedRequest(const Poco::DynamicStruct::Ptr _pPayload) {
    MLOG(Trace, "Processing a failed request.");
    CRequestMessage::Ptr req;
    req = CRequestMessage::Parse(*_pPayload, m_eProtocolVersion);
    MLOG(Trace, "Failure for request with ID: " << req->GetId() << ".");
    CleanRequest(req->GetId());
}

void CMessenger::ProcessResponse(const Mqtt::CMessage& _Msg, const Poco::DynamicStruct::Ptr _pPayload) {
    MLOG(Trace, "Processing a response.");
    CResponseMessage::Ptr rsp;
    rsp = CResponseMessage::Parse(*_pPayload, m_eProtocolVersion);
    m_RequestsMutex.lock();
    requests_t::iterator itr = m_Requests.find(rsp->GetId());

    if (itr != m_Requests.end()) {
        CRequestMessage::Ptr req = itr->second;
        m_Requests.erase(itr);
        m_RequestsMutex.unlock();
        InspectRetainedFlag(_Msg.IsRetained(), "rsp:" + req->GetName(), "");
        InspectQoS(_Msg.GetQos(), Mqtt::CMessage::QOS0, "rsp:" + req->GetName(), "");
        req->ReportResponse(rsp);
        MLOG(Debug, "Response processed (n='" << req->GetName() << "', id=" << rsp->GetId() << ")");
    } else {
        m_RequestsMutex.unlock();
    }
}

void CMessenger::SendResponse(const std::string& _szReceiverTag, Messenger::CResponseMessage::Ptr _pResponse) {
    std::string topic;

    if (_pResponse.isNull()) {
        throw Poco::NullPointerException("Response pointer is not valid.");
    }

    topic = _szReceiverTag + m_szResponseTopic;
    Mqtt::CMessage::Ptr pubmsg = new Mqtt::CMessage();
    std::string& payload = pubmsg->GetPayloadHolder();
    _pResponse->ToJson(m_eProtocolVersion, payload);
    MLOG(Trace, "Sending response for request with ID: " << _pResponse->GetId() << ".");
    pubmsg->SetQos(Mqtt::CMessage::QOS0);

    if (payload.length() < 262144) { // do not print payload larger than 256kB
        MLOG(Trace, "Sending response: " << payload << " to: " << topic << ".");
    }

    try {
        m_ObloClient.Publish(topic, pubmsg);
    } catch (const Poco::RuntimeException& exc) {
        std::stringstream ss;
        ss << "Error while sending response: " << exc.message();
        throw Poco::RuntimeException(ss.str());
    }

    MLOG(Debug, "Response message sent (d=" << _szReceiverTag << ", id=" << _pResponse->GetId() << ")");
}

bool StartsWith(const std::string& _szString, const std::string& _szStarting) {
    unsigned int stringSize = _szString.length();
    unsigned int startingSize = _szStarting.length();

    if (stringSize >= startingSize) {
        return (0 == _szString.compare(0, startingSize, _szStarting));
    } else {
        return false;
    }
}

bool EndsWith(const std::string& _szString, const std::string& _szEnding) {
    unsigned int stringSize = _szString.length();
    unsigned int endingSize = _szEnding.length();

    if (stringSize >= endingSize) {
        return (0 == _szString.compare(stringSize - endingSize, endingSize, _szEnding));
    } else {
        return false;
    }
}

void CMessenger::OnMessageArrived(const std::string& _szTopic, const Mqtt::CMessage::Ptr _pMsg) {
    std::string topic = _szTopic;
    Poco::DynamicStruct::Ptr payload;
    MLOG(Trace, "Message arrived: " << _pMsg->GetPayload());

    try {
        payload = ParseObject(_pMsg->GetPayload());
    } catch (const Poco::RuntimeException& e) {
        MLOG(Error, e.message());
        return;
    }

    // Protect processing, in case that someone wants to close the connection
    Poco::RWLock::ScopedReadLock lock(m_ProcessingSync);

    try {
        if (EndsWith(topic, m_szEventTopic)) {
            topic = Poco::replace(topic, m_szEventTopic, EMPTY_STRING);
            ProcessEvent(*_pMsg, topic, payload);
        } else if (EndsWith(topic, m_szRequestTopic)) {
            topic = Poco::replace(topic, m_szRequestTopic, EMPTY_STRING);

            if (StartsWith(topic, m_szUserTag)) {
                // Trim user tag.
                std::string functionality = Poco::replace(topic, m_szUserTag, EMPTY_STRING);

                if (!functionality.empty()) {
                    functionality.replace(0, 1, EMPTY_STRING); // erase "/"
                }

                ProcessRequest(*_pMsg, payload, functionality);
            } else {
                MLOG(Warning, "Unhandled request arrived");
            }
        } else if (EndsWith(topic, m_szResponseTopic)) {
            topic = Poco::replace(topic, m_szResponseTopic, EMPTY_STRING);

            if (topic == m_szUserTag) {
                ProcessResponse(*_pMsg, payload);
            } else {
                MLOG(Warning, "Unhandled response arrived");
            }
        } else if (EndsWith(topic, m_szStatusTopic)) {
            topic = Poco::replace(topic, m_szStatusTopic, EMPTY_STRING);
            ProcessLastwill(*_pMsg, topic, payload);
        } else {
            MLOG(Warning, "Unhandled message arrived on: " << topic << ".");
        }
    } catch (const Poco::DataFormatException& e) {
        MLOG(Error, e.message() << " (t=" << _szTopic << ", p=" << _pMsg->GetPayload() << ")");
    } catch (const Poco::Exception& e) {
        MLOG(Error, e.displayText());
    }
}

void CMessenger::OnPublishFailed(int _nCode, const string& _szTopic, const Mqtt::CMessage::Ptr _pMsg) {
    if (EndsWith(_szTopic, m_szRequestTopic)) {
        Poco::DynamicStruct::Ptr payload;

        try {
            payload = ParseObject(_pMsg->GetPayload());
        } catch (const Poco::RuntimeException& e) {
            MLOG(Error, e.message());
            return;
        }

        try {
            ProcessFailedRequest(payload);
        } catch (const Poco::RuntimeException& e) {
            MLOG(Error, e.message());
        }
    } else {
        std::stringstream ss;
        ss << "Publish failed (t=" << _szTopic << ")";
        OnTrace(ss.str());
    }
}

Poco::DynamicStruct::Ptr CMessenger::ParseObject(const string& _szObject) {
    Poco::JSON::Parser parser;
    Poco::DynamicStruct::Ptr object;

    try {
        Poco::Dynamic::Var payload = parser.parse(_szObject);
        Poco::JSON::Object::Ptr jsonObj = payload.extract<Poco::JSON::Object::Ptr>();
        object = new Poco::DynamicStruct(*jsonObj);
    } catch (const Poco::Exception& e) {
        std::stringstream ss;
        ss << "Error while parsing a JSON string: " << _szObject << ".";
        throw Poco::RuntimeException(ss.str());
    }

    return object;
}

void CMessenger::InspectRetainedFlag(bool _bRetained, const string& _szName, const string& _szSender) {
    if (_bRetained) {
        stringstream ss;
        ss << "RETAINED message detected (n='" << _szName << "', sender=" << _szSender << ")";
        OnWarning(ss.str());
    }
}

void CMessenger::InspectQoS(Mqtt::CMessage::QosLevel _eQos, Mqtt::CMessage::QosLevel _eQosExpected, const string& _szName, const string& _szSender) {
    if (_eQos != _eQosExpected) {
        stringstream ss;
        ss << "INVALID QoS detected (n='" << _szName << "', sender=" << _szSender
           << ", qos=" << _eQos << ", expected=" << _eQosExpected << ")";
        OnWarning(ss.str());
    }
}

}
