#include "messenger/MqttClient.h"
#include "messenger/Common.h"

#include <cstring>
#include <sstream>

namespace Mqtt {

CClientImpl::CClientImpl(const std::string& _szServerUri, const std::string& _szClientId):
    m_szClientId(_szClientId),
    m_pStatusListener(NULL) {
    MQTTClient_create(&m_Cli, _szServerUri.c_str(), m_szClientId.c_str(), 0, NULL);
}

CClientImpl::~CClientImpl() {
    MQTTClient_destroy(&m_Cli);
}

void CClientImpl::Connect(CConnectConfig::Ptr _pConfig) {
    if ( !IsConnected() ) {
        m_bSessionPresent = false;
        MQTTClient_connectOptions* pOptions;

        if ( !_pConfig->GetSsl().isNull() && _pConfig->GetServerURI().compare(0, 6, "ssl://")!=0 ) {
            throw Poco::InvalidArgumentException("URI must start with <ssl://> in order to use secure connection.");
        }

        pOptions = _pConfig->PackConfig();
        int rc = MQTTClient_connect(m_Cli, pOptions);
        bool bSessionPresent = pOptions->returned.sessionPresent;
        CConnectConfig::FreeConfig(pOptions);

        if (rc != MQTTCLIENT_SUCCESS) {
            MQTTClient_disconnect(m_Cli, 100);
            std::string error = GetErrorDescription(rc);
            throw Poco::RuntimeException(error, rc);
        }

        m_bSessionPresent = bSessionPresent;
    }
}

bool CClientImpl::IsSessionPresent() const {
    return m_bSessionPresent;
}

void CClientImpl::Disconnect(int _nTimeout) {
    if ( IsConnected() ) {
        int rc = MQTTClient_disconnect(m_Cli, _nTimeout);

        if (rc != MQTTCLIENT_SUCCESS) {
            std::string error = GetErrorDescription(rc);
            throw Poco::RuntimeException(error, rc);
        }
    }
}

std::string CClientImpl::GetClientId() const {
    return m_szClientId;
}

bool CClientImpl::IsConnected() const {
    return MQTTClient_isConnected(m_Cli) != 0;
}

void CClientImpl::PublishSync(const std::string& _szTopic, const CMessage::Ptr _pMsg) {
    MQTTClient_message* msg;
    _pMsg->WriteMessage(&msg);
    int rc = MQTTClient_publishMessage(m_Cli, _szTopic.c_str(), msg, NULL);
    CMessage::FreeMessage(&msg);

    if (rc != MQTTCLIENT_SUCCESS) {
        m_pStatusListener->PublishFailed(rc, _szTopic, _pMsg);
    }
}

void CClientImpl::RegisterListener(IStatusListener* _pListener) {
    Poco::Mutex::ScopedLock lock(m_Mutex);
    m_pStatusListener = _pListener;
    int rc = MQTTClient_setCallbacks(m_Cli, this,
                                     &CClientImpl::OnConnectionLost,
                                     &CClientImpl::OnMessageArrived,
                                     &CClientImpl::OnDeliveryComplete);

    if (rc != MQTTCLIENT_SUCCESS) {
        std::string error = GetErrorDescription(rc);
        throw Poco::RuntimeException(error, rc);
    }
}

char** GetPahoTopics(const CClientImpl::TopicCollection& _Topics) {
    CClientImpl::TopicCollection::const_iterator itr;
    char** topics = static_cast<char**>(malloc(_Topics.size()*sizeof(char*)));
    int i = 0;

    // Write topic
    for (itr=_Topics.begin(); itr!=_Topics.end(); ++itr, ++i) {
        std::string szTopic = (*itr);
        char* topic = strdup(szTopic.c_str());
        topics[i] = topic;
    }

    return topics;
}

void FreePahoTopics(char** _Topics, int _nCnt) {
    for (int i=0; i<_nCnt; ++i) {
        char* topic = _Topics[i];

        if ( topic ) {
            free(topic);
        }
    }

    if ( _Topics ) {
        free(_Topics);
    }
}

int* GetPahoQoss(const CClientImpl::QosCollection& _Qos) {
    CClientImpl::QosCollection::const_iterator itr;
    int* qoss = static_cast<int*>(malloc(_Qos.size()*sizeof(int)));
    int i = 0;

    // Write qos
    for (itr=_Qos.begin(); itr!=_Qos.end(); ++itr, ++i) {
        qoss[i] = (*itr);
    }

    return qoss;
}

void FreePahoQoss(int* _Qoss) {
    if ( _Qoss ) {
        free(_Qoss);
    }
}

void CClientImpl::Subscribe(const CClientImpl::TopicCollection& _TopicFilters,
                            const CClientImpl::QosCollection& _Qos) {
    unsigned int n = _TopicFilters.size();

    if ( n==0 ) {
        return;
    }

    if ( n!=_Qos.size()) {
        throw Poco::InvalidArgumentException("Collection sizes don't match");
    }

    char** topics = GetPahoTopics(_TopicFilters);
    int* qoss = GetPahoQoss(_Qos);
    int rc = MQTTClient_subscribeMany(m_Cli, n, topics, qoss);
    FreePahoTopics(topics, n);
    FreePahoQoss(qoss);

    if (rc != MQTTCLIENT_SUCCESS) {
        std::string error = GetErrorDescription(rc);
        throw Poco::RuntimeException(error, rc);
    }
}

void CClientImpl::Subscribe(const std::string& _szTopicFilter, int _nQos) {
    int rc = MQTTClient_subscribe(m_Cli, _szTopicFilter.c_str(), _nQos);

    if (rc != MQTTCLIENT_SUCCESS) {
        std::string error = GetErrorDescription(rc);
        throw Poco::RuntimeException(error, rc);
    }
}

void CClientImpl::Unsubscribe(const std::string& _szTopicFilter) {
    int rc = MQTTClient_unsubscribe(m_Cli, _szTopicFilter.c_str());

    if (rc != MQTTCLIENT_SUCCESS) {
        std::string error = GetErrorDescription(rc);
        throw Poco::RuntimeException(error, rc);
    }
}

void CClientImpl::Unsubscribe(const CClientImpl::TopicCollection& _TopicFilters) {
    int n = _TopicFilters.size();

    if ( n==0 ) {
        return;
    }

    char** topics = GetPahoTopics(_TopicFilters);
    int rc = MQTTClient_unsubscribeMany(m_Cli, n, topics);
    FreePahoTopics(topics, n);

    if (rc != MQTTCLIENT_SUCCESS) {
        std::string error = GetErrorDescription(rc);
        throw Poco::RuntimeException(error, rc);
    }
}

void CClientImpl::OnConnectionLost(void* _pContext, char* _szCause) {
    if (_pContext) {
        CClientImpl* cli = static_cast<CClientImpl*>(_pContext);
        IStatusListener* cb = cli->GetStatusListener();

        if (cb) {
            cb->ConnectionLost(_szCause ? std::string(_szCause) : std::string());
        }
    }
}

int CClientImpl::OnMessageArrived(void* _pContext, char* _szTopicName,
                                  int _nTopicLen, MQTTClient_message* _pMsg) {
    if (_pContext) {
        CClientImpl* cli = static_cast<CClientImpl*>(_pContext);
        IStatusListener* cb = cli->GetStatusListener();

        if (cb) {
            std::string topic;

            if (_nTopicLen) {
                topic = std::string(_szTopicName, _szTopicName+_nTopicLen);
            } else {
                topic = std::string(_szTopicName);
            }

            CMessage::Ptr msg = CMessage::ParseMessage(_pMsg);
            cb->MessageArrived(topic, msg);
        }
    }

    MQTTClient_freeMessage(&_pMsg);
    MQTTClient_free(_szTopicName);
    return (!0);
}

void CClientImpl::OnDeliveryComplete(void* _pContext, MQTTClient_token _Token) {
    // TODO Check this.
}

IStatusListener* CClientImpl::GetStatusListener() const {
    Poco::Mutex::ScopedLock lock(m_Mutex);
    return m_pStatusListener;
}

#define MQTT_NOT_AUTHORIZED 5
std::string CClientImpl::GetErrorDescription(int _nErrorCode) {
    std::string desc;

    switch (_nErrorCode) {
        case MQTTCLIENT_SUCCESS:
            desc = "Successful completion";
            break;

        case MQTTCLIENT_FAILURE:
            desc = "Failure";
            break;

        case MQTTCLIENT_DISCONNECTED:
            desc = "The client is disconnected";
            break;

        case MQTTCLIENT_MAX_MESSAGES_INFLIGHT:
            desc = "The maximum number of in-flight messages has been reached";
            break;

        case MQTTCLIENT_BAD_UTF8_STRING:
            desc = "An invalid UTF-8 string";
            break;

        case MQTTCLIENT_NULL_PARAMETER:
            desc = "A NULL parameter has been supplied";
            break;

        case MQTTCLIENT_TOPICNAME_TRUNCATED:
            desc = "Topic string includes embedded NULL characters";
            break;

        case MQTTCLIENT_BAD_STRUCTURE:
            desc = "A structure parameter does not have the correct eyecatcher and version number";
            break;

        case MQTTCLIENT_BAD_QOS:
            desc = "A QoS value that falls outside of the acceptable range (0, 1, 2)";
            break;

        case MQTT_BAD_SUBSCRIBE:
            desc = "Bad subscribe";
            break;

        case MQTT_NOT_AUTHORIZED:
            desc = "Bad username and password";
            break;

        default:
            std::stringstream ss;
            ss << "Unknown error (" << _nErrorCode << ")";
            desc = ss.str();
            break;
    }

    return desc;
}

}
