#include "messenger/MqttConfiguration.h"

namespace Mqtt {

#define CONNECTION_INITIALIZER MQTTClient_connectOptions_initializer
#define SECURE_INITIALIZER MQTTClient_SSLOptions_initializer
#define WILL_INITIALIZER MQTTClient_willOptions_initializer

const MQTTClient_connectOptions& InitConnectionConfig() {
    static MQTTClient_connectOptions connConfig = CONNECTION_INITIALIZER;
    return connConfig;
}

const MQTTClient_SSLOptions& InitSecureConnectionConfig() {
    static MQTTClient_SSLOptions secConfig = SECURE_INITIALIZER;
    return secConfig;
}

const MQTTClient_willOptions& InitWillConfig() {
    static MQTTClient_willOptions willConfig = WILL_INITIALIZER;
    return willConfig;
}

std::string CWillConfig::GetTopic() const {
    return m_szTopic;
}

void CWillConfig::SetTopic(const std::string& _szTopic) {
    m_szTopic = _szTopic;
}

CMessage::Ptr CWillConfig::GetMessage() const {
    return m_pMsg;
}

void CWillConfig::SetMessage(const CMessage::Ptr _pMsg) {
    m_pMsg = _pMsg;
}

MQTTClient_willOptions* CWillConfig::PackConfig() const {
    MQTTClient_willOptions* config;
    config = new MQTTClient_willOptions(InitWillConfig());
    config->topicName = Messenger::DuplicateString(m_szTopic);
    config->message = Messenger::DuplicateString(m_pMsg->GetPayload());
    config->retained = m_pMsg->IsRetained() ? 1 : 0;
    config->qos = (int) m_pMsg->GetQos();
    return config;
}

void CWillConfig::FreeConfig(MQTTClient_willOptions* _pConfig) {
    MQTTClient_willOptions* config = _pConfig;
    SAFE_DELETE_ARRAY_O(config->message);
    SAFE_DELETE_ARRAY_O(config->topicName);
    SAFE_DELETE_O(config);
}

CSslConfig::CSslConfig(): m_bVerifyServer(true) {
}

std::string CSslConfig::GetTrustStore() const {
    return m_szTrustStore;
}

void CSslConfig::SetTrustStore(const std::string& _szTrustStore) {
    m_szTrustStore = _szTrustStore;
}

std::string CSslConfig::GetKeyStore() const {
    return m_szKeyStore;
}

void CSslConfig::SetKeyStore(const std::string& _szKeyStore) {
    m_szKeyStore = _szKeyStore;
}

std::string CSslConfig::GetPrivateKey() const {
    return m_szPrivateKey;
}

void CSslConfig::SetPrivateKey(const std::string& _szPrivateKey) {
    m_szPrivateKey = _szPrivateKey;
}

std::string CSslConfig::GetPrivateKeyPassword() const {
    return m_szPrivateKeyPassword;
}

void CSslConfig::SetPrivateKeyPassword(const std::string& _szPrivateKeyPassword) {
    m_szPrivateKeyPassword = _szPrivateKeyPassword;
}

bool CSslConfig::GetVerifyServer() const {
    return m_bVerifyServer;
}

void CSslConfig::SetVerifyServer(bool _bVerifyServer) {
    m_bVerifyServer = _bVerifyServer;
}

MQTTClient_SSLOptions* CSslConfig::PackConfig() const {
    MQTTClient_SSLOptions* config;
    config = new MQTTClient_SSLOptions(InitSecureConnectionConfig());
    config->trustStore = m_szTrustStore.empty() ? NULL : Messenger::DuplicateString(m_szTrustStore);
    config->keyStore = m_szKeyStore.empty() ? NULL : Messenger::DuplicateString(m_szKeyStore);
    config->privateKey = m_szPrivateKey.empty() ? NULL : Messenger::DuplicateString(m_szPrivateKey);
    config->privateKeyPassword = m_szPrivateKeyPassword.empty() ? NULL : Messenger::DuplicateString(m_szPrivateKeyPassword);
    config->enableServerCertAuth = m_bVerifyServer ? 1 : 0;
    return config;
}

void CSslConfig::FreeConfig(MQTTClient_SSLOptions* _pConfig) {
    MQTTClient_SSLOptions* config = _pConfig;
    SAFE_DELETE_ARRAY_O(config->privateKeyPassword);
    SAFE_DELETE_ARRAY_O(config->privateKey);
    SAFE_DELETE_ARRAY_O(config->keyStore);
    SAFE_DELETE_ARRAY_O(config->trustStore);
    SAFE_DELETE_O(config);
}

CConnectConfig::CConnectConfig(const std::string& _szUsername,
                               const std::string& _szPassword,
                               bool _bCleanSession,
                               int _nKeepAliveInterval,
                               int _nTimeout,
                               int _nRetryinterval,
                               bool _bReliable):
    m_pWill(NULL),
    m_pSsl(NULL) {
    m_szUsername = _szUsername;
    m_szPassword = _szPassword;
    m_bCleanSession = _bCleanSession;
    m_nKeepAliveInterval = _nKeepAliveInterval;
    m_nTimeout = _nTimeout;
    m_nRetryInterval = _nRetryinterval;
    m_bReliable = _bReliable;
}

std::string CConnectConfig::GetServerURI() const {
    return m_szServerURI;
}

void CConnectConfig::SetServerURI(const std::string& _szServerURI) {
    m_szServerURI = _szServerURI;
}

int CConnectConfig::GetKeepAliveInterval() const {
    return m_nKeepAliveInterval;
}

void CConnectConfig::SetKeepAliveInterval(int _nKeepAliveInterval) {
    m_nKeepAliveInterval = _nKeepAliveInterval;
}

bool CConnectConfig::GetCleanSession() const {
    return m_bCleanSession;
}

void CConnectConfig::SetCleanSession(bool _bCleanSession) {
    m_bCleanSession = _bCleanSession;
}

bool CConnectConfig::IsReliable() const {
    return m_bReliable;
}

void CConnectConfig::SetReliable(bool _bReliable) {
    m_bReliable = _bReliable;
}

CWillConfig::Ptr CConnectConfig::GetWill() const {
    return m_pWill;
}

void CConnectConfig::SetWill(const CWillConfig::Ptr& _pWill) {
    m_pWill = _pWill;
}

std::string CConnectConfig::GetUsername() const {
    return m_szUsername;
}

void CConnectConfig::SetUsername(const std::string& _szUsername) {
    m_szUsername = _szUsername;
}

std::string CConnectConfig::GetPassword() const {
    return m_szPassword;
}

void CConnectConfig::SetPassword(const std::string& _szPassword) {
    m_szPassword = _szPassword;
}

int CConnectConfig::GetTimeout() const {
    return m_nTimeout;
}

void CConnectConfig::SetTimeout(int _nTimeout) {
    m_nTimeout = _nTimeout;
}

int CConnectConfig::GetRetryInterval() const {
    return m_nRetryInterval;
}

void CConnectConfig::SetRetryInterval(int _nRetryInterval) {
    m_nRetryInterval = _nRetryInterval;
}

CSslConfig::Ptr CConnectConfig::GetSsl() const {
    return m_pSsl;
}

void CConnectConfig::SetSsl(const CSslConfig::Ptr& _pSsl) {
    m_pSsl = _pSsl;
}

MQTTClient_connectOptions* CConnectConfig::PackConfig() const {
    MQTTClient_connectOptions* config;
    config = new MQTTClient_connectOptions(InitConnectionConfig());
    config->MQTTVersion = MQTTVERSION_3_1_1;
    config->keepAliveInterval = m_nKeepAliveInterval;
    config->cleansession = m_bCleanSession ? 1 : 0;
    config->reliable = m_bReliable ? 1 :0;

    if ( !m_pWill.isNull() ) {
        config->will = m_pWill->PackConfig();
    }

    if( !m_szServerURI.empty() ) {
        char** serverURIs = new char*[1];
        serverURIs[0] = Messenger::DuplicateString(m_szServerURI);
        config->serverURIcount = 1;
        config->serverURIs = serverURIs;
    }

    config->username = Messenger::DuplicateString(m_szUsername);
    config->password = Messenger::DuplicateString(m_szPassword);
    config->connectTimeout = m_nTimeout;
    config->retryInterval = m_nRetryInterval;

    if ( !m_pSsl.isNull() ) {
        config->ssl = m_pSsl->PackConfig();
    }

    return config;
}

void CConnectConfig::FreeConfig(MQTTClient_connectOptions* _pConfig) {
    MQTTClient_connectOptions* config = _pConfig;

    if ( config->ssl ) {
        CSslConfig::FreeConfig(config->ssl);
    }

    SAFE_DELETE_ARRAY_O(config->password);
    SAFE_DELETE_ARRAY_O(config->username);

    if ( config->will ) {
        CWillConfig::FreeConfig(config->will);
    }

    for (int i=0; i<config->serverURIcount; ++i) {
        delete [] config->serverURIs[i];
    }

    SAFE_DELETE_ARRAY_O(config->serverURIs);
    SAFE_DELETE_O(config);
}

}
