#include "messenger/Network.h"
#include <Poco/Timespan.h>
#include <sstream>

#define CLILOG(lvl, msg) SMLOG(m_pLogListener, lvl, msg)

namespace Messenger {

Poco::Random CResurector::m_RandomGenerator;

CResurector::CResurector(IConnector& _Connector, unsigned int _nInterval):
    m_Connector(_Connector),
    m_nInterval(_nInterval*1000),
    m_nMinimum(0),
    m_nMaximum(0),
    m_pTimer(NULL),
    m_pLogListener(NULL) {
}

CResurector::~CResurector() {
    m_nMinimum = 0; // Stop further scheduling
    SAFE_DELETE_O(m_pTimer);
}

void CResurector::SetInterval(unsigned int _nInterval) {
    m_nInterval = _nInterval*1000;
}

void CResurector::SetIntervalRange(unsigned int _nMinimum, unsigned int _nMaximum) {
    m_nMinimum = _nMinimum*1000;
    m_nMaximum = _nMaximum*1000;
    // Set interval to minimum
    m_nInterval = _nMinimum*1000;
}

void CResurector::RegisterLogListener(ILogListener* _pLogListener) {
    m_pLogListener = _pLogListener;
}

void CResurector::OnConnectionLost(const std::string& _szCause) {
    Poco::Mutex::ScopedLock lock(m_Mutex);
    std:: string cause = _szCause.empty() ? "unknown cause" : _szCause;
    CLILOG(Information, "Connection lost reported: " << cause);
    m_DisconnectionTime.update();

    if ( m_pTimer && m_pTimer->getPeriodicInterval()==0 ) {
        SAFE_DELETE_O(m_pTimer);
    }

    if (m_pTimer) {
        m_pTimer->restart(m_nInterval);
    } else {
        m_pTimer = new Poco::Timer();
        m_pTimer->setStartInterval(m_nInterval);
        m_pTimer->setPeriodicInterval(m_nInterval);
        m_pTimer->start(Poco::TimerCallback<CResurector>(*this, &CResurector::OnReconnect));
    }

    CLILOG(Information, "Resurector initialized.");
}

void CResurector::OnReconnect(Poco::Timer& timer) {
    Poco::Mutex::ScopedLock lock(m_Mutex);
    CLILOG(Trace, "Reconnection triggered.");

    try {
        m_Connector.Connect();
        Poco::Timespan elapsed(m_DisconnectionTime.elapsed());
        CLILOG(Information, "Offline for: " << elapsed.days() << "d "
               << elapsed.hours() << "h " << elapsed.minutes() << "m "
               << elapsed.seconds() << "s.");
        m_pTimer->restart(0);
    } catch (const Poco::Exception& e) {
        ScheduleNextRun();
        CLILOG(Debug, "Failed to reconnect (e=" + e.message() << ", next=" << m_pTimer->getPeriodicInterval() << ")");
    }
}

void CResurector::ScheduleNextRun() {
    if( m_nMinimum>0 && m_nMaximum>m_nMinimum ) {
        unsigned int next = GetRandom(m_nMinimum, m_nMaximum);
        // no need to check pointer, scheduling is called from timer's callback.
        m_pTimer->restart(next);
    }
}

unsigned int CResurector::GetRandom(unsigned int _nMinimum, unsigned int _nMaximum) {
    unsigned int range = _nMaximum - _nMinimum;
    unsigned int rnd = m_RandomGenerator.next() % range;
    rnd += _nMinimum;
    return rnd;
}

}
