#include "messenger/NetworkProxy.h"
#include <Poco/Observer.h>

namespace Mqtt {

const std::string CConnectionRunner::NO_LOST = "no_lost";

CConnectionRunner::CConnectionRunner(IConnListener& _ConnDelegate):
    m_ConnDelegate(_ConnDelegate) {
    Reinitialize();
}

void CConnectionRunner::Reinitialize() {
    m_szCause = NO_LOST;
    CSynchronizedRunner::Reinitialize();
}

void CConnectionRunner::Report(const std::string& _szCause) {
    m_szCause = _szCause;
    Continue();
}

void CConnectionRunner::DoJob() {
    if (m_szCause != NO_LOST) {
        m_ConnDelegate.OnConnectionLost(m_szCause);
        m_szCause = NO_LOST;
    } else {
        Pause();
    }
}

CRecvMessageTask::CRecvMessageTask(INetworkProxyListener& _Listener, CRecvHolder::Ptr _pRecvPacket):
    Poco::Task(""),
    m_Listener(_Listener), m_pRecvPacket(_pRecvPacket) {
}

void CRecvMessageTask::runTask() {
    std::string topic = m_pRecvPacket->GetTopic();
    Mqtt::CMessage::Ptr pMsg = m_pRecvPacket->GetMessage();
    m_Listener.OnMessageArrived(topic, pMsg);
}

CAsyncNetworkProxy::CAsyncNetworkProxy(INetworkProxyListener& _Listener):
    m_Listener(_Listener),
    m_SendingQueue(*this),
    m_RecvQueue(*this),
    m_pRecvThreadPool(NULL),
    m_pRecvTaskManager(NULL),
    m_Conn(_Listener),
    m_bStoped(false) {
    m_SendingThread.start(m_SendingQueue);
    m_RecvThread.start(m_RecvQueue);
    m_ConnThread.start(m_Conn);
}

CAsyncNetworkProxy::~CAsyncNetworkProxy() {
    m_bStoped = true;
    m_SendingQueue.Stop();
    m_RecvQueue.Stop();
    m_Conn.Stop();
    m_SendingThread.join();
    m_ConnThread.join();
    DeactivateThreadPool();
    m_RecvThread.join();
}

void CAsyncNetworkProxy::ActivateThreadPool(int _nMinThreads, int _nMaxThreads) {
    Poco::Mutex::ScopedLock lock(m_RecvMutex);

    if (m_pRecvThreadPool == NULL) {
        m_pRecvThreadPool = new Poco::ThreadPool(_nMinThreads, _nMaxThreads);
        m_pRecvTaskManager = new Poco::TaskManager(*m_pRecvThreadPool);
        Poco::Observer<CAsyncNetworkProxy, Poco::TaskFinishedNotification> obs(*this, &CAsyncNetworkProxy::HandleNotification);
        m_pRecvTaskManager->addObserver(obs);
    } else {
        int c = m_pRecvThreadPool->capacity();
        int diff = _nMaxThreads - c;
        m_pRecvThreadPool->addCapacity(diff);
    }
}

void CAsyncNetworkProxy::DeactivateThreadPool() {
    Poco::Mutex::ScopedLock lock(m_RecvMutex);

    if (m_pRecvThreadPool) {
        SAFE_DELETE_O(m_pRecvTaskManager);
        SAFE_DELETE_O(m_pRecvThreadPool);
    }
}

void CAsyncNetworkProxy::Publish(const std::string& _szTopic, const CMessage::Ptr _pMsg) {
    if (!m_bStoped) {
        // TODO add queue limit!
        CPublishHolder::Ptr holder = new CPublishHolder(_szTopic, _pMsg);
        m_SendingQueue.Add(holder);
    }
}

void CAsyncNetworkProxy::ConnectionLost(const std::string& _szCause) {
    if (!m_bStoped) {
        m_Conn.Report(_szCause);
    }
}

void CAsyncNetworkProxy::MessageArrived(const std::string& _szTopic, const CMessage::Ptr _pMsg) {
    if (!m_bStoped) {
        CRecvHolder::Ptr holder = new CRecvHolder(_szTopic, _pMsg);
        m_RecvQueue.Add(holder);
    }
}

void CAsyncNetworkProxy::DeliveryComplete(const IDeliveryToken& _Token) {
}

void CAsyncNetworkProxy::PublishFailed(int _nCode, const std::string& _szTopic, const CMessage::Ptr _pMsg) {
    if (!m_bStoped) {
        m_Listener.OnPublishFailed(_nCode, _szTopic, _pMsg);
    }
}

void CAsyncNetworkProxy::OnProcess(CPublishHolder::Ptr _Object) {
    if (!m_bStoped) {
        std::string topic = _Object->GetTopic();
        Mqtt::CMessage::Ptr pMsg = _Object->GetMessage();
        m_Listener.OnPublishRequired(topic, pMsg);
    }
}

void CAsyncNetworkProxy::OnProcess(CRecvHolder::Ptr _Object) {
    Poco::Mutex::ScopedLock lock(m_RecvMutex);

    if (m_pRecvThreadPool == NULL) {
        std::string topic = _Object->GetTopic();
        Mqtt::CMessage::Ptr pMsg = _Object->GetMessage();
        m_Listener.OnMessageArrived(topic, pMsg);
    } else {
        bool processed = false;

        while (!processed) {
            if (!m_bStoped) {
                try {
                    m_pRecvTaskManager->start(new CRecvMessageTask(m_Listener, _Object));
                    processed = true;
                } catch (const Poco::NoThreadAvailableException& e) {
                    m_RecvCondition.wait(m_RecvMutex);
                }
            } else {
                break;
            }
        }
    }
}

void CAsyncNetworkProxy::HandleNotification(Poco::TaskFinishedNotification* _pNotification) {
    Poco::Mutex::ScopedLock lock(m_RecvMutex);
    m_RecvCondition.signal();
    _pNotification->release();
}

}
