#include "AppModule.h"
#include "Globals.h"

namespace MessageBroker {

CCommandExecutor::CCommandExecutor() {
}

Poco::DynamicStruct::Ptr CCommandExecutor::ExecuteCommand(const std::string& _szCmdName, const Poco::DynamicStruct& _Params) {
    CommandHandlers::const_iterator itr;
    m_CmdMutex.lock();
    itr = m_CmdHandlers.find(_szCmdName);
    m_CmdMutex.unlock();

    if ( itr!=m_CmdHandlers.end() ) {
        CCmdExecutorDelegate::Ptr func = (*itr).second;
        return func->ExecuteCmd(_szCmdName, _Params);
    } else {
        throw Poco::NotFoundException("Command <" + _szCmdName + "> is not registered.", -1);
    }
}

Poco::DynamicStruct::Ptr CCommandExecutor::ExecuteApiCommand(const std::string& _szCmdName, const Poco::DynamicStruct& _Params) {
    CommandHandlers::const_iterator itr;
    m_CmdMutex.lock();
    itr = m_ApiCmdHandlers.find(_szCmdName);
    m_CmdMutex.unlock();

    if ( itr!=m_ApiCmdHandlers.end() ) {
        CCmdExecutorDelegate::Ptr func = (*itr).second;
        return func->ExecuteCmd(_szCmdName, _Params);
    } else {
        throw Poco::NotFoundException("Command <" + _szCmdName + "> is not registered.", -1);
    }
}

void CCommandExecutor::AddCommandExecutor(const std::string& _szCommandName, const CCmdExecutorDelegate::Ptr _pFunc) {
    CommandHandler handler;
    handler.first = _szCommandName;
    handler.second = _pFunc;
    m_CmdMutex.lock();
    m_CmdHandlers.insert(handler);
    m_CmdMutex.unlock();
}

void CCommandExecutor::AddApiCommandExecutor(const std::string& _szCommandName, const CCmdExecutorDelegate::Ptr _pFunc) {
    CommandHandler handler;
    handler.first = _szCommandName;
    handler.second = _pFunc;
    m_CmdMutex.lock();
    m_ApiCmdHandlers.insert(handler);
    m_CmdMutex.unlock();
}

CModuleManager* CAppModule::m_pModuleMngr = NULL;

CAppModule::CAppModule():
    m_bRunning(false) {
}

void CAppModule::Start() {
    m_bRunning = true;
}

void CAppModule::Stop() {
    m_bRunning = false;
}

Poco::DynamicStruct::Ptr CAppModule::SendCommand(const ID& _ModuleId, const std::string& _szCmdName, const Poco::DynamicStruct& _Params, bool _bExternalApi) {
    // Module is set on initialization, so there is no need to check if it is NULL.
    return m_pModuleMngr->SendCommand(_ModuleId, _szCmdName, _Params, _bExternalApi);
}

void CAppModule::UpdateProperty(const ID& _ModuleId,
                                const std::string& _szProperty,
                                const Poco::Dynamic::Var& _Value) {
    Poco::DynamicStruct params;
    params.insert(PARAM_SERVICE, _ModuleId);
    params.insert(PARAM_PROPERTY, _szProperty);
    params.insert(PARAM_VALUE, _Value);
    SendCommand(_ModuleId, CMD_UPDATE_PROP, params);
}

const Poco::Dynamic::Var CAppModule::GetProperty(const CAppModule::ID& _ModuleId, const std::string& _szProperty) {
    Poco::DynamicStruct params;
    params.insert(PARAM_SERVICE, _ModuleId);
    params.insert(PARAM_PROPERTY, _szProperty);
    Poco::DynamicStruct::Ptr ret = SendCommand(_ModuleId, CMD_GET_PROP, params);
    Poco::Dynamic::Var value = (*ret)[PARAM_VALUE];
    return value;
}

void CAppModule::SendEvent(CEvent* _pEvt) {
    std::stringstream ss;
    ss << "Sending event (n=" << _pEvt->GetName() << ")";
    trace(ss.str());
    m_pModuleMngr->SendEvent(_pEvt);
}

void CAppModule::ProcessEvent(const std::string& _szName, const CEvent* const _pEvt) {
}

void CAppModule::SetMenager(CModuleManager* _pModuleMngr) {
    m_pModuleMngr =_pModuleMngr;
}

CModuleManager::CModuleManager() {
    CAppModule::SetMenager(this);
}

void CModuleManager::StartModules() {
    Poco::ScopedReadRWLock lock(m_ModuleLock);
    Modules::const_iterator itr = m_Modules.begin();
    Modules::const_iterator itrEnd = m_Modules.end();

    for ( ; itr!=itrEnd; ++itr ) {
        ModuleElement elm = *itr;
        elm.second->Start();
    }
}

void CModuleManager::StopModules() {
    Poco::ScopedReadRWLock lock(m_ModuleLock);
    Modules::const_reverse_iterator itr = m_Modules.rbegin();
    Modules::const_reverse_iterator itrEnd = m_Modules.rend();

    for ( ; itr!=itrEnd; ++itr ) {
        ModuleElement elm = *itr;
        elm.second->Stop();
    }
}

Poco::DynamicStruct::Ptr CModuleManager::SendCommand(const CAppModule::ID& _ModuleId,
        const std::string& _szCmdName,
        const Poco::DynamicStruct& _Params, bool _bExternalApi) {
    m_ModuleLock.readLock();
    Modules::const_iterator itr = m_Modules.begin();
    Modules::const_iterator itrEnd = m_Modules.end();
    CAppModule::Ptr module = NULL;

    for ( ; itr!=itrEnd; ++itr ) {
        ModuleElement elm = *itr;

        if ( elm.first==_ModuleId ) {
            module = elm.second;
            break;
        }
    }

    m_ModuleLock.unlock();

    if ( !module.isNull() ) {
        return _bExternalApi ? module->ExecuteApiCommand(_szCmdName, _Params) : module->ExecuteCommand(_szCmdName, _Params);
    } else {
        throw Poco::NotFoundException("Module <" + _ModuleId + "> is not registered.", -1);
    }
}

void CModuleManager::RegisterModule(CAppModule::ID _Id, CAppModule::Ptr _pAppModule) {
    Poco::ScopedWriteRWLock lock(m_ModuleLock);
    ModuleElement elm(_Id, _pAppModule);
    m_Modules.push_back(elm);
}

void CModuleManager::UnregisterModule(CAppModule::ID _Id) {
    Poco::ScopedWriteRWLock lock(m_ModuleLock);
    Modules::iterator itr = m_Modules.begin();
    Modules::iterator itrEnd = m_Modules.end();

    for ( ; itr!=itrEnd; ++itr ) {
        ModuleElement elm = *itr;

        if ( elm.first==_Id ) {
            m_Modules.erase(itr);
            break;
        }
    }
}

void CModuleManager::OnProcess(CEvent* _pEvt) {
    Poco::ScopedReadRWLock lock(m_ModuleLock);
    std::string name = _pEvt->GetName();
    Modules::iterator itr;

    for ( itr=m_Modules.begin(); itr!=m_Modules.end(); ++itr ) {
        CAppModule::Ptr m = itr->second;
        m->ProcessEvent(name, _pEvt);
    }

    delete _pEvt;
}

CAsyncProcessor::CAsyncProcessor():
    m_AsyncProcessor(*this) {
    m_AsyncThread.start(m_AsyncProcessor);
}

CAsyncProcessor::~CAsyncProcessor() {
    m_AsyncProcessor.Stop();
    m_AsyncThread.join();
}

void CAsyncProcessor::AsyncRun(CAsyncCmd* _pCmp) {
    m_AsyncProcessor.Add(_pCmp);
}

MessageBroker::CEventDispatcher::CEventDispatcher():
    m_AsyncDispatcher(*this) {
    m_AsyncThread.start(m_AsyncDispatcher);
}

CEventDispatcher::~CEventDispatcher() {
    m_AsyncDispatcher.Stop();
    m_AsyncThread.join();
}

void CEventDispatcher::SendEvent(CEvent* _pEvt) {
    m_AsyncDispatcher.Add(_pEvt);
}

}
