/****************************************************************************
 *  Copyright (c) 2016 RT-RK d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.rt-rk.com
 *  -----------------------------------------------------
 *
 * \project Oblo Message Broker - Broker application.
 * \file AppModule.h
 * \brief
 * Application module classes.
 *
 * @Author Milan Tucic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef OMB_APP_MODULE_H
#define OMB_APP_MODULE_H

#include "Poco/DynamicStruct.h"
#include "Poco/RWLock.h"

#include "Logger.h"
#include "Database.h"
#include "messenger/Runners.h"

#include <map>

namespace MessageBroker {

#define MAKE_CMD_EXEC_DELEGATE(tp, ms) \
    MessageBroker::CCmdExecutorDelegate::CreateDelegate<tp, & tp :: ms>(this)

class CCmdExecutorDelegate {
public:
    typedef Poco::SharedPtr<CCmdExecutorDelegate> Ptr;
    CCmdExecutorDelegate():
        m_pObject(NULL),
        m_pStubExec(NULL) {
    }

    template <class T, Poco::DynamicStruct::Ptr (T::*TCmdExecuteMethod)(const std::string& _szCmdName, const Poco::DynamicStruct& _Params)>
    static CCmdExecutorDelegate::Ptr CreateDelegate(T* _pObject) {
        CCmdExecutorDelegate::Ptr d = new CCmdExecutorDelegate();
        d->m_pObject = _pObject;
        d->m_pStubExec = &MethodStubExec<T, TCmdExecuteMethod>;
        return d;
    }
    Poco::DynamicStruct::Ptr ExecuteCmd(const std::string& _szCmdName, const Poco::DynamicStruct& _Params) const {
        return (*m_pStubExec)(m_pObject, _szCmdName, _Params);
    }

private:
    typedef Poco::DynamicStruct::Ptr (*stub_type_exec)(void* m_pObject, const std::string& _szCmdName, const Poco::DynamicStruct& _Params);

    void* m_pObject;
    stub_type_exec m_pStubExec;

    template <class T, Poco::DynamicStruct::Ptr (T::*TCmdExecuteMethod)(const std::string& _szCmdName, const Poco::DynamicStruct& _Params)>
    static Poco::DynamicStruct::Ptr MethodStubExec(void* _pObject, const std::string& _szCmdName, const Poco::DynamicStruct& _Params) {
        T* p = static_cast<T*>(_pObject);
        return (p->*TCmdExecuteMethod)(_szCmdName, _Params);
    }
};

class CCommandExecutor {
public:
    CCommandExecutor();
    virtual Poco::DynamicStruct::Ptr ExecuteCommand(const std::string& _szCmdName,
            const Poco::DynamicStruct& _Params);
    virtual Poco::DynamicStruct::Ptr ExecuteApiCommand(const std::string& _szCmdName,
            const Poco::DynamicStruct& _Params);
protected:
    typedef std::pair<std::string, CCmdExecutorDelegate::Ptr> CommandHandler;
    typedef std::map<std::string, CCmdExecutorDelegate::Ptr> CommandHandlers;
    virtual void AddCommandExecutor(const std::string& _szCommandName, const CCmdExecutorDelegate::Ptr _pFunc);
    virtual void AddApiCommandExecutor(const std::string& _szCommandName, const CCmdExecutorDelegate::Ptr _pFunc);
private:
    CommandHandlers m_CmdHandlers;
    CommandHandlers m_ApiCmdHandlers;
    Poco::Mutex m_CmdMutex;
};

class CAsyncCmd {
public:
    CAsyncCmd(const std::string& _szName):
        m_szName(_szName) {
    }
    virtual ~CAsyncCmd() {}
    std::string GetName() const {
        return m_szName;
    }
private:
    std::string m_szName;
};

class CAsyncProcessor: private Messenger::IQueueProcessor<CAsyncCmd*> {
public:
    CAsyncProcessor();
    virtual ~CAsyncProcessor();
protected:
    virtual void OnProcess(CAsyncCmd* _pReq)=0;
    void AsyncRun(CAsyncCmd* _pCmp);
private:
    typedef Messenger::CQueueRunner<CAsyncCmd*> AsyncProcessor;
    AsyncProcessor m_AsyncProcessor;
    Poco::Thread m_AsyncThread;
};

class CEvent {
public:
    CEvent(const std::string& _szEvtName): m_szName(_szEvtName) {}
    virtual ~CEvent() {}

    std::string GetName() const {
        return m_szName;
    }
    virtual Poco::DynamicStruct::Ptr GetParams() const {
        return NULL;
    }
private:
    const std::string m_szName;
};

class CEventDispatcher: private Messenger::IQueueProcessor<CEvent*> {
public:
    CEventDispatcher();
    virtual ~CEventDispatcher();
    void SendEvent(CEvent* _pEvt);
protected:
    virtual void OnProcess(CEvent* _pReq)=0;
private:
    typedef Messenger::CQueueRunner<CEvent*> AsyncDispatcher;
    AsyncDispatcher m_AsyncDispatcher;
    Poco::Thread m_AsyncThread;
};

class CModuleManager;

class CAppModule: public CCommandExecutor, protected Logger::CLogger {
public:
    typedef Poco::SharedPtr<CAppModule> Ptr;
    typedef std::string ID;
    CAppModule();
    virtual ~CAppModule() {}
    virtual void Start();
    virtual void Stop();
    bool IsRunning() const {
        return m_bRunning;
    }
    virtual void GetDbTables(Database::CDatabase::Tables& _Tables) const {}
protected:
    virtual Poco::DynamicStruct::Ptr SendCommand(const CAppModule::ID& _ModuleId,
            const std::string& _szCmdName, const Poco::DynamicStruct& _Params, bool _bExternalApi=false);
    void UpdateProperty(const CAppModule::ID& _ModuleId,
                        const std::string& _szProperty, const Poco::Dynamic::Var& _Value);
    const Poco::Dynamic::Var GetProperty(const CAppModule::ID& _ModuleId, const std::string& _szProperty);
    void SendEvent(CEvent* _pEvt);
    virtual void ProcessEvent(const std::string& _szName, const CEvent* const _pEvt);
    Poco::Mutex m_ModuleMutex;
private:
    friend class CModuleManager;
    static void SetMenager(CModuleManager* _pModuleMngr);
    static CModuleManager* m_pModuleMngr;
    bool m_bRunning;
};

class CModuleManager: public CEventDispatcher {
public:
    typedef std::pair<CAppModule::ID, CAppModule::Ptr> ModuleElement;
    typedef std::list<ModuleElement> Modules;

    CModuleManager();
    virtual ~CModuleManager() {}
    void StartModules();
    void StopModules();
    Poco::DynamicStruct::Ptr SendCommand(const CAppModule::ID& _ModuleId,
                                         const std::string& _szCmdName,
                                         const Poco::DynamicStruct& _Params,
                                         bool _bExternalApi);

protected:
    void RegisterModule(CAppModule::ID _Id, CAppModule::Ptr _pAppModule);
    void UnregisterModule(CAppModule::ID _Id);

    Modules m_Modules;
    Poco::RWLock m_ModuleLock;
private:
    virtual void OnProcess(CEvent* _pEvt);
};

}

#endif
