#include "Clients.h"
#include <sstream>
#include "messenger/Messages.h"
#include "Globals.h"

namespace MessageBroker {

const std::string LASTWILL_TOPIC_TYPE = "/lwm";

CLogger::CLogger(const std::string& _szLoggerName,
                 const std::string& _szTag): m_szTag(_szTag) {
    setLoggerName(_szLoggerName);
}

void CLogger::OnTrace(const std::string& _Msg) {
    std::stringstream ss;
    ss << "(" << m_szTag << ") " << _Msg;
    trace(ss.str());
}

void CLogger::OnDebug(const std::string& _Msg) {
    std::stringstream ss;
    ss << "(" << m_szTag << ") " << _Msg;
    debug(ss.str());
}

void CLogger::OnInformation(const std::string& _Msg) {
    std::stringstream ss;
    ss << "(" << m_szTag << ") " << _Msg;
    information(ss.str());
}

void CLogger::OnWarning(const std::string& _Msg) {
    std::stringstream ss;
    ss << "(" << m_szTag << ") " << _Msg;
    warning(ss.str());
}

void CLogger::OnError(const std::string& _Msg) {
    std::stringstream ss;
    ss << "(" << m_szTag << ") " << _Msg;
    error(ss.str());
}

CClient::CClient(const std::string& _szClientId, const std::string& _szUri, const std::string& _szLoggerName, const std::string& _szTag):
    CLogger(_szLoggerName, _szTag),
    m_pConnListener(NULL), m_bSubscribeOnConnect(false), m_DnsCache(_szUri) {
    m_pResurector = new Messenger::CResurector(*this);
    m_pResurector->RegisterLogListener(this);
    m_pClient = new Mqtt::CBuffClient(_szClientId, _szUri, *this, this);
    m_pClient->RegisterConnectionListener(this);
}

CClient::~CClient() {
    SAFE_DELETE(m_pResurector);
    SAFE_DELETE(m_pClient);
}

void CClient::Start() {
    try {
        Connect();
    } catch (const Poco::Exception& exc) {
        std::stringstream ss ;
        ss << "Error on connect: " << exc.message();
        OnError(ss.str());
        m_pResurector ? m_pResurector->OnConnectionLost("inital connect failed") : void();
    }
}

void CClient::Stop() {
    try {
        Disconnect();
    } catch (const Poco::RuntimeException& exc) {
        std::stringstream ss ;
        ss << "Disconnection error: " << exc.message();
        OnError(ss.str());
    }
}

void CClient::SetConnectionListener(IConnectionStatusListener* _pCL) {
    m_pConnListener = _pCL;
}

void CClient::DeactivateResurector() {
    SAFE_DELETE(m_pResurector);
}

void CClient::SetDnsCacheLimit(int _nLimit) {
    m_DnsCache.SetCacheLimit(_nLimit);
}

void CClient::Connect() {
    try {
        // Resolve and cache address once it is connected.
        string resolved = m_DnsCache.Resolve();
        SetServerURI(resolved);

        if (m_bSubscribeOnConnect || !m_pClient->IsCleanSession()) {
            m_pClient->Connect();
        } else {
            // In case that session cleaning is requested, do basic connect (without processing stored subscriptions and publishes).
            m_pClient->ConnectBasic();
        }

        // Mark connecction as OK.
        m_DnsCache.SetServerVerified();
        OnInformation(" Connected on: " + m_pClient->GetServerUri());

        if (!m_szWillTopic.empty()) {
            Mqtt::CMessage::Ptr status = GenerateStatusMessage(true);
            m_pClient->Publish(m_szWillTopic, status);
            OnDebug("Status (lwm:online) sent.");
        }

        m_pConnListener ? m_pConnListener->OnConnected(m_szTag) : void();
    } catch (const Poco::Exception& e) {
        m_DnsCache.SetServerError();
        m_pConnListener ? m_pConnListener->OnConnectFailed(m_szTag, e.code(), e.message()) : void();
        e.rethrow();
    }
}

void CClient::Disconnect() {
    if (!m_szWillTopic.empty()) {
        Mqtt::CMessage::Ptr status = GenerateStatusMessage(false);
        m_pClient->PublishSync(m_szWillTopic, status);
        OnDebug("Status (lwm:offline) sent.");
    }

    m_pClient->Disconnect();
}

Mqtt::CMessage::Ptr CClient::GenerateStatusMessage(bool _bOnline) const {
    Mqtt::CMessage::Ptr mp = new Mqtt::CMessage();
    Messenger::CLastwillMessage lwm;
    lwm.SetStatus(_bOnline);
    mp->SetPayload(lwm.ToJson(Messenger::PROTOCOL_OBLO_1));
    mp->SetQos(Mqtt::CMessage::QOS1);
    return mp;
}

void CClient::OnConnectionLost(const std::string& _szCause) {
    m_pConnListener ? m_pConnListener->OnConnectionLost(m_szTag, _szCause) : void();
    m_pResurector ? m_pResurector->OnConnectionLost(_szCause) : void();
}

void CClient::Subscribe(const CClient::SubscriptionCollection& _Subscriptions) {
    Mqtt::CClientImpl::TopicCollection topics;
    Mqtt::CClientImpl::QosCollection qoss;
    SubscriptionCollection::const_iterator itr;

    for (itr = _Subscriptions.begin(); itr != _Subscriptions.end(); ++itr) {
        topics.push_back(itr->first);
        qoss.push_back(itr->second);
    }

    m_pClient->Subscribe(topics, qoss);
}

void CClient::Subscribe(const std::string& _szTopic, Mqtt::CMessage::QosLevel _eQos) {
    m_pClient->Subscribe(_szTopic, _eQos);
}

void CClient::Unsubscribe(const std::string& _szTopic) {
    m_pClient->Unsubscribe(_szTopic);
}

void CClient::PublishMsg(const std::string& topic, const Mqtt::CMessage& _Msg) {
    OnTrace("Publishing to: " + topic);
    m_pClient->Publish(topic, new Mqtt::CMessage(_Msg));
    OnTrace("Publish done for: " + topic);
}

void CClient::ActivateStatusMessage(const std::string& _szTag) {
    m_szWillTopic = _szTag + LASTWILL_TOPIC_TYPE;
    Mqtt::CMessage::Ptr status = GenerateStatusMessage(false);
    m_pClient->SetWillConfiguration(m_szWillTopic, status);
}

void CClient::RegisterObserver(Mqtt::CMsgDelegate::Ptr _pObserver) {
    m_ObserverMutex.lock();
    m_Observers.push_front(_pObserver);
    m_ObserverMutex.unlock();
}

void CClient::SetReconnectionInterval(unsigned int _nInterval) {
    m_pResurector ? m_pResurector->SetInterval(_nInterval) : throw
    Poco::NullPointerException("Resurector not active.");
}

void CClient::SetReconnectionInterval(unsigned int _nMinimum, unsigned int _nMaximum) {
    m_pResurector ? m_pResurector->SetIntervalRange(_nMinimum, _nMaximum) : throw
    Poco::NullPointerException("Resurector not active.");
}

void CClient::SetServerURI(const std::string& _szServerURI) {
    m_pClient->SetServerURI(_szServerURI);
}

void CClient::SetUsername(const std::string& _szUsername) {
    m_pClient->SetUsername(_szUsername);
}

void CClient::SetPassword(const std::string& _szPassword) {
    m_pClient->SetPassword(_szPassword);
}

void CClient::SetCleanSession(bool _bCleanSession) {
    m_pClient->SetCleanSession(_bCleanSession);
}

bool CClient::IsCleanSessionSet() const {
    return m_pClient->IsCleanSession();
}

void CClient::SetConnectionTimeoutInterval(int _nInterval) {
    m_pClient->SetConnectionTimeoutInterval(_nInterval);
}

void CClient::SetKeepAliveInterval(int _nInterval) {
    m_pClient->SetKeepAliveInterval(_nInterval);
}

#ifdef USE_SECURE_CONNECTION
void CClient::SetSecureConfiguration(const std::string& _szTrustStore,
                                     const std::string& _szKeyStore,
                                     const std::string& _szPrivateKey, const std::string& _szPrivateKeyPassword) {
    m_pClient->SetSecureConfiguration(_szTrustStore, _szKeyStore, _szPrivateKey,
                                      _szPrivateKeyPassword);
}
#endif

std::string CClient::GetClientId() const {
    return m_pClient->GetClientId();
}

std::string CClient::GetServerURI() const {
    return m_pClient->GetServerUri();
}

bool CClient::IsConnected() const {
    return m_pClient->IsConnected();
}

bool CClient::IsSessionPressent() const {
    return m_pClient->IsSessionPresent();
}

void CClient::IgnoreSessionPresent(bool _bIgnore) {
    m_pClient->IgnoreSessionPresent(_bIgnore);
}

void CClient::SetSubscribeOnConnect(bool _bSubscribe) {
    m_bSubscribeOnConnect = _bSubscribe;
}

void CClient::OnMessageArrived(const std::string& _szTopic, const Mqtt::CMessage::Ptr _pMsg) {
    Observers::const_iterator itr;
    Poco::Mutex::ScopedLock lock(m_ObserverMutex);
    const Mqtt::CMessage& ref = *_pMsg;

    for (itr = m_Observers.begin(); itr != m_Observers.end(); ++itr) {
        (*itr)->operator()(_szTopic, ref);
    }
}

void CClient::OnPublishFailed(int _nCode, const std::string& _szTopic, const Mqtt::CMessage::Ptr _pMsg) {
    // TODO method will be called in case of QoS0 messages
}

} // MessageBroker
