/****************************************************************************
 *  Copyright (c) 2016 RT-RK d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.rt-rk.com
 *  -----------------------------------------------------
 *
 * \project Oblo Message Broker - Broker application.
 * \file Clients.h
 * \brief
 * OBLO client classes.
 *
 * @Author Milan Tucic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef OMB_CLIENTS_H
#define OMB_CLIENTS_H

#include "Logger.h"
#include "DnsCache.h"
#include "messenger/Messenger.h"
#include "messenger/Client.h"

namespace MessageBroker {

class CLogger: protected Messenger::ILogListener, protected Logger::CLogger {
public:
    CLogger(const std::string& _szLoggerName, const std::string& _szTag = "");
protected:
    virtual void OnTrace(const std::string& _Msg);
    virtual void OnDebug(const std::string& _Msg);
    virtual void OnInformation(const std::string& _Msg);
    virtual void OnWarning(const std::string& _Msg);
    virtual void OnError(const std::string& _Msg);
    std::string m_szTag;
};

class IConnectionStatusListener {
public:
    virtual void OnConnected(const std::string& _szCliId) = 0;
    virtual void OnConnectFailed(const std::string& _szCliId, int16_t _nCode,
                                 const std::string& _szReason) = 0;
    virtual void OnConnectionLost(const std::string& _szCliId,
                                  const std::string& _szReason) = 0;
};

/**
 * @brief The CClient class
 */
class CClient: public CLogger, public Mqtt::IDeliveryListener,
    public Messenger::IConnector, public Messenger::IConnectionListener {
public:
    typedef std::pair<std::string, int> Subscription;
    typedef std::vector<Subscription> SubscriptionCollection;
    /**
     * @brief CClient constructor.
     * @param _szClientId Client identifier.
     * @param _szUri MQTT broker URI.
     * @param _szLoggerName Logger name.
     * @param _szTag Client/logger tag.
     */
    CClient(const std::string& _szClientId,
            const std::string& _szUri, const std::string& _szLoggerName, const std::string& _szTag);
    /**
     * @brief Destructor.
     */
    virtual ~CClient();
    /**
     * @brief Start client with configuration previously set.
     * @param _pProgListener Optional listener for the connection status.
     * @see SetCleanSession, SetConnectionTimeoutInterval, SetInterval, SetUsername, SetPassword, SetSecureConfiguration
     */
    void Start();
    /**
     * @brief Stop client.
     */
    void Stop();
    /**
     * @brief RewriteConnectionListener
     * @param _pCL Connection listener
     */
    void SetConnectionListener(IConnectionStatusListener* _pCL);
    /**
     * @brief DeactivateResurector
     */
    void DeactivateResurector();
    /**
     * @brief SetCacheLimit Set limit reported on number of URI re-tries, before reseting to default URI.
     * @param _nLimit Number of allowed server URI errors (@see SetServerError()). Set to 0 to disable cache.
     */
    void SetDnsCacheLimit(int _nLimit);
    /**
     * @brief Subscribe Subscribe on multiple topics at once
     * @param _Subscriptions subscription topics and qos
     */
    void Subscribe(const SubscriptionCollection& _Subscriptions);
    /**
     * @brief Subscribe MQTT subscribe
     * @param _szTopic Topic name
     * @param _eQos QoS
     */
    void Subscribe(const std::string& _szTopic, Mqtt::CMessage::QosLevel _eQos);
    /**
     * @brief Unsubscribe MQTT unsubscribe
     * @param _szTopic Topic name
     */
    void Unsubscribe(const std::string& _szTopic);
    /**
     * @brief Publish a MQTT message.
     * @param topic MQTT topic.
     * @param _Msg MQTT message.
     */
    virtual void PublishMsg(const std::string& topic, const Mqtt::CMessage& _Msg);
    /**
     * @brief Activate last will message to be sent on connect and disconnect.
     * @param _szTag User tag to use for this operation.
     */
    void ActivateStatusMessage(const std::string& _szTag);
    /**
     * @brief Register message observer.
     * @param _pObserver Pointer to an observer.
     */
    virtual void RegisterObserver(Mqtt::CMsgDelegate::Ptr _pObserver);
    /**
     * @brief Set reconnection interval. It is used on reconnect and can be updated in run-time.
     * @param _nInterval Fixed period to run connection retries.
     */
    void SetReconnectionInterval(unsigned int _nInterval);
    /**
     * @brief Set reconnection interval. It is used on reconnect and can be updated in run-time.
     * @param _nMinimum Shortest period until next connection retry in seconds.
     * @param _nMaximum Longest period until next connection retry in seconds.
     */
    void SetReconnectionInterval(unsigned int _nMinimum, unsigned int _nMaximum);
    /**
     * @brief Setter for server URI. Must be called before Connect().
     * @param _szServerURI Server URI.
     */
    void SetServerURI(const std::string& _szServerURI);
    /**
     * @fn void SetUsername(const std::string& _szUsername);
     * @brief Setter for username. Must be called before Connect().
     * @param _szUsername Username to set.
     */
    void SetUsername(const std::string& _szUsername);
    /**
     * @fn void SetPassword(const std::string& _szPassword);
     * @brief Setter for password. Must be called before Connect().
     * @param _szPassword Password to set.
     */
    void SetPassword(const std::string& _szPassword);
    /**
     * @fn void SetCleanSession(bool _bCleanSession);
     * @brief Setter for clean session value. Must be called before Connect().
     * @param _bCleanSession Clean session on disconnect.
     */
    void SetCleanSession(bool _bCleanSession);
    /**
     * @brief IsCleanSessionSet Getter for clean session flag value
     * @return true/false
     */
    bool IsCleanSessionSet() const;
    /**
     * @fn void SetConnectionTimeoutInterval(int _nInterval);
     * @brief Setter for connection timeout interval. Must be called before Connect().
     * @param _nInterval Interval to wait connection to complete.
     */
    void SetConnectionTimeoutInterval(int _nInterval);
    void SetKeepAliveInterval(int _nInterval);
    /**
     * @fn void SetSecureConfiguration(const std::string& _szTrustStore, const std::string& _szKeyStore="", const std::string& _szPrivateKey="", const std::string& _szPrivateKeyPassword="");
     * @brief Set secure configuration. Connect must be called again, to apply this configuration.
     * @param _szTrustStore Secure connection trust store.
     * @param _szKeyStore Secure connection key store.
     * @param _szPrivateKey Secure connection private key.
     * @param _szPrivateKeyPassword Secure conenction pasword for a provate key.
     */
    void SetSecureConfiguration(const std::string& _szTrustStore,
                                const std::string& _szKeyStore = "", const std::string& _szPrivateKey = "",
                                const std::string& _szPrivateKeyPassword = "");
    /**
     * @brief Get MQTT client identifier.
     * @return Client identifier.
     */
    std::string GetClientId() const;
    /**
     * @brief Get currently used broker address.
     * @return Broker address.
     */
    std::string GetServerURI() const;
    /**
     * @brief Is client connected
     * @return true/false
     */
    bool IsConnected() const;
    /**
     * @brief IsSessionPressent
     * @return true/false if session is present on the broker
     */
    bool IsSessionPressent() const;
    /**
     * @brief Ignore if session is persisted on the broker.
     * @param _bIgnore true/false
     */
    void IgnoreSessionPresent(bool _bIgnore);
    /**
     * @brief SubscribeOnConnect Subscribe after connection is established.
     *        By default, if clean session is set, subscription will be skipped. Set to TRUE to force subscription.
     * @param _bForce Force subscription
     */
    void SetSubscribeOnConnect(bool _bSubscribe);
private:
    /**
     * @brief Connect client.
     */
    virtual void Connect();
    /**
     * @brief Disconnect client.
     */
    virtual void Disconnect();
    /**
     * @brief GenerateWillMessage Generate last will message with corresponding status
     * @param _bOnline Indication of the presence
     * @return Generated message
     */
    Mqtt::CMessage::Ptr GenerateStatusMessage(bool _bOnline) const;
    /**
     * @brief OnConnectionLost
     * @param _szCause
     */
    virtual void OnConnectionLost(const std::string& _szCause);
    /**
     * @brief Method called when a message arrives.
     * @param _szTopic Message topic.
     * @param _pMsg Message payload.
     */
    virtual void OnMessageArrived(const std::string& _szTopic,
                                  const Mqtt::CMessage::Ptr _pMsg);
    /**
     * @brief Method called when a message delivery failes.
     * @param _nCode Error code.
     * @param _szTopic Message topic.
     * @param _pMsg Message payload.
     */
    virtual void OnPublishFailed(int _nCode, const std::string& _szTopic,
                                 const Mqtt::CMessage::Ptr _pMsg);

    Mqtt::CBuffClient* m_pClient;

    typedef std::list< Mqtt::CMsgDelegate::Ptr > Observers;
    Observers m_Observers;
    Poco::Mutex m_ObserverMutex;

    std::string m_szWillTopic;

    Messenger::CResurector* m_pResurector;
    IConnectionStatusListener* m_pConnListener;

    bool m_bSubscribeOnConnect;

    CDnsCache m_DnsCache;
};

} // MessageBroker

#endif
