#include "Database.h"
#include "Globals.h"
#include "Poco/File.h"

#include "Poco/Data/RecordSet.h"

namespace MessageBroker {
namespace Database {

using namespace Poco::Data::Keywords;
using Poco::Data::Session;
using Poco::Data::Statement;
using Poco::Data::RecordSet;

CDatabase* CDatabase::m_Instance = NULL;

ITable::ITable(const std::string& _szName):
    m_pSession(NULL),
    m_szName(_szName) {
    setLoggerName("Database");
}

ITable::~ITable() {
}

std::string ITable::GetName() const {
    return m_szName;
}

const std::string CParamsTable::NAME = "Params";

CParamsTable::CParamsTable():
    ITable(NAME) {
}

void CParamsTable::CreateTable() {
    // Create Params table
    Statement statement(*m_pSession);
    statement << "CREATE TABLE " << NAME << " (Key VARCHAR(50) NOT NULL, Value VARCHAR(160) NOT NULL, PRIMARY KEY (Key))";
    statement.execute();
    statement.reset(*m_pSession);
}

void CParamsTable::InitTable() {
    // Init Params table
    Statement statement(*m_pSession);
    statement << "insert into " << NAME << " values(\"" << VERSION_KEY << "\", \"" << VERSION << "\");";
#ifdef USE_OMB_BRIDGE
    statement << "insert into " << NAME << " values(\"" << Bridge::KEY_ACCOUNT_ID << "\", \"" << Bridge::DEFAULT_ACCOUNT << "\");";
    statement << "insert into " << NAME << " values(\"" << Bridge::KEY_RESET_SESSION << "\", \"1\");";
#endif
    statement.execute();
    statement.reset(*m_pSession);
}

unsigned int CParamsTable::UpdateTable(unsigned int _nFromVersion) {
    unsigned int updatedTo = _nFromVersion + 1;

    if (_nFromVersion == 3) {
        // Session info is not used since version 4
        Statement statement(*m_pSession);
        statement << "insert or replace into " << NAME << " values(\"" << Bridge::KEY_ACCOUNT_ID << "\", \"" << Bridge::DEFAULT_ACCOUNT << "\");";
        statement << "insert or replace into " << NAME << " values(\"" << Bridge::KEY_RESET_SESSION << "\", \"1\");";
        statement.execute();
        statement.reset(*m_pSession);
    }

    return updatedTo;
}

std::string CParamsTable::GetParam(std::string _szKey, const std::string& _szDefValue) {
    Statement select(*m_pSession);
    std::string value = _szDefValue;
    select << "SELECT Value FROM " << CParamsTable::NAME << " WHERE Key=?",
           into(value),
           use(_szKey);
    select.execute();
    return value;
}

bool CParamsTable::GetParamsLike(std::string _szKey, std::map<std::string, std::string>& params, bool keepSubkey){
    Statement select(*m_pSession);
    std::string key, value;
    int count = 0;
    select << "SELECT * FROM " << CParamsTable::NAME << " WHERE Key LIKE '%" << _szKey << "%'",
            into(key),
            into(value),
            range(0, 1);

    while(!select.done()){
        try{
            count = select.execute();
        }catch(Poco::Exception& exc){
            Poco::Logger::root().fatal("Error while reading from database: " + exc.displayText());
            return false;
        }

        if(count > 0){
            keepSubkey ? params[key] = value : params[key.erase(0, _szKey.length())] = value;
        }
    }
    return count;
}

void CParamsTable::DeleteParamsLike(const std::string& _szSubKey){
    Statement deleteMe(*m_pSession);
    deleteMe << "DELETE FROM " << CParamsTable::NAME << " WHERE Key LIKE '%" << _szSubKey << "%'";
    deleteMe.execute();
}

bool CParamsTable::GetParamBool(std::string _szKey, bool _bDefValue) {
    std::string defVal = _bDefValue ? "1" : "0";
    std::string value = GetParam(_szKey, defVal);
    bool param = (value=="0" ? false : true);
    return param;
}

void CParamsTable::SetParam(std::string _szKey, std::string _szValue) {
    Statement insert(*m_pSession);
    insert << "INSERT OR REPLACE INTO " << CParamsTable::NAME << " VALUES(?, ?)",
           use(_szKey),
           use(_szValue);
    insert.execute();
}

void CParamsTable::SetParam(std::string _szKey, bool _bValue) {
    string val = _bValue ? "1" : "0";
    SetParam(_szKey, val);
}

bool CDatabase::m_bInitialized = false;

void CDatabase::InitializeInstance(const string& configPath) {
    if(CDatabase::m_Instance != NULL)
        throw Poco::InvalidAccessException("Somebody already initialized SharedValue factory");
    m_Instance = new CDatabase(configPath);
}

CDatabase::CDatabase(const string& _fileName): m_szDbFile(_fileName) {
    Poco::File dbFile(m_szDbFile);
    setLoggerName("Database");

    if ( !m_bInitialized ) {
        // register SQLite connector
        Poco::Data::SQLite::Connector::registerConnector();
        m_bInitialized = true;
    }

    if ( !dbFile.exists() ) {
        dbFile.createFile();
        information("Database file <" + m_szDbFile + "> created.");
    }

    m_pSession = new Poco::Data::Session("SQLite", m_szDbFile);
}

CDatabase::~CDatabase() {
    Tables::iterator itr;

    for ( itr=m_Tables.begin(); itr!=m_Tables.end(); ++itr ) {
        ITable* p = *itr;
        SAFE_DELETE(p);
    }

    SAFE_DELETE(m_pSession);
}

CDatabase& CDatabase::GetInstance() {
    if(m_Instance == NULL) {
        InitializeInstance(DB_FILE);
    }
    return *m_Instance;
}

void CDatabase::Recreate() {
    Poco::File dbFile(m_szDbFile);

    if (dbFile.exists()) {
        dbFile.remove();
    }

    dbFile.createFile();
    information("Database file '" + m_szDbFile + "' re-created.");
    SAFE_DELETE(m_pSession);
    information("Previous db session deleted.");
    m_pSession = new Poco::Data::Session("SQLite", m_szDbFile);
    information("New db session created.");
    Tables::iterator itr;

    for (itr = m_Tables.begin(); itr != m_Tables.end(); ++itr) {
        ITable* table = *itr;
        table->m_pSession = m_pSession;
    }

    information("Tables updated with new db session.");
}

void CDatabase::RegisterTables(const CDatabase::Tables& _Tables) {
    Tables::const_iterator itr;

    for ( itr=_Tables.begin(); itr!=_Tables.end(); ++itr ) {
        ITable* table = *itr;
        table->m_pSession = m_pSession;
        m_Tables.push_back(table);
    }
}

void CDatabase::InitDb() {
    Tables::const_iterator itr;

    for (itr=m_Tables.begin(); itr!=m_Tables.end(); ++itr) {
        ITable* table = *itr;

        if ( !ContainsTable(table->GetName()) ) {
            m_pSession->begin();

            try {
                // Create tables
                table->CreateTable();
                std::stringstream ss;
                ss << "Table <" << table->GetName() << "> is created.";
                information(ss.str());
                // Initialize content
                table->InitTable();
                ss.str("");
                ss << "Table <" << table->GetName() << "> is initialized.";
                information(ss.str());
                m_pSession->commit();
            } catch (const Poco::Exception& e) {
                std::stringstream ss;
                ss << "Init table failed, rolling back (m=" << e.message() <<").";
                error(ss.str());
                m_pSession->rollback();
                throw Poco::IllegalStateException("Unable to initialize database");
            }
        }
    }
}

bool CDatabase::ContainsTable(std::string _szTableName) {
    Statement select(*m_pSession);
    string szTableType = "table";
    int cnt = 0;
    select << "SELECT count(*) FROM sqlite_master WHERE type=? AND name=?",
           into(cnt),
           use(szTableType),
           use(_szTableName);
    select.execute();
    return cnt>0;
}

void CDatabase::CheckVersion() {
    CParamsTable* paramsTable = GetTable<CParamsTable>();
    std::string szVersion = paramsTable->GetParam(VERSION_KEY);
    unsigned int ver = Poco::NumberParser::parseUnsigned(szVersion);

    if ( ver<VERSION ) {
        std::stringstream ss;
        ss << "Upgrading DB v." << ver << " to v." << VERSION <<".";
        debug(ss.str());
        Tables::const_iterator itr;
        m_pSession->begin();

        try {
            for (itr=m_Tables.begin(); itr!=m_Tables.end(); ++itr) {
                ITable* table = *itr;
                unsigned int tmpVer = ver;

                // Upgrading table
                while ( tmpVer!=VERSION ) {
                    tmpVer = table->UpdateTable(tmpVer);
                    std::stringstream ss;
                    ss << "Table <" << table->GetName() << "> is upgraded to version: " << tmpVer;
                    debug(ss.str());
                }

                std::stringstream ss;
                ss << "Table <" << table->GetName() << "> is upgraded.";
                information(ss.str());
            }

            paramsTable->SetParam(VERSION_KEY, Poco::NumberFormatter::format(VERSION));
            ss.str("");
            ss << "Database v." << ver << " is upgraded to v." << VERSION <<".";
            information(ss.str());
            m_pSession->commit();
        } catch (const Poco::Exception& e) {
            ss.str("");
            ss << "Database upgrade failed, rolling back (m=" << e.message() <<").";
            critical(ss.str());
            m_pSession->rollback();
            throw Poco::IllegalStateException("Unable to upgrade database");
        }
    } else if ( ver>VERSION ) {
        std::stringstream ss;
        ss << "DOWNGRADE detected, DB version missmatch - current: v." << ver << ", expected v." << VERSION <<".";
        warning(ss.str());
    }
}

} // Database
} // MessageBroker
