/****************************************************************************
 *  Copyright (c) 2019 OBLO Living LLC
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.obloliving.com
 *  -----------------------------------------------------
 *
 * \project Oblo Message Broker - Broker application.
 * \file bridge/DnsCache.h
 * \brief
 * OBLO cache for resolved domain names.
 *
 * @Author Milan Tucic
 *
 * \notes
 *
 * \history
 * 23.1.2019 Initial version
 *****************************************************************************/
#include "DnsCache.h"
#include <Poco/Net/DNS.h>
#include <Poco/URI.h>
#include <Poco/Net/IPAddress.h>

namespace MessageBroker {

using namespace std;

CDnsCache::CDnsCache(const std::string& _szServerURI):
    m_szDomainNameURI(_szServerURI),
    m_nLimit(-1),
    m_nCount(0) {
}

void CDnsCache::SetCacheLimit(int _nLimit) {
    m_nLimit = _nLimit - 1;
}

std::string CDnsCache::Resolve() {
    string uri;

    if (m_nLimit > 0) {
        // Resolve is requested.
        if (m_szResolvedURI.empty()) {
            m_szResolvedURI = ResolveServerURI(m_szDomainNameURI);
        }

        uri = m_szResolvedURI;
    } else {
        // Resolve is disabled.
        uri = m_szDomainNameURI;
    }

    return uri;
}

void CDnsCache::SetServerVerified() {
    // Reset counter
    m_nCount = 0;
}

void CDnsCache::SetServerError() {
    if (m_nCount < m_nLimit) {
        ++m_nCount;
    } else {
        m_szResolvedURI = "";
        m_nCount = 0;
    }
}

std::string CDnsCache::ResolveServerURI(const std::string& _szServerURI) {
    Poco::URI uri(_szServerURI);
    string host = uri.getHost();

    if (IsDomainName(host)) {
        string address;
        // Resolve only if host is not IP address
        Poco::Net::HostEntry entry = Poco::Net::DNS::resolve(host);
        const Poco::Net::HostEntry::AddressList& addresses = entry.addresses();
        Poco::Net::HostEntry::AddressList::const_iterator itr;

        for (itr = addresses.begin(); itr != addresses.end(); ++itr) {
            address = itr->toString();

            if (itr->family() == Poco::Net::IPAddress::IPv4) {
                break; // Force first IPv4
            }
        }

        if (!address.empty()) {
            // Address found, switch host.
            uri.setHost(address);
        }
    }

    return uri.toString();
}

bool CDnsCache::IsDomainName(const std::string& _szHost) {
    bool isDomainName = true;

    try {
        Poco::Net::IPAddress::parse(_szHost);
        isDomainName = false;
    } catch (...) {}

    return isDomainName;
}

} // MessageBroker
