/****************************************************************************
 *  Copyright (c) 2011 RT-RK d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.rt-rk.com
 *  -----------------------------------------------------
 *
 * \project Oblo Message Broker - Broker application.
 * \file Logger.cpp
 * \brief
 * Implementation of the Logger
 *
 * @Author Ivan Lazarevic, Milan Tucic
 * \notes
 *
 * \history
 * 03.03.2015. Initial version.
 *****************************************************************************/
#include "Logger.h"
#include "Globals.h"

#include <stdio.h>

#include <Poco/ConsoleChannel.h>
#include <Poco/FileChannel.h>
#include <Poco/AsyncChannel.h>
#include <Poco/SplitterChannel.h>
#include <Poco/SyslogChannel.h>
#include <Poco/FormattingChannel.h>
#include <Poco/PatternFormatter.h>
#include <sstream>

using Poco::AsyncChannel;
using Poco::SplitterChannel;
using Poco::FileChannel;
using Poco::PatternFormatter;
using Poco::FormattingChannel;

namespace MessageBroker {
namespace Logger {

/// *** Shared values *** ///
const string CLogConfig::KEY_BASE = "log.";
const string CLogConfig::WITH_CONSOLE = "with_console";
const bool CLogConfig::WITH_CONSOLE_DEFAULT = true;
const string CLogConfig::WITH_FILE = "with_file";
const bool CLogConfig::WITH_FILE_DEFAULT = true;
const string CLogConfig::WITH_SYSLOG = "with_syslog";
const bool CLogConfig::WITH_SYSLOG_DEFAULT = false;
const string CLogConfig::APPNAME = "syslog.app_name";
const string CLogConfig::APPNAME_DEFAULT = "oblomb";
const string CLogConfig::FACILITY = "syslog.facility";
const string CLogConfig::FACILITY_DEFAULT = "LOG_LOCAL2";
const string CLogConfig::OPTIONS = "syslog.option";
const string CLogConfig::OPTIONS_DEFAULT = "LOG_PID";
const string CLogConfig::FILE = "file_name";
const string CLogConfig::FILE_DEFAULT = "oblomb.log";
const string CLogConfig::PATTERN = "pattern";
const string CLogConfig::PATTERN_DEFAULT = "%e.%m.%y %L%H:%M:%S.%i %q [%s]: %t";
const string CLogConfig::PURGE_CNT = "purge_cnt";
const string CLogConfig::PURGE_CNT_DEFAULT = "3";
const string CLogConfig::FILE_SIZE = "file_size";
const string CLogConfig::FILE_SIZE_DEFAULT = "100K";
const string CLogConfig::COMPRESS = "compress";
const bool CLogConfig::COMPRESS_DEFAULT = true;
const string CLogConfig::DEFAULT_LEVEL = "level.default";
const string CLogConfig::DEFAULT_LEVEL_DEFAULT = "I";
const string CLogConfig::MODULE_LEVEL = "level.";
const string CLogConfig::MODULE_LEVEL_DEFAULT = "";

bool CLogConfig::WithConsole() const {
    bool v = m_pConfigValues->getBool(m_szKeyBase + WITH_CONSOLE, WITH_CONSOLE_DEFAULT);
    return v;
}

bool CLogConfig::WithFile() const {
    bool v = m_pConfigValues->getBool(m_szKeyBase + WITH_FILE, WITH_FILE_DEFAULT);
    return v;
}

bool CLogConfig::WithSyslog() const {
    bool v = m_pConfigValues->getBool(m_szKeyBase + WITH_SYSLOG, WITH_SYSLOG_DEFAULT);
    return v;
}

string CLogConfig::GetAppName() const {
    string v = m_pConfigValues->getString(m_szKeyBase + APPNAME, APPNAME_DEFAULT);
    return v;
}

std::string CLogConfig::GetFacility() const {
    string v = m_pConfigValues->getString(m_szKeyBase + FACILITY, FACILITY_DEFAULT);
    return v;
}

std::string CLogConfig::GetOptions() const {
    string v = m_pConfigValues->getString(m_szKeyBase + OPTIONS, OPTIONS_DEFAULT);
    return v;
}

string CLogConfig::GetFileName() const {
    string v = m_pConfigValues->getString(m_szKeyBase + FILE, FILE_DEFAULT);
    return v;
}

std::string CLogConfig::GetPattern() const {
    string v = m_pConfigValues->getString(m_szKeyBase + PATTERN, PATTERN_DEFAULT);
    return v;
}

std::string CLogConfig::GetPurgeCnt() const {
    string v = m_pConfigValues->getString(m_szKeyBase + PURGE_CNT, PURGE_CNT_DEFAULT);
    return v;
}

std::string CLogConfig::GetFileSize() const {
    string v = m_pConfigValues->getString(m_szKeyBase + FILE_SIZE, FILE_SIZE_DEFAULT);
    return v;
}

bool CLogConfig::GetCompress() const {
    bool v = m_pConfigValues->getBool(m_szKeyBase + COMPRESS, COMPRESS_DEFAULT);
    return v;
}

std::string CLogConfig::GetDefaultLevel() const {
    string v = m_pConfigValues->getString(m_szKeyBase + DEFAULT_LEVEL, DEFAULT_LEVEL_DEFAULT);
    return v;
}

std::string CLogConfig::GetModuleLevel(const std::string& _szModuleName) const {
    string v = m_pConfigValues->getString(m_szKeyBase + MODULE_LEVEL + _szModuleName, MODULE_LEVEL_DEFAULT);
    return v;
}

void CLogger::setLoggerName(const std::string& _szModuleName) {
    m_Logger = &(Poco::Logger::get(_szModuleName));
    CLogConfig::Ptr config = SharedValues::CFactory::GetHelper<CLogConfig>();
    string level = config->GetModuleLevel(_szModuleName);
    if(level != CLogConfig::MODULE_LEVEL_DEFAULT) {
        m_Logger->setLevel(getPriority(level));
    }
}

void CLogger::Initialize() {
    CLogConfig::Ptr config = SharedValues::CFactory::GetHelper<CLogConfig>();
    string appName = config->GetAppName();
    string appFacility = config->GetFacility();
    string appOptions = config->GetOptions();
    string fileName = config->GetFileName();
    string fileSize = config->GetFileSize();
    string purgeCount = config->GetPurgeCnt();
    bool compress = config->GetCompress();
    AutoPtr<PatternFormatter> pattern(new PatternFormatter);
    pattern->setProperty("pattern", config->GetPattern());
    AutoPtr<SplitterChannel> splitterChannel(new SplitterChannel);

    if (config->WithConsole()) {
        AutoPtr<FormattingChannel> consoleChannel(new FormattingChannel(pattern, getConsoleChannel()));
        splitterChannel->addChannel(consoleChannel);
    }

    if (config->WithFile()) {
        AutoPtr<FormattingChannel> fileChannel(new FormattingChannel(pattern, getFileChannel(fileName, fileSize, purgeCount, compress)));
        splitterChannel->addChannel(fileChannel);
    }

    if (config->WithSyslog()){
        Poco::AutoPtr<Poco::FormattingChannel> syslogChannel(new Poco::FormattingChannel(pattern, getSyslogChannel(appName, appFacility, appOptions)));
        splitterChannel->addChannel(syslogChannel);
    }

    Poco::Logger::root().setChannel(splitterChannel);
    Poco::Logger::setLevel(Poco::Logger::ROOT, getPriority(config->GetDefaultLevel()));
}

void CLogger::InitializeDefaultLevels(){
    CLogConfig::Ptr config = SharedValues::CFactory::GetHelper<CLogConfig>();
    string defaultLevel = config->GetDefaultLevel();
    Poco::Logger::setLevel(Poco::Logger::ROOT, getPriority(defaultLevel));

    std::vector<string> loggers;
    string level;
    Poco::Logger::names(loggers);
    for(unsigned int i = 0; i < loggers.size(); ++i){
        level = config->GetModuleLevel(loggers[i]);
        if(level != CLogConfig::MODULE_LEVEL_DEFAULT && level != defaultLevel) {
            Poco::Logger::setLevel(loggers[i], getPriority(level));
        }
    }
}

void CLogger::InitializeLevels(const std::map<string, string>& levels){
    for(std::map<string, string>::const_iterator it = levels.begin(); it != levels.end(); ++it){
        Poco::Logger::setLevel(it->first, atoi(it->second.c_str()));
    }
}

AutoPtr<Channel> CLogger::getFileChannel(const string& _fileName, const string& _maxSize, const std::string& _purgeCount, bool _compress) {
    AutoPtr<FileChannel> fileChannel(new FileChannel);
    fileChannel->setProperty("path", _fileName);
    fileChannel->setProperty("rotation", _maxSize);
    fileChannel->setProperty("archive","number");
    fileChannel->setProperty("purgeCount", _purgeCount);
    string compress = _compress ? "true" : "false";
    fileChannel->setProperty("compress", compress);
    AutoPtr<AsyncChannel> fileAsync(new AsyncChannel(fileChannel));
    return fileAsync;
}

AutoPtr<Channel> CLogger::getConsoleChannel() {
    AutoPtr<Poco::ColorConsoleChannel> colorConsoleChannel(new Poco::ColorConsoleChannel);
    AutoPtr<AsyncChannel> consoleAsync(new AsyncChannel(colorConsoleChannel));
    return consoleAsync;
}

Poco::AutoPtr<Poco::Channel> CLogger::getSyslogChannel(const std::string& _appName, const std::string& _facility, const std::string& _options) {
    Poco::AutoPtr<Poco::SyslogChannel> syslogChannel(new Poco::SyslogChannel());
    syslogChannel->setProperty("name", _appName);
    syslogChannel->setProperty("facility", _facility);
    syslogChannel->setProperty("options", _options);
    Poco::AutoPtr<Poco::AsyncChannel> syslogAsync(new Poco::AsyncChannel(syslogChannel));
    return syslogAsync;
}

int CLogger::getPriority(const std::string &_priority) {
    int defaultPriority = Poco::Logger::root().getLevel();

    if (_priority.length()<1) {
        return defaultPriority;
    }

    int criteria = _priority[0];

    switch (criteria) {
        case 'I':
        case 'i':
            return Poco::Message::PRIO_INFORMATION;

        case 'N':
        case 'n':
            return Poco::Message::PRIO_NOTICE;

        case 'T':
        case 't':
            return Poco::Message::PRIO_TRACE;

        case 'C':
        case 'c':
            return Poco::Message::PRIO_CRITICAL;

        case 'D':
        case 'd':
            return Poco::Message::PRIO_DEBUG;

        case 'E':
        case 'e':
            return Poco::Message::PRIO_ERROR;

        case 'W':
        case 'w':
            return Poco::Message::PRIO_WARNING;

        case 'F':
        case 'f':
            return Poco::Message::PRIO_FATAL;
        case 'Q':
        case 'q':
            return 0;
            break;
        default:
            return defaultPriority;
    }
}

void CLogger::log(const string& _msg) {
    m_Logger ? m_Logger->log(_msg) : void();
}

void CLogger::critical(const string& _msg) {
    if(m_Logger)
        m_Logger->critical() ? m_Logger->critical(_msg) : void();
}

void CLogger::debug(const string& _msg) {
    if(m_Logger)
        m_Logger->debug() ? m_Logger->debug(_msg) : void();
}

void CLogger::error(const string& _msg) {
    if(m_Logger)
        m_Logger->error() ? m_Logger->error(_msg) : void();
}

void CLogger::fatal(const string& _msg) {
    if(m_Logger)
        m_Logger->fatal() ? m_Logger->fatal(_msg) : void();
}

void CLogger::information(const string& _msg) {
    if(m_Logger)
        m_Logger->information() ? m_Logger->information(_msg) : void();
}

void CLogger::notice(const string& _msg) {
    if(m_Logger)
        m_Logger->notice() ? m_Logger->notice(_msg) : void();
}

void CLogger::trace(const string& _msg) {
    if(m_Logger)
        m_Logger->trace() ? m_Logger->trace(_msg) : void();
}

void CLogger::warning(const string& _msg) {
    if(m_Logger)
        m_Logger->warning() ? m_Logger->warning(_msg) : void();
}

} // Logger
} // MessageBroker
