/****************************************************************************
 *  Copyright (c) 2011 RT-RK d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.rt-rk.com
 *  -----------------------------------------------------
 *
 * \project Oblo Message Broker - Broker application.
 * \file Logger.h
 * \brief
 * Implementation of the Logger
 *
 * @Author Ivan Lazarevic, Milan Tucic
 * \notes
 *
 * \history
 * 03.03.2015. Initial version.
 * 16.07.2016. Configuration replaced with the shared values.
*****************************************************************************/
#ifndef OMB_LOGGER_H
#define OMB_LOGGER_H

#include <Poco/Logger.h>
#include <Poco/AutoPtr.h>
#include <Poco/SharedPtr.h>

#include "SharedValues.h"

using std::string;
using Poco::AutoPtr;
using Poco::Channel;


namespace MessageBroker {
namespace Logger {

class CLogConfig: public SharedValues::CHelper {
public:
    typedef Poco::SharedPtr<CLogConfig> Ptr;

    static const string KEY_BASE;
    static const string WITH_CONSOLE;
    static const bool WITH_CONSOLE_DEFAULT;
    static const string WITH_FILE;
    static const bool WITH_FILE_DEFAULT;
    static const string WITH_SYSLOG;
    static const bool WITH_SYSLOG_DEFAULT;
    static const string APPNAME;
    static const string APPNAME_DEFAULT;
    static const string FACILITY;
    static const string FACILITY_DEFAULT;
    static const string OPTIONS;
    static const string OPTIONS_DEFAULT;
    static const string FILE;
    static const string FILE_DEFAULT;
    static const string PATTERN;
    static const string PATTERN_DEFAULT;
    static const string PURGE_CNT;
    static const string PURGE_CNT_DEFAULT;
    static const string FILE_SIZE;
    static const string FILE_SIZE_DEFAULT;
    static const string COMPRESS;
    static const bool COMPRESS_DEFAULT;
    static const string DEFAULT_LEVEL;
    static const string DEFAULT_LEVEL_DEFAULT;
    static const string MODULE_LEVEL;
    static const string MODULE_LEVEL_DEFAULT;

    CLogConfig(): CHelper(KEY_BASE) {}
    bool WithConsole() const;
    bool WithFile() const;
    bool WithSyslog() const;
    string GetAppName() const;
    string GetFacility() const;
    string GetOptions() const;
    string GetFileName() const;
    string GetPattern() const;
    string GetPurgeCnt() const;
    string GetFileSize() const;
    bool GetCompress() const;
    string GetDefaultLevel() const;
    string GetModuleLevel(const string& _szModuleName) const;
};

class CLogger {
public:
    /**
     * @fn  CLogger::CLogger();
     * @brief   Default constructor.
     */
    CLogger():m_Logger(NULL) {
    }

    ~CLogger() {
        m_Logger = NULL;
    }
    /**
     * @fn  void CLogger::Initialize()
     * @brief   Initializes logger.
     */
    static void Initialize();

    /**
     * @fn  void CLogger::InitializeDefaultLevels()
     * @brief   Initializes currently defined loggers with initial levels from config.
     */
    static void InitializeDefaultLevels();

    /**
     * @fn  void CLogger::InitializeLevels(const std::map<string, string>& levels)
     * @brief   Initializes log levels save in database on boot
     */
    static void InitializeLevels(const std::map<string, string>& levels);
    /**
     * @fn  void CLogger::setLoggerName(Poco::Logger* logger, string _szModuleName)
     * @brief   Sets logger with given name.
     * @param   Poco::Logger* logger Logger
     * @param   string _szModuleName   Name of the logger.
     */
    void setLoggerName(const string& _szModuleName);

protected:
    //Log messages
    /** Log message */
    void log(const string& _msg);
    /** Fatal message */
    void fatal(const std::string& _msg);
    /** Critical message */
    void critical(const std::string& _msg);
    /** Error message */
    void error(const std::string& _msg);
    /** Warning message */
    void warning(const std::string& _msg);
    /** Notice message */
    void notice(const std::string& _msg);
    /** Information message */
    void information(const std::string& _msg);
    /** Debug message */
    void debug(const std::string& _msg);
    /** Trace message */
    void trace(const std::string& _msg);
private:
    /**
     * @fn  AutoPtr<Channel> CLogger::getFileChannel(const string& _fileName, const string& _maxSize, string purgeCount)
     * @brief   Gets channel for logging to file.
     * @param   const string& _fileName Name of the log file.
     * @param   const string& _maxSize  Maximum size of the log file.
     * @param   const string& _purgeCount       Number of files to be saved.
     * @param   bool _compress       Compress archived files.
     *
     * @returns Channel for file logging.
     */
    static AutoPtr<Channel> getFileChannel(const string& _fileName, const string& _maxSize, const string& _purgeCount, bool _compress);
    /**
     * @fn  AutoPtr<Channel> CLogger::getConsoleChannel()
     * @brief   Gets channel for logging to file.
     * @returns Channel for console logging.
     */
    static AutoPtr<Channel> getConsoleChannel();
    /**
     * @fn  AutoPtr<Poco::Channel> CLogger::getSyslogChannel(const string& _appName, const string& _facility, const string& _options)
     * @brief   Gets channel for logging to syslog.
     * @param   const string& _appName Name of aplication in syslog.
     * @param   const string& _facility A facility code is used to specify the type of program that is logging the message.
     * @param   const string& _options Option what will be log in syslog.
     *
     * @returns Channel for syslog logging.
     */
    static Poco::AutoPtr<Poco::Channel> getSyslogChannel(const std::string& _appName, const std::string& _facility, const std::string& _options);
    /**
     * @fn  int CLogger::getpriority(const string _priority)
     * @brief   Gets priority for logger based on user input.
     * @param   const string _priority  Wanted priority.
     * @returns Channel for console logging.
     */
    static int getPriority(const string& _priority);

    /**Logger*/
    Poco::Logger* m_Logger;
};

} // Logger
} // MessageBroker
#endif // OMB_LOGGER_H
