#include "ObloMb.h"
#include "Globals.h"
#include "SharedValues.h"
#include "Poco/Net/Socket.h"

#include "core/omb-messenger/OmbMessenger.h"
#include "core/mosquitto/Mosquitto.h"
#ifdef USE_OMB_BRIDGE
    #include "feature/bridge/Bridge.h"
#endif
#ifdef USE_OMB_LINKER
    #include "feature/linker/Linker.h"
#endif

#include <messenger/MqttLogger.h>

namespace MessageBroker {

COmbApplication::COmbApplication():
    m_bInitialized(false),
    m_bHelpRequested(false),
    m_pPahoMqttLogger(NULL) {
}

COmbApplication::~COmbApplication() {
}

/// Handles help option.
void COmbApplication::handleHelp(const std::string& name, const std::string& value) {
    m_bHelpRequested = true;
    PrintLogHeader();
    HelpFormatter helpFormatter(options());
    helpFormatter.setCommand(commandName());
    helpFormatter.setUsage("OPTIONS");
    helpFormatter.format(std::cout);
    stopOptionsProcessing();
}

void COmbApplication::defineOptions(OptionSet& options) {
    Application::defineOptions(options);
    options.addOption(
        Option("help", "h",
               "display help information on command line arguments").required(
            false).repeatable(false).callback(
            OptionCallback<COmbApplication>(this, &COmbApplication::handleHelp)));
    options.addOption(
        Option("config-file", "f",
               "load configuration data from a file").required(false).repeatable(
            true).argument(" file").callback(
            OptionCallback<COmbApplication>(this, &COmbApplication::handleConfig)));
}

void COmbApplication::handleConfig(const std::string& name, const std::string& value) {
    SharedValues::CFactory::InitializeInstance(value);
}

void COmbApplication::initialize(Poco::Util::Application& app) {
    Poco::Mutex::ScopedLock lock(m_AppMtx);
    if(m_bHelpRequested) return;
    if ( !m_bInitialized ) {
        Logger::CLogger::Initialize();
        Poco::Net::initializeNetwork();
        setLoggerName("");
        PrintLogHeader();
        information("Initialization...");
        SharedValues::CApp::Ptr appConfig = SharedValues::CFactory::GetHelper<SharedValues::CApp>();
        std::vector<string> modules = appConfig->GetEnabledModules();
        std::vector<string>::const_iterator enblItr;
        for ( enblItr=modules.begin(); enblItr!=modules.end(); ++enblItr ) {
            string mdl = *enblItr;
            CAppModule* p;

            try {
                p = CreateModule(mdl);
                RegisterModule(mdl, p);
            } catch (const Poco::Exception& e) {
                error(e.displayText());
            }
        }

        Database::CDatabase::InitializeInstance(appConfig->GetDBFilePath());
        Database::CDatabase& db = Database::CDatabase::GetInstance();
        Modules::const_iterator itr;
        Database::CDatabase::Tables tables;
        {
            Poco::RWLock::ScopedReadLock lock(m_ModuleLock);

            for ( itr=m_Modules.begin(); itr!=m_Modules.end(); ++itr ) {
                const ModuleElement& elm = *itr;
                elm.second->GetDbTables(tables);
            }
        }
        tables.push_back(new Database::CParamsTable());
        db.RegisterTables(tables);
        db.InitDb();

        try {
            db.CheckVersion();
        } catch (const Poco::Exception& e) {
            error("DB update failed, starting recovery procedure.");
            db.Recreate();
            db.InitDb();
            information("DB reinitialized successfully.");
        }

        m_pPahoMqttLogger = new CPahoMqttLogger();
        Messenger::CMqttLogger::getInstance().SetListener(m_pPahoMqttLogger);
        Poco::Util::ServerApplication::initialize(app);
        m_bInitialized = true;
        information("Initialized.");
    } else {
        information("Already initialized.");
    }
}

void COmbApplication::uninitialize() {
    Poco::Mutex::ScopedLock lock(m_AppMtx);

    if ( m_bInitialized ) {
        // Terminate application
        try {
            StopModules();
            information("Modules stopped.");
        } catch (const Poco::Exception& e) {
            error("Stop modules error: " + e.displayText());
        }

        Messenger::CMqttLogger::getInstance().SetListener(NULL);
        SAFE_DELETE(m_pPahoMqttLogger);
        Poco::Net::uninitializeNetwork();
        Poco::Util::ServerApplication::uninitialize();
        m_bInitialized = false;
        information("Uninitialized.");
        information("Application terminated regularly.");
    } else {
        information("Already uninitialized.");
    }
}

void COmbApplication::reinitialize(Poco::Util::Application& app) {
    Poco::Util::ServerApplication::reinitialize(app);
}
std::string COmbApplication::MakePrettyText(std::string _szText, char _Symbol, unsigned int _nLineSize) {
    int cnt = (_nLineSize - _szText.length())/2;
    std::string symbols(cnt, _Symbol);
    std::string line = symbols + _szText + symbols;

    if ( line.length()%2 ) {
        line += _Symbol;
    }

    return line;
}
#define HEADING_LINE_SIZE 62
#define HEADING_PRETTY_SYMBOL ' '
void COmbApplication::PrintLogHeader() {
    m_StartTime = Poco::LocalDateTime();
    std::string platform = " Platform: " + std::string(PLATFORM) + " ";
    std::string customer = " Customer: " + std::string(CUSTOMER) + " ";
#ifdef OMB_VERSION
    std::string version = " Version: " + std::string(OMB_VERSION) + " ";
#endif
    std::stringstream ss;
    ss << std::endl
       << MakePrettyText("", '-', HEADING_LINE_SIZE) << std::endl
       << MakePrettyText(" OBLO Message Broker ", HEADING_PRETTY_SYMBOL, HEADING_LINE_SIZE) << std::endl
       << MakePrettyText(platform, HEADING_PRETTY_SYMBOL, HEADING_LINE_SIZE) << std::endl
       << MakePrettyText(customer, HEADING_PRETTY_SYMBOL, HEADING_LINE_SIZE) << std::endl
#ifdef OMB_VERSION
       << MakePrettyText(version, HEADING_PRETTY_SYMBOL, HEADING_LINE_SIZE) << std::endl
#endif
       << MakePrettyText("", '-', HEADING_LINE_SIZE);
    information(ss.str());
}
void COmbApplication::PrintRunTime() {
    Poco::LocalDateTime now;
    Poco::Timespan diff = now - m_StartTime;
    std::stringstream ss;
    ss << "Terminating after: "
       << diff.days() << "d "
       << diff.hours() << "h "
       << diff.minutes() << "min "
       << diff.seconds() << "sec...";
    information(ss.str());
}

void COmbApplication::InitDbLoggers(){
    Database::CParamsTable* paramsTable =
        Database::CDatabase::GetTable<Database::CParamsTable>();
    std::map<string, string> loggers;
    if(paramsTable->GetParamsLike("log.level.", loggers, false)){
        CLogger::InitializeLevels(loggers);
    }
}

CAppModule* COmbApplication::CreateModule(const std::string& _szName) {
    CAppModule* p = NULL;

    if ( _szName==Mosquitto::MODULE_ID ) {
        p = new Mosquitto::CMosquitto();
    } else if ( _szName==OmbMessenger::MODULE_ID ) {
        p = new OmbMessenger::COmbMessenger();
#ifdef USE_OMB_BRIDGE
    } else if ( _szName==Bridge::MODULE_ID ) {
        p = new Bridge::CBridge();
#endif
#ifdef USE_OMB_LINKER
    } else if ( _szName==Linker::MODULE_ID ) {
        p = new Linker::CLinker();
#endif
    } else {
        throw Poco::InvalidArgumentException("Unknown module name: " + _szName);
    }

    return p;
}

int COmbApplication::main(const std::vector<std::string>& args) {
    int code = Poco::Util::ServerApplication::EXIT_OK;
    if(m_bHelpRequested) return code;
    information("Starting...");

    try {
        // Start modules
        StartModules();
        information("Modules started.");
        InitDbLoggers();
        // Wait for termination
        waitForTerminationRequest();
        PrintRunTime();
    } catch (const Poco::Exception& e) {
        error("Application error: " + e.message());
        code = Poco::Util::ServerApplication::EXIT_SOFTWARE;
    }

    return code;
}
}

int main(int argc, char* argv[]) {
    MessageBroker::COmbApplication app;

    try {
        app.run(argc, argv);
    } catch (const Poco::Exception& e) {
        std::cout << "Uncaught error: " << e.message() << std::endl;
    }

    return 0;
}
