/****************************************************************************
 *  Copyright (c) 2018 RT-RK d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.rt-rk.com
 *  -----------------------------------------------------
 *
 * \project Oblo Message Broker - Broker application.
 * \file ClientManager.cpp
 * \brief
 * Client manager.
 *
 * @Author Milan Tucic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#include "ClientManager.h"
#include "ClientFactory.h"
#include <core/Globals.h>
#include <sstream>

namespace MessageBroker {
namespace Mosquitto {

using namespace std;

CClientManager::CClientManager(IClientInfoProvider& _rInfoProvider):
    m_rInfoProvider(_rInfoProvider) {
    setLoggerName(Mosquitto::MODULE_ID);
}

CClientManager::~CClientManager() {
}

void CClientManager::AddClient(CClient::Ptr _pCli) {
    Poco::RWLock::ScopedWriteLock lock(m_CliMngrMutex);
    CliMap& m = m_Clients[_pCli->GetTypeId()];
    m[_pCli->GetObjectID()] = _pCli;
    stringstream ss;
    ss << "Client added (id='" << _pCli->GetID() << "')";
    information(ss.str());
}

void CClientManager::RemoveClient(const string& _szClientId) {
    Poco::RWLock::ScopedWriteLock lock(m_CliMngrMutex);
    CClientID cli;
    CClientFactory::ParseClientID(_szClientId, cli);
    Clients::iterator itr = m_Clients.find(cli.GetObjectType());

    if (itr != m_Clients.end()) {
        CliMap& cm = itr->second;
        cm.erase(cli.GetObjectID());
        stringstream ss;
        ss << "Client removed (id='" << _szClientId << "')";
        information(ss.str());
    }
}

void CClientManager::SetStatus(const string& _szClientId, bool _bOnline) {
    CClient::Ptr cli = GetClient(_szClientId);

    if (cli.isNull()) {
        cli = CClientFactory::GetInstance().Create(_szClientId, m_rInfoProvider);
        AddClient(cli);
    }

    // TODO!!! when to delete a client???
    cli->SetOnline(_bOnline);
    stringstream ss;
    ss << "Status changed (id='" << cli->GetID() << "', s=" << cli->IsOnline() << ")";
    information(ss.str());
}

void CClientManager::Clear() {
    Poco::RWLock::ScopedWriteLock lock(m_CliMngrMutex);
    m_Clients.clear();
}

bool CClientManager::IsAuthorizedPublish(const string& _szClientId, const string& _szTopic) {
    CClient::Ptr cli = GetClient(_szClientId);

    if (cli.isNull()) {
        cli = CClientFactory::GetInstance().Create(_szClientId, m_rInfoProvider);
        AddClient(cli); // Add client
    }

    bool authorized = cli->IsAuthorizedPublish(_szTopic);
    return authorized;
}

bool CClientManager::IsAuthorizedSubscribe(const string& _szClientId, const string& _szTopic) {
    CClient::Ptr cli = GetClient(_szClientId);

    if (cli.isNull()) {
        cli = CClientFactory::GetInstance().Create(_szClientId, m_rInfoProvider);
        AddClient(cli); // Add client
    }

    bool authorized = cli->IsAuthorizedSubscribe(_szTopic);
    return authorized;
}


void CClientManager::GetClients(const string& _szType, vector<CClient::Ptr>& _Clients) {
    Poco::RWLock::ScopedReadLock lock(m_CliMngrMutex);
    Clients::const_iterator itr = m_Clients.find(_szType);

    if (itr != m_Clients.end()) {
        const CliMap& m = itr->second;
        CliMap::const_iterator citr;

        for (citr = m.begin(); citr != m.end(); ++citr) {
            _Clients.push_back(citr->second);
        }
    }
}

void CClientManager::GetOnlineClients(const string& _szType, std::vector<CClient::Ptr>& _Clients) {
    Poco::RWLock::ScopedReadLock lock(m_CliMngrMutex);
    Clients::const_iterator itr = m_Clients.find(_szType);

    if (itr != m_Clients.end()) {
        const CliMap& m = itr->second;
        CliMap::const_iterator citr;

        for (citr = m.begin(); citr != m.end(); ++citr) {
            CClient::Ptr cli = citr->second;

            if (cli->IsOnline()) {
                _Clients.push_back(cli);
            }
        }
    }
}

CClient::Ptr CClientManager::GetClient(const string& _szClientId) {
    Poco::RWLock::ScopedReadLock lock(m_CliMngrMutex);
    CClientID id;
    CClientFactory::ParseClientID(_szClientId, id);
    CClient::Ptr client = NULL;
    const CliMap& m = m_Clients[id.GetObjectType()]; // Allocates new, if doesn't exist
    CliMap::const_iterator citr = m.find(id.GetObjectID());

    if (citr != m.end()) {
        client = citr->second;
    }

    return client;
}
}
}
