/****************************************************************************
 *  Copyright (c) 2018 RT-RK d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.rt-rk.com
 *  -----------------------------------------------------
 *
 * \project Oblo Message Broker - Broker application.
 * \file Clients.cpp
 * \brief
 * Client implementations.
 *
 * @Author Milan Tucic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#include "Clients.h"
#include <Poco/String.h>
#include <Poco/Format.h>
#include <Poco/Timestamp.h>
#include <Poco/Logger.h>
#include <sstream>

namespace MessageBroker {
namespace Mosquitto {

using namespace std;

const string CClientGtwModule::TYPE_ID = "app/";
const string CClientClient::TYPE_ID = "cli/";
const string CClientLocalAmc::TYPE_ID = "amc/";
const string CClientBridge::TYPE_ID = "stratus";
const string CClientUser::TYPE_ID = "user";
const string CClientWiseHub::TYPE_ID = "hub/";
const string CClientWiseIPD::TYPE_ID = "ipd/";

string& CClientID::GetObjectType() {
    return m_szObjectType;
}

string& CClientID::GetObjectID() {
    return m_szObjectId;
}

CClient::CClient(CClientID& _rClientID, IClientInfoProvider& _rInfoProvider):
    m_rInfoProvider(_rInfoProvider) {
    m_szObjectID = _rClientID.GetObjectID();
    m_bOnline = false;
}

CClient::~CClient() {
}

string CClient::GetID() const {
    return GetTypeId() + GetObjectID();
}

string CClient::GetObjectID() const {
    return m_szObjectID;
}

bool CClient::IsOnline() const {
    return m_bOnline;
}

void CClient::SetOnline(bool _bOnline) {
    m_bOnline = _bOnline;
}

bool CClient::Validate(const string& _szTopic, const char* _szRegex) const {
    bool valid = false;
    string reg = _szRegex; // copy to a string
    Poco::replaceInPlace(reg, "/", "\\/"); // regex slash
    Poco::replaceInPlace(reg, "<ANY>", "[^\\/]+");
    Poco::replaceInPlace(reg, "<GTW_ID>", m_rInfoProvider.GetGatewayID().c_str());
    Poco::replaceInPlace(reg, "<ID>", m_szObjectID.c_str());
    // Add wildcards for begining and end of string
    // Add optional project ID and account ID prefix = "projectID/accountID/regex"
    stringstream ss;
    ss << "^(" << m_rInfoProvider.GetProjectID() << "\\/" << m_rInfoProvider.GetAccountID() << "\\/)?" << reg << "$";
    reg = ss.str();
    Poco::RegularExpression regex(reg);
    valid = regex.match(_szTopic);
    return valid;
}

bool CClientGtwModule::IsAuthorizedPublish(const string& _szTopic) const {
    bool authorized =
        Validate(_szTopic, "gtw/<GTW_ID>/<ID>/(evt|lwm)") // Event and status topics
        || Validate(_szTopic, "gtw/<GTW_ID>/<ID>/<ANY>(/<ANY>)?/evt") // Event topic for functionality level 1 + optional level 2
        || Validate(_szTopic, "gtw/<GTW_ID>/(?!<ID>)<ANY>/(req|rsp)") // Request and response to other gateway modules
        || Validate(_szTopic, "gtw/<GTW_ID>/(?!<ID>)<ANY>/<ANY>/req") // Request and response to other gateway modules
        || Validate(_szTopic, "gtw/(?!<GTW_ID>)<ANY>/<ANY>/(req|rsp)")// Request and response to other gateways
        || Validate(_szTopic, "(svc|cli|amc)/<ANY>/(req|rsp)")// Request and response to cloud services, clients and app manager clients
        || Validate(_szTopic, "svc/<ANY>/gtw/<GTW_ID>/(req|rsp)")// Request and response to cloud services; depricated, by spec functionalities should not have req
        || Validate(_szTopic, "user/(req|rsp)");// Request and response to user (temporary registration client)
    return authorized;
}

bool CClientGtwModule::IsAuthorizedSubscribe(const string& _szTopic) const {
    bool authorized =
        Validate(_szTopic, "gtw/<GTW_ID>/<ID>/(req|rsp)") // Request and response topics
        || Validate(_szTopic, "gtw/<GTW_ID>/<ID>/<ANY>(/<ANY>)?/req") // Request topic for functionality level 1 + optional level 2
        || Validate(_szTopic, "gtw/<GTW_ID>/(?!<ID>)<ANY>/(evt|lwm)") // Event and status of other gateway modules
        || Validate(_szTopic, "gtw/(?!<GTW_ID>)<ANY>/<ANY>/(evt|lwm)")// Event and status of other gateways
        || Validate(_szTopic, "(svc|cli|amc)/<ANY>/(evt|lwm)")// Event and status of cloud services, clients and app manager clients
        || Validate(_szTopic, "svc/<ANY>/<ANY>/evt")// Event of cloud services, functionality level 1
        || Validate(_szTopic, "svc/<ANY>/gtw/<GTW_ID>/evt")// Event of cloud services
        || Validate(_szTopic, "user/(evt|lwm)");// Event and status of user (temporary registration client)
    return authorized;
}

bool CClientClient::IsAuthorizedPublish(const string& _szTopic) const {
    bool authorized =
        Validate(_szTopic, "cli/<ID>/(evt|lwm)") // Event and status topics
        || Validate(_szTopic, "gtw/<ANY>/<ANY>/(req|rsp)") // Request and response to gateway modules
        || Validate(_szTopic, "svc/<ANY>/(req|rsp)") // Request and response to cloud services
        || Validate(_szTopic, "svc/<ANY>/gtw/<ANY>/(req|rsp)"); // Request and response to cloud services, gtw functionality
    return authorized;
}

bool CClientClient::IsAuthorizedSubscribe(const string& _szTopic) const {
    bool authorized =
        Validate(_szTopic, "cli/<ID>/(req|rsp)") // Request and response topics
        || Validate(_szTopic, "gtw/<ANY>/<ANY>/(evt|lwm)") // Event and status of any gateway, any module
        || Validate(_szTopic, "gtw/<ANY>/ohm/azc/evt") // Event and status of any gateway, any module
        || Validate(_szTopic, "svc/<ANY>/(evt|lwm)") // Event and status to cloud services
        || Validate(_szTopic, "svc/<ANY>/gtw/<ANY>/evt"); // Event of cloud services
    return authorized;
}

bool CClientLocalAmc::IsAuthorizedPublish(const string& _szTopic) const {
    bool authorized =
        Validate(_szTopic, "amc/<ID>/(evt|lwm)") // Event and status topics
        || Validate(_szTopic, "gtw/<ANY>/<ANY>/(req|rsp)"); // Request and response to gateway modules
    return authorized;
}

bool CClientLocalAmc::IsAuthorizedSubscribe(const string& _szTopic) const {
    bool authorized =
        Validate(_szTopic, "cli/<ID>/(req|rsp)") // Request and response topics
        || Validate(_szTopic, "gtw/<ANY>/<ANY>/(evt|lwm)"); // Event and status of gateway modules
    return authorized;
}

bool CClientBridge::IsAuthorizedPublish(const string& _szTopic) const {
    bool authorized =
        Validate(_szTopic, "gtw/<ANY>/<ANY>/(req|rsp)") // Request and response to gateway modules
        || Validate(_szTopic, "gtw/<ANY>/<ANY>/<ANY>/req") // Request to gateway modules functionality level 1
        || Validate(_szTopic, "gtw/(?!<GTW_ID>)<ANY>/<ANY>/(evt|lwm)")// Event and status of remote gateways
        || Validate(_szTopic, "gtw/(?!<GTW_ID>)<ANY>/<ANY>/<ANY>/evt")// Event of remote gateways, functionality level 1
        || Validate(_szTopic, "svc/<ANY>/(evt|lwm)")// Event and status of cloud services
        || Validate(_szTopic, "svc/<ANY>/<ANY>/evt")// Event of cloud services, functionality level 1
        || Validate(_szTopic, "svc/<ANY>/gtw/<ANY>(/<ANY>)?/evt");// Event and status of cloud services, functionality level 1 + optional level 2
    return authorized;
}

bool CClientBridge::IsAuthorizedSubscribe(const string& _szTopic) const {
    bool authorized =
        Validate(_szTopic, "gtw/<ANY>/<ANY>/(evt|lwm)") // Event and status of gateway modules
        || Validate(_szTopic, "gtw/<ANY>/<ANY>/<ANY>/evt") // Event of gateway modules functionality level 1
        || Validate(_szTopic, "gtw/(?!<GTW_ID>)<ANY>/<ANY>/(req|rsp)")// Request and response to remote gateways
        || Validate(_szTopic, "gtw/(?!<GTW_ID>)<ANY>/<ANY>/<ANY>/req")// Request to remote gateways, functionality level 1
        || Validate(_szTopic, "svc/<ANY>/(req|rsp)")// Request and response to cloud services
        || Validate(_szTopic, "svc/<ANY>/gtw/<ANY>(/<ANY>)?/(req|rsp)");// Request and response to cloud services, functionality level 1 + optional level 2; depricated, by spec rsp should not be used by a functionality
    return authorized;
}

bool CClientUser::IsAuthorizedPublish(const string& _szTopic) const {
    bool authorized =
        Validate(_szTopic, "user/(evt|lwm)") // Event and status topics
        || Validate(_szTopic, "gtw/<GTW_ID>/omb/auth/req") // Request to register
        || Validate(_szTopic, "gtw/<GTW_ID>/omb/rsp"); // Response on reqistration
    return authorized;
}

bool CClientUser::IsAuthorizedSubscribe(const string& _szTopic) const {
    bool authorized =
        Validate(_szTopic, "user/(req|rsp)"); // Request and response topics
    return authorized;
}

bool CClientWiseHub::IsAuthorizedPublish(const string& _szTopic) const {
    bool authorized =
        Validate(_szTopic, "hub/<GTW_ID>/(evt|sts)") // Event and status topics
        || Validate(_szTopic, "hub/<GTW_ID>/<ANY>/<ANY>/evt") // Event functionality topics
        || Validate(_szTopic, "ipd/<ANY>/(req|rsp)") // Request and response to any IPD
        || Validate(_szTopic, "ipd/<ANY>/<ANY>/<ANY>/req"); // Request and response to any IPD's functionality
    return authorized;
}

bool CClientWiseHub::IsAuthorizedSubscribe(const string& _szTopic) const {
    bool authorized =
        Validate(_szTopic, "hub/<GTW_ID>/(req|rsp)") // Request and response topics
        || Validate(_szTopic, "hub/<GTW_ID>/<ANY>/<ANY>/req") // Request functionality topics
        || Validate(_szTopic, "ipd/<ANY>/(evt|sts)") // Event and status of any IPD
        || Validate(_szTopic, "ipd/<ANY>/<ANY>/<ANY>/evt"); // Event of any IPD's functionality
    return authorized;
}

bool CClientWiseIPD::IsAuthorizedPublish(const string& _szTopic) const {
    bool authorized =
        Validate(_szTopic, "ipd/<ID>/(evt|sts)") // Event and status topics
        || Validate(_szTopic, "ipd/<ID>/<ANY>/<ANY>/evt") // Event functionality topics
        || Validate(_szTopic, "hub/<GTW_ID>/(req|rsp)") // Request and response to HUB
        || Validate(_szTopic, "hub/<GTW_ID>/<ANY>/<ANY>/req"); // Request and response to HUB's functionality
    return authorized;
}

bool CClientWiseIPD::IsAuthorizedSubscribe(const string& _szTopic) const {
    bool authorized =
        Validate(_szTopic, "ipd/<ID>/(req|rsp)") // Request and response topics
        || Validate(_szTopic, "ipd/<ID>/<ANY>/<ANY>/req") // Request functionality topics
        || Validate(_szTopic, "hub/<GTW_ID>/(evt|sts)") // Event and status of HUB
        || Validate(_szTopic, "hub/<GTW_ID>/<ANY>/<ANY>/evt"); // Event of HUB's functionality
    return authorized;
}

}
}
