/****************************************************************************
 *  Copyright (c) 2018 RT-RK d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.rt-rk.com
 *  -----------------------------------------------------
 *
 * \project Oblo Message Broker - Broker application.
 * \file Clients.h
 * \brief
 * Client implementations.
 *
 * @Author Milan Tucic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef OMB_MOSQUITTO_CLIENTS_H
#define OMB_MOSQUITTO_CLIENTS_H

#include <Poco/SharedPtr.h>
#include <Poco/RegularExpression.h>
#include <list>

namespace MessageBroker {
namespace Mosquitto {

typedef Poco::SharedPtr<Poco::RegularExpression> RegexPtr;
typedef std::list<RegexPtr> AccessControlList;

class IClientInfoProvider {
public:
    virtual const std::string& GetProjectID() = 0;
    virtual const std::string& GetAccountID() = 0;
    virtual const std::string& GetGatewayID() = 0;
};

class CClientID {
public:
    std::string& GetObjectType();
    std::string& GetObjectID();

private:
    std::string m_szObjectType;
    std::string m_szObjectId;
};

class CClient {
public:
    typedef Poco::SharedPtr<CClient> Ptr;

    CClient(CClientID& _rClientID, IClientInfoProvider& _rInfoProvider);
    virtual ~CClient();

    std::string GetID() const;
    std::string GetObjectID() const;
    virtual std::string GetTypeId() const = 0;
    bool IsOnline() const;
    void SetOnline(bool _bOnline);

    virtual bool IsAuthorizedPublish(const std::string& _szTopic) const = 0;
    virtual bool IsAuthorizedSubscribe(const std::string& _szTopic) const = 0;

protected:
    bool Validate(const std::string& _szTopic, const char* _szRegex) const;

private:
    IClientInfoProvider& m_rInfoProvider;
    std::string m_szObjectID;
    bool m_bOnline;
};

class CDummyClient: public CClient {
public:
    CDummyClient(CClientID& _rClientID, IClientInfoProvider& _rInfoProvider): CClient(_rClientID, _rInfoProvider) {}
    std::string GetTypeId() const {
        return "";
    }
    bool IsAuthorizedPublish(const std::string& _szTopic) const {
        return false;
    }
    bool IsAuthorizedSubscribe(const std::string& _szTopic) const {
        return false;
    }
};

class CClientGtwModule: public CClient {
public:
    static const std::string TYPE_ID;
    CClientGtwModule(CClientID& _rClientID, IClientInfoProvider& _rInfoProvider): CClient(_rClientID, _rInfoProvider) {}
    std::string GetTypeId() const {
        return TYPE_ID;
    }
    bool IsAuthorizedPublish(const std::string& _szTopic) const;
    bool IsAuthorizedSubscribe(const std::string& _szTopic) const;
};

class CClientClient: public CClient {
public:
    static const std::string TYPE_ID;
    CClientClient(CClientID& _rClientID, IClientInfoProvider& _rInfoProvider): CClient(_rClientID, _rInfoProvider) {}
    std::string GetTypeId() const {
        return TYPE_ID;
    }
    bool IsAuthorizedPublish(const std::string& _szTopic) const;
    bool IsAuthorizedSubscribe(const std::string& _szTopic) const;
};

class CClientLocalAmc: public CClient {
public:
    static const std::string TYPE_ID;
    CClientLocalAmc(CClientID& _rClientID, IClientInfoProvider& _rInfoProvider): CClient(_rClientID, _rInfoProvider) {}
    std::string GetTypeId() const {
        return TYPE_ID;
    }
    bool IsAuthorizedPublish(const std::string& _szTopic) const;
    bool IsAuthorizedSubscribe(const std::string& _szTopic) const;
};

class CClientBridge: public CClient {
public:
    static const std::string TYPE_ID;
    CClientBridge(CClientID& _rClientID, IClientInfoProvider& _rInfoProvider): CClient(_rClientID, _rInfoProvider) {}
    std::string GetTypeId() const {
        return TYPE_ID;
    }
    bool IsAuthorizedPublish(const std::string& _szTopic) const;
    bool IsAuthorizedSubscribe(const std::string& _szTopic) const;
};

class CClientUser: public CClient {
public:
    static const std::string TYPE_ID;
    CClientUser(CClientID& _rClientID, IClientInfoProvider& _rInfoProvider): CClient(_rClientID, _rInfoProvider) {}
    std::string GetTypeId() const {
        return TYPE_ID;
    }
    bool IsAuthorizedPublish(const std::string& _szTopic) const;
    bool IsAuthorizedSubscribe(const std::string& _szTopic) const;
};

class CClientWiseHub: public CClient {
public:
    static const std::string TYPE_ID;
    CClientWiseHub(CClientID& _rClientID, IClientInfoProvider& _rInfoProvider): CClient(_rClientID, _rInfoProvider) {}
    std::string GetTypeId() const {
        return TYPE_ID;
    }
    bool IsAuthorizedPublish(const std::string& _szTopic) const;
    bool IsAuthorizedSubscribe(const std::string& _szTopic) const;
};

class CClientWiseIPD: public CClient {
public:
    static const std::string TYPE_ID;
    CClientWiseIPD(CClientID& _rClientID, IClientInfoProvider& _rInfoProvider): CClient(_rClientID, _rInfoProvider) {}
    std::string GetTypeId() const {
        return TYPE_ID;
    }
    bool IsAuthorizedPublish(const std::string& _szTopic) const;
    bool IsAuthorizedSubscribe(const std::string& _szTopic) const;
};

}
}

#endif
