/****************************************************************************
 *  Copyright (c) 2018 RT-RK d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.rt-rk.com
 *  -----------------------------------------------------
 *
 * \project Oblo Message Broker - Broker application.
 * \file Mosquitto.cpp
 * \brief
 * Mosquitto wrapper.
 *
 * @Author Milan Tucic
 *
 * \notes
 *
 * \history
 * created 2016
 *****************************************************************************/
#include "core/mosquitto/Mosquitto.h"
#include "core/mosquitto/Tables.h"
#include "core/Globals.h"
#include "ClientFactory.h"
#include "Poco/JSON/Parser.h"

#include "stdlib.h"
#include <sstream>

namespace MessageBroker {
namespace Mosquitto {

using namespace std;

IMosquitto* CRunnable::m_pContext = NULL;

CRunnable::CRunnable(IMosquitto* _pMosquitto) {
    m_bRunning = true;
    m_pBinder = (broker_binder_t*) malloc(sizeof(broker_binder_t));

    if (m_pBinder) {
        memset(m_pBinder, 0, sizeof(broker_binder_t));
        m_pContext = _pMosquitto;
        m_pBinder->on_log_print = _OnLogPrint;
        m_pBinder->on_acl_check = _OnAccessControlCheck;
        m_pBinder->on_unpwd_check = _OnCredentialCheck;
        m_pBinder->on_client_connected = _OnClientConnected;
        m_pBinder->on_client_disconnected = _OnClientDisonnected;
        m_pBinder->on_client_do_disconnect_check = _OnClientDoDisonnectCheck;
        m_pBinder->get_pem_password = _GetPemPassword;
        broker_set_binder(m_pBinder);
    }
}

CRunnable::~CRunnable() {
    if (m_pBinder) {
        free(m_pBinder);
    }
}

void CRunnable::run() {
    while (m_bRunning) {
        char* name = strdup("mosquitto");
        char* configFlag = strdup("-c");
        CConfig::Ptr conf = SharedValues::CFactory::GetHelper<CConfig>();
        string szConfigFile = conf->GetConfigFile();
        char* configFile = strdup(szConfigFile.c_str());
        char* params[] = { name, configFlag, configFile };
        broker_main(3, params);
        free(name);
        free(configFlag);
        free(configFile);

        if (m_bRunning) {
            _OnLogPrint("Terminated unexpectedly, restart in 10s", BROKER_LOG_WARNING);
            Poco::Thread::sleep(10000);
        }
    }

    _OnLogPrint("Thread terminated", BROKER_LOG_INFORMATION);
}

void CRunnable::stop() {
    if (m_bRunning) {
        m_bRunning = false;
        broker_set_running(0);
    }
}

void CRunnable::_OnLogPrint(const char* _szLog, broker_log_level_t _Level) {
    m_pContext->OnLog(_szLog, _Level);
}

int CRunnable::_OnAccessControlCheck(const char* _szClientId, const char* _szUsername, const char* _szTopic, broker_permissions_t _nAccess) {
    string clientid = _szClientId ? _szClientId : "";
    string username = _szUsername ? _szUsername : "";
    string topic = _szTopic ? _szTopic : "";
    bool ret = m_pContext->OnAccessControlCheck(clientid,
               username, topic, _nAccess);
    return ret ? BROKER_RET_SUCCESS : BROKER_RET_ACL_DENIED;
}

int CRunnable::_OnCredentialCheck(const char* _szUsername, const char* _szPassword, int _nPort) {
    string username = _szUsername ? _szUsername : "";
    string password = _szPassword ? _szPassword : "";
    return m_pContext->OnCredentialCheck(username, password, _nPort) ? BROKER_RET_SUCCESS : BROKER_RET_AUTH_FAILED;
}

void CRunnable::_OnClientConnected(const char* _szClientId, const char* _szUsername) {
    string clientid = _szClientId ? _szClientId : "";
    string username = _szUsername ? _szUsername : "";
    m_pContext->OnClientConnected(clientid, username);
}

void CRunnable::_OnClientDisonnected(const char* _szClientId, const char* _szUsername) {
    string clientid = _szClientId ? _szClientId : "";
    string username = _szUsername ? _szUsername : "";
    m_pContext->OnClientDisconnected(clientid, username);
}

int CRunnable::_OnClientDoDisonnectCheck(const char* _szClientId) {
    string clientid = _szClientId ? _szClientId : "";
    bool ret = m_pContext->OnClientDoDisconnectCheck(clientid);
    return ret ? 1 : 0;
}

const char* CRunnable::_GetPemPassword(unsigned int _nPort) {
    return m_pContext->GetPemPassword(_nPort);
}

const string CMD_USER_ONLINE = "cmd_user_online";
const string CMD_USER_OFFLINE = "cmd_user_offline";
const string PRM_USERNAME = "username";
const string PRM_TIME = "time";

CMosquitto::CMosquitto():
    m_pMosq(NULL),
    m_pExpirationTimer(NULL),
    m_pCredsTable(NULL),
    m_pTmpUserTable(NULL),
    m_pUserDetails(NULL),
    m_ClientManager(*this) {
    setLoggerName(Mosquitto::MODULE_ID);
    AddCommandExecutor(CMD_ADD_CREDENTIALS, MAKE_CMD_EXEC_DELEGATE(CMosquitto, ExecuteAddCredentials));
    AddCommandExecutor(CMD_REMOVE_USER, MAKE_CMD_EXEC_DELEGATE(CMosquitto, ExecuteUserRemovement));
}

CMosquitto::~CMosquitto() {
    Stop();
}

void CMosquitto::Start() {
    Poco::Mutex::ScopedLock lock(m_ModuleMutex);

    if (!IsRunning()) {
        m_pCredsTable = Database::CDatabase::GetTable<CCredsTable>();
        m_pTmpUserTable = Database::CDatabase::GetTable<CTmpUsersTable>();
        m_pUserDetails = Database::CDatabase::GetTable<CUserDetails>();
        trace("Starting...");
        m_szGtwId = SharedValues::CFactory::GetHelper<SharedValues::CGateway>()->GetId();
        CConfig::Ptr conf = SharedValues::CFactory::GetHelper<CConfig>();
        m_UnlockedListeners = conf->GetUnlockedListeners();
        m_pExpirationTimer = new Poco::Timer(USER_EXP_CHECK_PERIOD, USER_EXP_CHECK_PERIOD);
        m_pExpirationTimer->start(Poco::TimerCallback<CMosquitto>(*this, &CMosquitto::OnExpirationCheck));
        m_pMosq = new CRunnable(this);
        m_MosqThread.start(*m_pMosq);
        Poco::Thread::sleep(1000);
        CAppModule::Start();
        information("Started.");
    }
}

void CMosquitto::Stop() {
    Poco::Mutex::ScopedLock lock(m_ModuleMutex);

    if (IsRunning()) {
        trace("Stoping...");

        if (m_pMosq) {
            m_pMosq->stop();
            m_MosqThread.join();
            SAFE_DELETE(m_pMosq);
        }

        if (m_pExpirationTimer) {
            m_pExpirationTimer->stop();
            SAFE_DELETE(m_pExpirationTimer);
        }

        m_ClientManager.Clear();
        CAppModule::Stop();
        information("Stoped.");
    }
}

const string& CMosquitto::GetProjectID() {
    return PROJECT_ID;
}

const string& CMosquitto::GetAccountID() {
    return ACCOUNT_ID;
}

const string& CMosquitto::GetGatewayID() {
    return m_szGtwId;
}

void CMosquitto::GetDbTables(MessageBroker::Database::CDatabase::Tables& _Tables) const {
    _Tables.push_back(new CCredsTable());
    _Tables.push_back(new CAccessControlTable());
    _Tables.push_back(new CTmpUsersTable());
    _Tables.push_back(new CUserDetails());
}

void CMosquitto::OnLog(const string& _szLog, broker_log_level_t _Level) {
    switch (_Level) {
        case BROKER_LOG_TRACE:
            trace(_szLog);
            break;

        case BROKER_LOG_DEBUG:
            debug(_szLog);
            break;

        case BROKER_LOG_NOTICE:
            debug(_szLog);
            break;

        case BROKER_LOG_INFORMATION:
            information(_szLog);
            break;

        case BROKER_LOG_WARNING:
            warning(_szLog);
            break;

        case BROKER_LOG_ERROR:
            error(_szLog);
            break;

        default:
            error(_szLog);
            break;
    }
}

bool CMosquitto::OnAccessControlCheck(const string& _szClientId, const string& _szUsername, const string& _szTopic, broker_permissions_t _nAccess) {
    bool valid = false;

    // Skip
    if (!_szClientId.empty()) {
        switch (_nAccess) {
            case BROKER_PERM_SUBSCRIBE:
                valid = true;
                break;

            case BROKER_PERM_READ: {
                valid = IsAuthorizedSubscribe(_szClientId, _szTopic);
                break;
            }

            case BROKER_PERM_WRITE: {
                valid = IsAuthorizedPublish(_szClientId, _szTopic);
                break;
            }

            default:
                break;
        }

        stringstream ss;
        ss << "User <" << _szClientId << "> is ";
        ss << (valid ? "" : "NOT ");
        ss << "authorized to ";
        ss << (_nAccess != BROKER_PERM_WRITE ? "subscribe to " : "publish to ");
        ss << _szTopic;
        debug(ss.str());
    } else {
        trace("Empty client ID on access control check.");
    }

    return valid;
}

bool CMosquitto::OnCredentialCheck(const string& _szUsername, const string& _szPassword, int _nPort) {
    // Don't check password for local clients.
    if (find(m_UnlockedListeners.begin(), m_UnlockedListeners.end(), _nPort) != m_UnlockedListeners.end()) {
        return true;
    }

    bool valid;
    valid = m_pCredsTable->CheckCredentials(_szUsername, _szPassword);
    stringstream ss;
    ss << "User <" << _szUsername << "> is ";
    ss << (valid ? "authorized" : "NOT authorized");
    ss << " -> password <" << _szPassword << ">.";
    valid ? debug(ss.str()) : warning(ss.str());
    return valid;
}

void CMosquitto::OnClientConnected(const string& _szClientId, const string& _szUsername) {
    if (!_szClientId.empty()) {
        m_ClientManager.SetStatus(_szClientId, true);
        DismissUserDisconnect(_szClientId);
    }
}

void CMosquitto::OnClientDisconnected(const string& _szClientId, const string& _szUsername) {
    if (!_szClientId.empty()) {
        m_ClientManager.SetStatus(_szClientId, false);
    }
}

bool CMosquitto::OnClientDoDisconnectCheck(const string& _szClientId) {
    Poco::FastMutex::ScopedLock lock(m_DoDisconnectMutex);
    bool disconnect = false;
    list<string>::iterator itr;

    for (itr = m_DoDisconnectClients.begin(); itr != m_DoDisconnectClients.end(); ++itr) {
        if (*itr == _szClientId) {
            disconnect = true;
            m_DoDisconnectClients.erase(itr);
            stringstream ss;
            ss << "User disconnect cleared (u='" << _szClientId << "')";
            information(ss.str());
            break;
        }
    }

    return disconnect;
}

const char* CMosquitto::GetPemPassword(unsigned int _nPort) {
    const char* password = NULL;
#ifdef OMB_PRIVATE_KEY_PASSWORD
    password = OMB_PRIVATE_KEY_PASSWORD;
#endif
    return password;
}

void CMosquitto::OnExpirationCheck(Poco::Timer& timer) {
    list<string> users;
    long time = GetCurrentTimeInHours();
    m_pTmpUserTable->GetExpiredUsers(users, time);
    list<string>::const_iterator itr;

    for (itr = users.begin(); itr != users.end(); ++itr) {
        string username = *itr;
        RemoveUser(username);
    }
}

long CMosquitto::GetCurrentTimeInSeconds() {
    Poco::Timestamp timestamp;
    return timestamp.epochTime();
}

long CMosquitto::GetCurrentTimeInHours() {
    return ConvertSecToHours(GetCurrentTimeInSeconds());
}

long CMosquitto::ConvertSecToHours(long _nSeconds) {
    return _nSeconds / 3600;
}

long CMosquitto::ConvertDaysToHours(long _nDays) {
    return _nDays * 24; // time in hours.
}

Poco::DynamicStruct::Ptr CMosquitto::ExecuteAddCredentials(const string& _szCmdName, const Poco::DynamicStruct& _Params) {
    Poco::DynamicStruct::Ptr ret = new Poco::DynamicStruct();
    string username = _Params[PARAM_USERNAME].convert<string>();
    string password = _Params[PARAM_PASSWORD].convert<string>();
    string name = _Params[PARAM_NAME].convert<string>();
    string appversion = _Params[PARAM_APPVERSION].convert<string>();
    string ostype = _Params[PARAM_OSTYPE].convert<string>();
    int expiration = 0;

    if (_Params.contains(PARAM_EXPIRATION)) {
        expiration = _Params[PARAM_EXPIRATION];
    }

    m_pCredsTable->StoreCredentials(username, password);
    m_pUserDetails->UpdateDetails(username, name, appversion, ostype);

    if (expiration > 0) {
        long time = 0;
        time = GetCurrentTimeInHours() + ConvertDaysToHours(expiration);
        m_pTmpUserTable->StoreTempUser(username, expiration, time);
    }

    information("Credentials added (u='" + username + "')");
    return ret;
}

Poco::DynamicStruct::Ptr CMosquitto::ExecuteUserRemovement(const string& _szCmdName, const Poco::DynamicStruct& _Params) {
    Poco::DynamicStruct::Ptr ret = new Poco::DynamicStruct();
    string username = _Params[PARAM_USERNAME].convert<string>();
    RemoveUser(username);
    DisconnectUser(username);
    return ret;
}

void CMosquitto::RemoveUser(const string& _szUsername) {
    m_pTmpUserTable->DeleteTempUser(_szUsername);
    m_pCredsTable->DeleteCredentials(_szUsername);
    information("Credentials removed (u='" + _szUsername + "')");
}

void CMosquitto::DisconnectUser(const string& _szUsername) {
    Poco::FastMutex::ScopedLock lock(m_DoDisconnectMutex);
    m_DoDisconnectClients.push_back(_szUsername);
    stringstream ss;
    ss << "User scheduled for disconnect (u='" << _szUsername << "', n=" << m_DoDisconnectClients.size() << ")";
    information(ss.str());
}

void CMosquitto::DismissUserDisconnect(const string& _szUsername) {
    Poco::FastMutex::ScopedLock lock(m_DoDisconnectMutex);
    list<string>::iterator itr;

    for (itr = m_DoDisconnectClients.begin(); itr != m_DoDisconnectClients.end(); ++itr) {
        if (*itr == _szUsername) {
            m_DoDisconnectClients.erase(itr);
            stringstream ss;
            ss << "User disconnect cleared (u='" << _szUsername << "')";
            information(ss.str());
            break;
        }
    }
}

bool CMosquitto::IsAuthorizedSubscribe(const string& _szClientId, const string& _szTopic) {
    bool valid = false;

    try {
        valid = m_ClientManager.IsAuthorizedSubscribe(_szClientId, _szTopic);
    } catch (const Poco::Exception& e) {
        string msg;
        Poco::format(msg, "Failed to validate (t='%s', e='%s')", _szTopic, e.displayText());
        error(msg);
    }

    /* TODO find better solution for this!!! */

    if (!valid && _szClientId == BRIDGE_USER && _szTopic.compare(0, 4, "cli/") == 0) {
        /* Stratus user is allowed to use topics for mobile clients but it is granted dynamically,
        depending of whether client user is connected or not to the OMB. */
        stringstream regex;
        string clients = BuildCliRegexBlock();
        regex << "^cli\\/" << clients << "\\/(evt|lwm)$";
        Poco::RegularExpression reg(regex.str());

        if (reg.match(_szTopic)) {
            valid = true;
        } else {
            regex.str("");
            regex << "^cli\\/(?!" << clients << ")[^\\/]+\\/(req|rsp)$";
            Poco::RegularExpression reg(regex.str());
            valid = reg.match(_szTopic);
        }
    } else if (!valid) {
        warning("User '" + _szClientId + "' is NOT authorized to sub on: " + _szTopic);
    }

    return valid;
}

bool CMosquitto::IsAuthorizedPublish(const string& _szClientId, const string& _szTopic) {
    bool valid = false;

    try {
        valid = m_ClientManager.IsAuthorizedPublish(_szClientId, _szTopic);
    } catch (const Poco::Exception& e) {
        string msg;
        Poco::format(msg, "Failed to validate (t='%s', e='%s')", _szTopic, e.displayText());
        error(msg);
    }

    /* TODO find better solution for this!!! */

    if (!valid && _szClientId == BRIDGE_USER && _szTopic.compare(0, 4, "cli/") == 0) {
        /* Stratus user is allowed to use topics for mobile clients but it is granted dynamically,
        depending of whether client user is connected to the OMB or not. */
        /* TODO find better solution for this */
        stringstream regex;
        string clients = BuildCliRegexBlock();
        regex << "^cli\\/" << clients << "\\/(req|rsp)$";
        Poco::RegularExpression reg(regex.str());

        if (reg.match(_szTopic)) {
            valid = true;
        } else {
            regex.str("");
            regex << "^cli\\/(?!" << clients << ")[^\\/]+\\/(evt|lwm)$";
            Poco::RegularExpression reg(regex.str());
            valid = reg.match(_szTopic);
        }
    } else if (!valid) {
        warning("User '" + _szClientId + "' is NOT authorized to pub on: " + _szTopic);
    }

    return valid;
}

string CMosquitto::BuildCliRegexBlock() {
    stringstream regex;
    vector<CClientClient::Ptr> clients;
    m_ClientManager.GetOnlineClients(CClientClient::TYPE_ID, clients);

    if (!clients.empty()) {
        regex << "(";
        vector<CClientClient::Ptr>::const_iterator begin = clients.begin();
        vector<CClientClient::Ptr>::const_iterator itr;

        for (itr = clients.begin(); itr != clients.end(); ++itr) {
            CClientClient::Ptr cli = * itr;
            string objId = cli->GetObjectID();

            if (itr != begin) {
                regex << "|";
            }

            regex << objId;
        }

        regex << ")";
    } else {
        regex << "(?!.*)" ;
    }

    return regex.str();
}

const string CConfig::KEY_BASE = "mosquitto.";

CConfig::CConfig(): SharedValues::CHelper(KEY_BASE) {
}

string CConfig::GetConfigFile() const {
    return m_pConfigValues->getString(m_szKeyBase + "config_file", "mosquitto.conf");
}

vector<int> CConfig::GetUnlockedListeners() const {
    string s = m_pConfigValues->getString(m_szKeyBase + "unlocked_listeners", "[]");
    Poco::JSON::Parser parser;
    Poco::Dynamic::Var content = parser.parse(s);
    Poco::JSON::Array::Ptr jsonArr = content.extract<Poco::JSON::Array::Ptr>();
    Poco::JSON::Array::ConstIterator itr;
    vector<int> listeners;

    for (itr = jsonArr->begin(); itr != jsonArr->end(); ++itr) {
        const Poco::Dynamic::Var& l = *itr;
        listeners.push_back(l.convert<int>());
    }

    return listeners;
}

}
}
