/****************************************************************************
 *  Copyright (c) 2016 RT-RK d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.rt-rk.com
 *  -----------------------------------------------------
 *
 * \project Oblo Message Broker - Broker application.
 * \file Mosquitto.h
 * \brief
 * Mosquitto wrapper.
 *
 * @Author Milan Tucic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef OMB_MOSQUITTO_H
#define OMB_MOSQUITTO_H

#include "core/AppModule.h"
#include "core/Logger.h"
#include "messenger/Runners.h"
#include "core/mosquitto/Tables.h"
#include "ClientManager.h"

#include "Poco/Runnable.h"
#include "Poco/Thread.h"
#include "Poco/SynchronizedObject.h"
#include "Poco/Timer.h"
#include <vector>

extern "C" {
#include "oblomb/binder.h"
}

namespace MessageBroker {
namespace Mosquitto {

class CConfig: public SharedValues::CHelper {
public:
    typedef Poco::SharedPtr<CConfig> Ptr;
    static const std::string KEY_BASE;

    CConfig();
    std::string GetConfigFile() const;
    std::vector<int> GetUnlockedListeners() const;
};

class IMosquitto {
public:
    virtual void OnLog(const std::string& _szLog, broker_log_level_t _Level) = 0;
    virtual bool OnAccessControlCheck(const std::string& _szClientId, const std::string& _szUsername,
                                      const std::string& _szTopic, broker_permissions_t _nAccess) = 0;
    virtual bool OnCredentialCheck(const std::string& _szUsername, const std::string& _szPassword, int _nPort) = 0;
    virtual void OnClientConnected(const std::string& _szClientId, const std::string& _szUsername) = 0;
    virtual void OnClientDisconnected(const std::string& _szClientId, const std::string& _szUsername) = 0;
    virtual bool OnClientDoDisconnectCheck(const std::string& _szClientId) = 0;
    virtual const char* GetPemPassword(unsigned int _nPort) = 0;
};

class CRunnable: public Poco::Runnable {
public:
    CRunnable(IMosquitto* _pMosquitto);
    ~CRunnable();
    virtual void run();
    void stop();
private:
    static void _OnLogPrint(const char* _szLog, broker_log_level_t _Level);
    static int _OnAccessControlCheck(const char* _szClientId, const char* _szUsername,
                                     const char* _szTopic, broker_permissions_t _nAccess);
    static int _OnCredentialCheck(const char* _szUsername, const char* _szPassword, int _nPort);
    static void _OnClientConnected(const char* _szClientId, const char* _szUsername);
    static void _OnClientDisonnected(const char* _szClientId, const char* _szUsername);
    static int _OnClientDoDisonnectCheck(const char* _szClientId);
    static const char* _GetPemPassword(unsigned int _nPort);

    bool m_bRunning;
    broker_binder_t* m_pBinder;
    static IMosquitto* m_pContext;
};

class CMosquitto: public CAppModule, public IMosquitto, public IClientInfoProvider {
public:
    CMosquitto();
    ~CMosquitto();
    void Start();
    void Stop();

    const std::string& GetProjectID();
    const std::string& GetAccountID();
    const std::string& GetGatewayID();

protected:
    virtual void GetDbTables(Database::CDatabase::Tables& _Tables) const;
private:
    virtual void OnLog(const std::string& _szLog, broker_log_level_t _Level);
    virtual bool OnAccessControlCheck(const std::string& _szClientId, const std::string& _szUsername, const std::string& _szTopic, broker_permissions_t _nAccess);
    virtual bool OnCredentialCheck(const std::string& _szUsername, const std::string& _szPassword, int _nPort);
    virtual void OnClientConnected(const std::string& _szClientId, const std::string& _szUsername);
    virtual void OnClientDisconnected(const std::string& _szClientId, const std::string& _szUsername);
    virtual bool OnClientDoDisconnectCheck(const std::string& _szClientId);
    virtual const char* GetPemPassword(unsigned int _nPort);

    void OnExpirationCheck(Poco::Timer& timer);

    static long GetCurrentTimeInSeconds();
    static long GetCurrentTimeInHours();
    static long ConvertSecToHours(long _nSeconds);
    static long ConvertDaysToHours(long _nDays);

    Poco::DynamicStruct::Ptr ExecuteAddCredentials(const std::string& _szCmdName,
            const Poco::DynamicStruct& _Params);
    Poco::DynamicStruct::Ptr ExecuteUserRemovement(const std::string& _szCmdName,
            const Poco::DynamicStruct& _Params);

    void RemoveUser(const std::string& _szUsername);
    void DisconnectUser(const std::string& _szUsername);
    void DismissUserDisconnect(const std::string& _szUsername);

    bool IsAuthorizedSubscribe(const std::string& _szClientId, const std::string& _szTopic);
    bool IsAuthorizedPublish(const std::string& _szClientId, const std::string& _szTopic);
    std::string BuildCliRegexBlock();

    CRunnable* m_pMosq;
    Poco::Thread m_MosqThread;

    Poco::Timer* m_pExpirationTimer;
    Poco::Thread m_ExpirationThread;
    std::vector<int> m_UnlockedListeners;

    CCredsTable* m_pCredsTable;
    CTmpUsersTable* m_pTmpUserTable;
    CUserDetails* m_pUserDetails;

    CClientManager m_ClientManager;
    std::string m_szGtwId;

    std::list<std::string> m_DoDisconnectClients;
    Poco::FastMutex m_DoDisconnectMutex;
};

}
}

#endif
