#include "core/mosquitto/Tables.h"
#include "core/Globals.h"

namespace MessageBroker {
namespace Mosquitto {

using namespace std;
using namespace Poco::Data::Keywords;
using Poco::Data::Session;
using Poco::Data::Statement;
using Poco::Data::RecordSet;

const std::string CCredsTable::NAME = "Credentials";

CCredsTable::CCredsTable():
    ITable(NAME) {
}

void CCredsTable::CreateTable() {
    // Create Credentials table
    Statement statement(*m_pSession);
    statement << "CREATE TABLE " << NAME << " (Username VARCHAR(50) NOT NULL, Password VARCHAR(50) NOT NULL, Name VARCHAR(50) NOT NULL, Appversion VARCHAR(50) NOT NULL, Type VARCHAR(50) NOT NULL,  PRIMARY KEY (Username))";
    statement.execute();
    statement.reset(*m_pSession);
}

void CCredsTable::InitTable() {
    // Init Credentials table
    Statement statement(*m_pSession);
    statement << "INSERT INTO " << NAME << " Values(\"user\", \"user\", \"reg\", \"2.0\", \"os\");";
    statement << "INSERT INTO " << NAME << " Values(\"client\", \"tmpclient\", \"wise\", \"1.0\", \"os\");";
    statement.execute();
    statement.reset(*m_pSession);
}

unsigned int CCredsTable::UpdateTable(unsigned int _nFromVersion) {
    unsigned int updatedTo = _nFromVersion + 1;

    if (_nFromVersion == 1) {
        Statement statement(*m_pSession);
        statement << "ALTER TABLE " << NAME << " ADD Name VARCHAR(50) NOT NULL DEFAULT \"name\";";
        statement << "ALTER TABLE " << NAME << " ADD Appversion VARCHAR(50) NOT NULL DEFAULT \"2.0\";";
        statement << "ALTER TABLE " << NAME << " ADD Type VARCHAR(50) NOT NULL DEFAULT \"os\";";
        statement.execute();
        statement.reset(*m_pSession);
    } else if (_nFromVersion == 2) {
        Statement statement(*m_pSession);
        statement << "INSERT INTO " << NAME << " Values(\"client\", \"tmpclient\", \"wise\", \"1.0\", \"os\");";
        statement.execute();
        statement.reset(*m_pSession);
    } else if (_nFromVersion == 4) {
        // Drop table, old one had expired users and additional columns
        Statement statement(*m_pSession);
        statement << "DROP TABLE " << NAME << ";";
        statement.execute();
        statement.reset(*m_pSession);
        // Create and re-initialize
        CreateTable();
        InitTable();
    }

    return updatedTo;
}

void CCredsTable::StoreCredentials(std::string _szUsername, std::string _szPassword) {
    // Insert statement
    Statement insert(*m_pSession);
    // Empty strings because of the backward compatibility
    insert << "INSERT OR REPLACE INTO " << CCredsTable::NAME << " VALUES(?, ?, \"\", \"\", \"\")",
           use(_szUsername),
           use(_szPassword);
    insert.execute();
}

void CCredsTable::DeleteCredentials(std::string _szUsername) {
    Statement del(*m_pSession);
    del << "DELETE FROM " << CCredsTable::NAME << " WHERE Username=?",
        use(_szUsername);
    del.execute();
}

bool CCredsTable::CheckCredentials(std::string _szUsername, std::string _szPassword) {
    Statement select(*m_pSession);
    Poco::Nullable<std::string> storedPassword;
    select << "SELECT Password FROM " << CCredsTable::NAME << " WHERE Username=?",
           into(storedPassword),
           use(_szUsername);
    select.execute();
    bool bValid = false;

    if (!storedPassword.isNull()) {
        bValid = _szPassword == storedPassword;
    }

    return bValid;
}

const std::string CAccessControlTable::NAME = "AccessControlList";

CAccessControlTable::CAccessControlTable():
    ITable(NAME) {
}

void CAccessControlTable::CreateTable() {
    // Create AccessControlList table
    Statement statement(*m_pSession);
    statement << "CREATE TABLE " << NAME << " (Owner VARCHAR(50) NOT NULL, Regex VARCHAR(200) NOT NULL, Perm TINYINT NOT NULL)";
    statement.execute();
    statement.reset(*m_pSession);
}

void CAccessControlTable::InitTable() {
    // Init AccessControlList table
}

unsigned int CAccessControlTable::UpdateTable(unsigned int _nFromVersion) {
    return Database::VERSION;
}

const std::string CTmpUsersTable::NAME = "TempUsers";

CTmpUsersTable::CTmpUsersTable():
    ITable(NAME) {
}

void CTmpUsersTable::CreateTable() {
    // Create TempUsers table
    Statement statement(*m_pSession);
    statement << "CREATE TABLE " << NAME << " (Username VARCHAR(50) NOT NULL, Expiration TINYINT NOT NULL, ValidUntil INT8 NOT NULL, PRIMARY KEY (Username))";
    statement.execute();
    statement.reset(*m_pSession);
}

void CTmpUsersTable::InitTable() {
    // Init TempUsers table
}

unsigned int CTmpUsersTable::UpdateTable(unsigned int _nFromVersion) {
    return Database::VERSION;
}

void CTmpUsersTable::StoreTempUser(std::string _szUsername, int _nExpiration, long _nValidUntil) {
    Statement insert(*m_pSession);
    insert << "INSERT OR REPLACE INTO " << CTmpUsersTable::NAME << " VALUES(?, ?, ?)",
           use(_szUsername),
           use(_nExpiration),
           use(_nValidUntil);
    insert.execute();
}

bool CTmpUsersTable::ContainsTempUser(std::string _szUsername) {
    Statement select(*m_pSession);
    int cnt = 0;
    select << "SELECT count(*) FROM " << CTmpUsersTable::NAME << " WHERE Username=?",
           into(cnt),
           use(_szUsername);
    select.execute();
    return cnt > 0;
}

void CTmpUsersTable::DeleteTempUser(std::string _szUsername) {
    Statement del(*m_pSession);
    del << "DELETE FROM " << CTmpUsersTable::NAME << " WHERE Username=?",
        use(_szUsername);
    del.execute();
}

long CTmpUsersTable::GetExpirationTime(std::string _szUsername) {
    Statement select(*m_pSession);
    Poco::Nullable<long> expiration;
    select << "SELECT Expiration FROM " << CTmpUsersTable::NAME << " WHERE Username=?",
           into(expiration),
           use(_szUsername);
    select.execute();
    return expiration.isNull() ? 0 : expiration.value();
}

void CTmpUsersTable::GetExpiredUsers(std::list<std::string>& _Users, long _nTime) {
    Statement select(*m_pSession);
    select << "SELECT Username FROM " << CTmpUsersTable::NAME << " WHERE ValidUntil<?",
           into(_Users),
           use(_nTime);
    select.execute();
}

const std::string CUserDetails::NAME = "UserDetails";

CUserDetails::CUserDetails(): ITable(NAME) {
}

void CUserDetails::CreateTable() {
    // Create Credentials table
    Statement statement(*m_pSession);
    statement << "CREATE TABLE " << NAME <<
              " (Username VARCHAR(50) NOT NULL, "
              "Name VARCHAR(50) NOT NULL, "
              "Version VARCHAR(50) NOT NULL, "
              "Type VARCHAR(50) NOT NULL, "
              "FirstReg VARCHAR(30) NOT NULL, "
              "LastReg VARCHAR(30) NOT NULL, "
              "PRIMARY KEY (Username));";
    statement.execute();
    statement.reset(*m_pSession);
}

void CUserDetails::InitTable() {
}

unsigned int CUserDetails::UpdateTable(unsigned int _nFromVersion) {
    return Database::VERSION;
}

void CUserDetails::UpdateDetails(string _szUsername, string _szName, string _szVersion, string _szType) {
    m_pSession->begin();

    try {
        Statement select(*m_pSession);
        int cnt = 0;
        select << "SELECT count(*) FROM " << NAME << " WHERE Username=?",
               into(cnt),
               use(_szUsername);
        select.execute();
        string date = Poco::DateTimeFormatter::DateTimeFormatter::format(Poco::LocalDateTime(), Poco::DateTimeFormat::SORTABLE_FORMAT);

        if (cnt > 0) {
            // User already in the DB, update columns
            Statement update(*m_pSession);
            update << "UPDATE " << NAME
                   << " SET Name=?, Version=?, LastReg=?"
                   << " WHERE Username=?;",
                   use(_szName),
                   use(_szVersion),
                   use(date),
                   use(_szUsername);
            update.execute();
        } else {
            // User is new
            Statement insert(*m_pSession);
            insert << "INSERT INTO " << NAME << " VALUES(?, ?, ?, ?, ?, ?)",
                   use(_szUsername),
                   use(_szName),
                   use(_szVersion),
                   use(_szType),
                   use(date),
                   use(date);
            insert.execute();
        }

        m_pSession->commit();
    } catch (const Poco::Exception& e) {
        m_pSession->rollback();
        warning("Failed to update user details: " + e.displayText());
    }
}

} // Mosquitto
} // MessageBroker

