#include "core/omb-messenger/OmbMessenger.h"
#include "core/Globals.h"

#include "Poco/Crypto/CipherFactory.h"
#include "Poco/Crypto/CipherKey.h"
#include "Poco/SHA1Engine.h"

#include <sstream>

namespace MessageBroker {
namespace OmbMessenger {

const std::string   IAssignmentHandler::GET_ASSIGNMENT      = "getUserAssignment";
const std::string   IAssignmentHandler::ACCOUNT_ID          = "account_id";
const int           IAssignmentHandler::ASSIGN_TIMEOUT      = 10;
const int           IAssignmentHandler::MAX_RETRANSMISSIONS = 3;

CRequestHandling::CRequestHandling(COmbMessenger* _pObj): m_pObj(_pObj) {
}

void CRequestHandling::AddRequestHandler(const std::string& _szType, const std::string& _szCommandName, const CRequestHandling::RespondFunctionPtr _pFunc) {
    Poco::Mutex::ScopedLock lock(m_Mutex);
    AddTypeRequestHandler(m_RequestHandlers, _szType, _szCommandName, _pFunc);
}

void CRequestHandling::CallHandler(const std::string& _szType, const std::string& _szCommandName, const Messenger::CRequestMessage::Ptr _pReqMsg, Messenger::CResponseMessage::Ptr _pRsp) {
    Poco::Mutex::ScopedLock lock(m_Mutex);
    CallTypeHandler(m_RequestHandlers, _szType, _szCommandName, _pReqMsg, _pRsp);
}

void CRequestHandling::AddFunctRequestHandler(const std::string& _szFunctionality, const std::string& _szType, const std::string& _szCommandName, const CRequestHandling::RespondFunctionPtr _pFunc) {
    Poco::Mutex::ScopedLock lock(m_Mutex);
    TypeRequestHandlers& functHandlers = m_FunctReqHandlers[_szFunctionality];
    AddTypeRequestHandler(functHandlers, _szType, _szCommandName, _pFunc);
}

void CRequestHandling::CallFunctHandler(const std::string& _szFunctionality, const std::string& _szType,
                                        const std::string& _szCommandName, const Messenger::CRequestMessage::Ptr _pReqMsg,
                                        Messenger::CResponseMessage::Ptr _pRsp) {
    Poco::Mutex::ScopedLock lock(m_Mutex);
    FunctTypeReqHandlers::const_iterator functItr;
    functItr = m_FunctReqHandlers.find(_szFunctionality);

    if (functItr != m_FunctReqHandlers.end()) {
        TypeRequestHandlers& typeHandlers = m_FunctReqHandlers[_szFunctionality];
        CallTypeHandler(typeHandlers, _szType, _szCommandName, _pReqMsg, _pRsp);
    } else {
        throw Poco::NotFoundException("Unknown functionality", -1);
    }
}

void CRequestHandling::AddTypeRequestHandler(CRequestHandling::TypeRequestHandlers& _TypeHandlers,
        const std::string& _szType, const std::string& _szCommandName,
        const CRequestHandling::RespondFunctionPtr _pFunc) {
    RequestHandler handler;
    handler.first = _szCommandName;
    handler.second = _pFunc;
    RequestHandlers& typeHandler = _TypeHandlers[_szType];
    typeHandler.insert(handler);
}

void CRequestHandling::CallTypeHandler(CRequestHandling::TypeRequestHandlers& _TypeHandlers, const std::string& _szType,
                                       const std::string& _szCommandName, const Messenger::CRequestMessage::Ptr _pReqMsg,
                                       Messenger::CResponseMessage::Ptr _pRsp) {
    TypeRequestHandlers::const_iterator itr;
    itr = _TypeHandlers.find(_szType);

    if (itr != _TypeHandlers.end()) {
        RequestHandlers& handlers = _TypeHandlers[_szType];
        string cmdName = _szCommandName;
        RequestHandlers::const_iterator rItr = handlers.find(cmdName);

        if (rItr == handlers.end()) {
            cmdName = REQ_NAME_ANY;
            rItr = handlers.find(cmdName); // Try to find universal handler.
        }

        if (rItr != handlers.end()) {
            RespondFunctionPtr func = handlers[cmdName];
            (m_pObj->*func)(_pReqMsg, _pRsp);
        } else {
            throw Poco::NotFoundException("Unknown request name", -1);
        }
    } else {
        throw Poco::NotFoundException("Unknown request type", -1);
    }
}

COmbMessenger::COmbMessenger():
    CRequestHandling(this),
    CManager(this),
    m_Resurector(*this, RECONNECT_PERIOD),
    m_pCipher(NULL),
    m_iRetransmissionCount(0){
    setLoggerName("OMB Messenger");
    m_Resurector.RegisterLogListener(this);
    AddFunctRequestHandler(Request::CRegisterUser::FUNCTIONALITY, Request::CRegisterUser::TYPE,
                           Request::CRegisterUser::NAME, &COmbMessenger::RegisterUser);
    AddRequestHandler(Request::CChangePassword::TYPE, Request::CChangePassword::NAME, &COmbMessenger::ChangePassword);
    AddRequestHandler(Request::CAddUser::TYPE, Request::CAddUser::NAME, &COmbMessenger::AddUser);
    AddRequestHandler(Request::CRemoveUser::TYPE, Request::CRemoveUser::NAME, &COmbMessenger::RemoveUser);
    AddRequestHandler(Request::CGetProp::TYPE, Request::CGetProp::NAME, &COmbMessenger::GetProp);
    AddRequestHandler(Request::CSetProp::TYPE, Request::CSetProp::NAME, &COmbMessenger::SetProp);
    AddRequestHandler(Request::CGetInfo::TYPE, Request::CGetInfo::NAME, &COmbMessenger::GetInfo);
    AddRequestHandler(Request::CServiceCmd::TYPE, Request::CServiceCmd::NAME, &COmbMessenger::ServiceCmd);
    AddRequestHandler(Request::CGetLoggers::TYPE, Request::CGetLoggers::NAME, &COmbMessenger::GetLoggers);
    AddRequestHandler(Request::CSetLoggers::TYPE, Request::CSetLoggers::NAME, &COmbMessenger::SetLoggers);
    AddRequestHandler(Request::CResetLoggers::TYPE, Request::CResetLoggers::NAME, &COmbMessenger::ResetLoggers);
#ifdef USE_TEST_API
    AddRequestHandler(Request::CGenPayload::TYPE, Request::CGenPayload::NAME, &COmbMessenger::GenPayload);
#endif
}

COmbMessenger::~COmbMessenger() {
    Stop();
}

void COmbMessenger::Start() {
    Poco::Mutex::ScopedLock lock(m_ModuleMutex);

    if (!IsRunning()) {
        trace("Starting...");
        CConfig::Ptr conf = SharedValues::CFactory::GetHelper<CConfig>();
        m_pOmb = new Messenger::CMessenger(conf->GetTag(),
                                           conf->GetUsername(),
                                           conf->GetPassword(),
                                           conf->GetAddr());
        m_pOmb->RegisterLogListener(this);
        m_pOmb->RegisterConnectionListener(&m_Resurector);
        // Get engine configuration.
        SharedValues::CGateway::Ptr gtw = SharedValues::CFactory::GetHelper<SharedValues::CGateway>();
        std::string gtwId = gtw->GetId();
        std::string gtwSecId = gtw->GetSecurity();
        // Create engine
        Poco::SHA1Engine engine;
        engine.update(gtwId);
        engine.update(gtwSecId);
        // Get hash
        std::string hashDigest(Poco::DigestEngine::digestToHex(engine.digest()));
        std::string hashKey = hashDigest.substr(0, 16);
        const Poco::Crypto::CipherKey::ByteVec key(hashKey.begin(), hashKey.end());
        Poco::Crypto::CipherKey::ByteVec vector;
        OnDebug("Encryption key: " + hashKey);
        // Create cipher
        m_pCipher = Poco::Crypto::CipherFactory::defaultFactory().createCipher(Poco::Crypto::CipherKey(ENCRYPTION_TYPE, key, vector));

        try {
            // Get UDM service.
            m_pUdm = m_pOmb->GetUser("svc/udm/gtw/" + gtwId);
            m_pUdm->RegisterForEvents(this);
        } catch (const Poco::Exception& e) {
            OnError(e.message());
        }

        // Connect OMB messenger.
        try {
            m_pOmb->RegisterResponder(this);
            m_pOmb->ActivateFunctionality(FUNCTIONALITY_AUTH);
            m_pOmb->Connect();
        } catch (const Poco::Exception& e) {
            OnError(e.message());
            m_Resurector.OnConnectionLost("Initial connect failed");
        }

        CAppModule::Start();
        information("Started.");
    }
}

void COmbMessenger::Stop() {
    Poco::Mutex::ScopedLock lock(m_ModuleMutex);

    if (IsRunning()) {
        trace("Stoping...");

        // Disconnect messenger.
        try {
            m_pOmb->Disconnect();
        } catch (const Poco::Exception& e) {
            OnError(e.message());
        }

        SAFE_DELETE(m_pOmb);
        CAppModule::Stop();
        information("Stoped.");
    }
}

void COmbMessenger::OnEvent(const std::string& _szSenderTag, const Messenger::CEventMessage::Ptr _pEvent) {
    OnDebug("New event received from: " + _szSenderTag);
#ifdef USE_OMB_BRIDGE

    if (_szSenderTag == m_pUdm->GetTag()) {
        if (_pEvent->GetName() == Event::CAssignment::NAME) {
            try {
                Event::CAssignment::Ptr pMsg = Event::CAssignment::Parse(_pEvent->GetParams());
                Poco::DynamicStruct params;
                params.insert(Bridge::PARAM_ACCOUNT, pMsg->GetAccountId());
                SendCommand(Bridge::MODULE_ID, Bridge::CMD_SET_ACCOUNT, params);
            } catch (const Poco::Exception& e) {
                OnError(e.message());
            }
        }
    }

#endif
}

void COmbMessenger::OnLastwill(const std::string& _szSenderTag, const Messenger::CLastwillMessage::Ptr _pLastwill) {
    std::stringstream ss;
    ss << "Lastwill (" << _pLastwill->IsOnline() << ") arrived from: " << _szSenderTag;
    OnDebug(ss.str());
}

const Messenger::CResponseMessage::Ptr COmbMessenger::Respond(const Messenger::CRequestMessage::Ptr _pRequest) {
    Messenger::CResponseMessage::Ptr response = NULL;

    if (IsRunning()) {
        response = new Messenger::CResponseMessage();

        try {
            CallHandler(_pRequest->GetType(), _pRequest->GetName(), _pRequest, response);
        } catch (const Poco::Exception& e) {
            OnTrace(e.message());
            int code = e.code();
            code = code == 0 ? -1 : code; // default code -1
            response->SetCode(code);
            response->SetMessage(e.message());
        }
    }

    return response;
}

const Messenger::CResponseMessage::Ptr COmbMessenger::FunctionalityRespond(const std::string& _szFunctionality, const Messenger::CRequestMessage::Ptr _pRequest) {
    Messenger::CResponseMessage::Ptr response = NULL;

    if (IsRunning()) {
        response = new Messenger::CResponseMessage();

        try {
            CallFunctHandler(_szFunctionality, _pRequest->GetType(), _pRequest->GetName(), _pRequest, response);
        } catch (const Poco::Exception& e) {
            OnTrace(e.message());
            int code = e.code();
            code = code == 0 ? -1 : code; // default code -1
            response->SetCode(code);
            response->SetMessage(e.message());
        }
    }

    return response;
}

void COmbMessenger::GetAssignment(){
    debug("Trying to get assignment...");
    Messenger::CRequestMessage::Ptr pReq = new Messenger::CRequestMessage();
    pReq->SetType(IAssignmentHandler::GET_ASSIGNMENT);
    pReq->SetName(IAssignmentHandler::GET_ASSIGNMENT);
    Poco::DynamicStruct::Ptr params = new Poco::DynamicStruct();
    pReq->SetParams(params);
    try {
        m_pUdm->Request(*pReq, MAKE_REQ_DELEGATE(COmbMessenger, OnAssignmentGet, OnAssignmentGetFail, this), IAssignmentHandler::ASSIGN_TIMEOUT);
    } catch (const Poco::Exception& e) {
        OnError("Failed to send assignment request. Reason: " + e.message());
    }
}

void COmbMessenger::OnAssignmentGet(unsigned int _nRequestId, const Messenger::CResponseMessage &_Response){
    m_iRetransmissionCount = 0;
    Poco::DynamicStruct::Ptr pParams = _Response.GetParams();
    debug("Assignment response: " + pParams->toString());
    std::string _accountId = pParams->operator [](IAssignmentHandler::ACCOUNT_ID);
    _accountId != Bridge::DEFAULT_ACCOUNT ? OnAssignmentSuccess(_accountId) : void();
}

void COmbMessenger::OnAssignmentGetFail(unsigned int _nRequestId){
    warning("Failed to retrieve account id, trying again...");
    if(m_iRetransmissionCount++ < IAssignmentHandler::MAX_RETRANSMISSIONS){
        debug("Retransmission of assignment...");
        GetAssignment();
    } else {
        m_iRetransmissionCount = 0;
    }
}

void COmbMessenger::OnAssignmentSuccess(const std::string &_accountId){
    try {
        Poco::DynamicStruct params;
        params.insert(Bridge::PARAM_ACCOUNT, _accountId);
        SendCommand(Bridge::MODULE_ID, Bridge::CMD_SET_ACCOUNT, params);
    } catch (const Poco::Exception& e) {
        OnError(e.message());
    }
}

void COmbMessenger::ProcessEvent(const std::string& _szName, const CEvent* const _pEvt) {
    if (IsRunning()) {
        Messenger::CEventMessage msg;
        msg.SetName(_szName);
        msg.SetParams(_pEvt->GetParams());

        try {
            m_pOmb->SendEvent(msg);
        } catch (const Poco::Exception& e) {
            std::stringstream ss;
            ss << "Unable to send event (n=" << _szName << ", m=" << e.message() << ")";
            error(ss.str());
        }
        if(_szName == MessageBroker::Bridge::EVENT_CLOUD_STATUS_CHANGED){
            int status = _pEvt->GetParams()->operator []("status");
            if(status == Bridge::STATUS_CONNECTED_NOT_ASSIGNED){
                GetAssignment();
            }
        }
    }
}

void COmbMessenger::Connect() {
    if (m_pOmb) {
        m_pOmb->Connect();
        GetAssignment();
    } else {
        throw Poco::IllegalStateException("Messenger is NULL.");
    }
}

void COmbMessenger::ChangePassword(const Messenger::CRequestMessage::Ptr _pReqMsg, Messenger::CResponseMessage::Ptr _pRsp) {
    string username = _pReqMsg->GetSender();
    Async::CChangePassword* opr = new Async::CChangePassword(username, username, NULL, NULL , NULL);
    AddOperation(opr);
}

void COmbMessenger::AddUser(const Messenger::CRequestMessage::Ptr _pReqMsg, Messenger::CResponseMessage::Ptr _pRsp) {
    if (_pReqMsg->GetSender().compare(0, 4, "gtw/") == 0) {
        Request::CAddUser::Ptr pReq = Request::CAddUser::Parse(_pReqMsg);
        Poco::DynamicStruct params;
        string username = Poco::format("%s/%s", pReq->GetGroup(), pReq->GetId());
        string password = GenerateHash(USER_PASSWORD_LENGTH);
        params.insert(Mosquitto::PARAM_USERNAME, username);
        params.insert(Mosquitto::PARAM_PASSWORD, password);
        int expiration = pReq->GetExpiration();
        expiration = (expiration > USER_EXPIRATION_DAYS) ? USER_EXPIRATION_DAYS : expiration; // maximum allowed is USER_EXPIRATION_DAYS
        params.insert(Mosquitto::PARAM_EXPIRATION, expiration);
        params.insert(Mosquitto::PARAM_NAME, pReq->GetName());
        params.insert(Mosquitto::PARAM_APPVERSION, pReq->GetVersion());
        params.insert(Mosquitto::PARAM_OSTYPE, pReq->GetType());
        SendCommand(Mosquitto::MODULE_ID, Mosquitto::CMD_ADD_CREDENTIALS, params);
        Poco::DynamicStruct::Ptr rspParams = new Poco::DynamicStruct();
        (*rspParams)["username"] = username;
        (*rspParams)["password"] = password;
        _pRsp->SetParams(rspParams);
    } else {
        throw Poco::NoPermissionException("Not authorized", -1);
    }
}

void COmbMessenger::RemoveUser(const Messenger::CRequestMessage::Ptr _pReqMsg, Messenger::CResponseMessage::Ptr _pRsp) {
    if (_pReqMsg->GetSender().compare(0, 4, "gtw/") == 0) {
        Request::CRemoveUser::Ptr pReq = Request::CRemoveUser::Parse(_pReqMsg);
        Poco::DynamicStruct params;
        params.insert(Mosquitto::PARAM_USERNAME, pReq->GetUsername());
        SendCommand(Mosquitto::MODULE_ID, Mosquitto::CMD_REMOVE_USER, params);
    } else {
        throw Poco::NoPermissionException("Not authorized", -1);
    }
}

void COmbMessenger::RegisterUser(const Messenger::CRequestMessage::Ptr _pReqMsg, Messenger::CResponseMessage::Ptr _pRsp) {
    Request::CRegisterUser::Ptr pReq = NULL;
    pReq = Request::CRegisterUser::Parse(_pReqMsg);
    std::string sender = _pReqMsg->GetSender();
    std::string id = pReq->GetId();
    std::string group = pReq->GetGroup();
    std::string username = pReq->GetUserName();
    std::string appversion = pReq->GetAppVersion();
    std::string ostype = pReq->GetOsType();
    Async::CRegUser* opr = new Async::CRegUser(sender, id, group, username, appversion, ostype);
    int regNum = opr->GetRegNumber();
    AddOperation(opr);
    Poco::DynamicStruct::Ptr params = new Poco::DynamicStruct();
    params->insert("reg_number", regNum);
    _pRsp->SetParams(params);
}

void COmbMessenger::GetProp(const Messenger::CRequestMessage::Ptr _pReqMsg, Messenger::CResponseMessage::Ptr _pRsp) {
    Request::CGetProp::Ptr pReq = NULL;
    pReq = Request::CGetProp::Parse(_pReqMsg);
    Poco::Dynamic::Var value = GetProperty(pReq->GetServiceName(), pReq->GetPropertyName());
    Poco::DynamicStruct::Ptr params = new Poco::DynamicStruct();
    params->insert(Request::CGetProp::KEY_VALUE, value);
    _pRsp->SetParams(params);
}

void COmbMessenger::SetProp(const Messenger::CRequestMessage::Ptr _pReqMsg,
                            Messenger::CResponseMessage::Ptr _pRsp) {
    Request::CSetProp::Ptr pReq = NULL;
    pReq = Request::CSetProp::Parse(_pReqMsg);
    UpdateProperty(pReq->GetServiceName(), pReq->GetPropertyName(),
                   pReq->GetPropertyValue());
}

void COmbMessenger::GetInfo(const Messenger::CRequestMessage::Ptr _pReqMsg, Messenger::CResponseMessage::Ptr _pRsp) {
    Poco::DynamicStruct::Ptr params = new Poco::DynamicStruct();
    params->insert("version", OMB_VERSION);
    params->insert("api_version", OMB_API_VERSION);
    _pRsp->SetParams(params);
}

void COmbMessenger::ServiceCmd(const Messenger::CRequestMessage::Ptr _pReqMsg, Messenger::CResponseMessage::Ptr _pRsp) {
    Request::CServiceCmd::Ptr pReq = NULL;
    pReq = Request::CServiceCmd::Parse(_pReqMsg);
    Poco::DynamicStruct::Ptr rsp = SendCommand(pReq->GetServiceName(), _pReqMsg->GetName(), *_pReqMsg->GetParams(), true);
    _pRsp->SetParams(rsp);
}

void COmbMessenger::GetLoggers(const Messenger::CRequestMessage::Ptr _pReqMsg, Messenger::CResponseMessage::Ptr _pRsp){
    Poco::DynamicStruct::Ptr loggers = new Poco::DynamicStruct();
    std::vector<std::string> vlog;
    Poco::Logger::names(vlog);
    for(unsigned int i = 0; i < vlog.size(); ++i){
        loggers->insert(vlog[i], Poco::Logger::get(vlog[i]).getLevel());
    }
    _pRsp->SetParams(loggers);
}

void COmbMessenger::SetLoggers(const Messenger::CRequestMessage::Ptr _pReqMsg, Messenger::CResponseMessage::Ptr _pRsp){
    information("Changing log levels...");
    Request::CSetLoggers::Ptr pReq = NULL;
    pReq = Request::CSetLoggers::Parse(_pReqMsg);
    if(pReq){
        Database::CParamsTable* paramsTable =
            Database::CDatabase::GetTable<Database::CParamsTable>();
        Poco::DynamicStruct loggers = pReq->GetLoggers();
        for(Poco::DynamicStruct::ConstIterator it = loggers.begin(); it != loggers.end(); ++it){
            Poco::Logger::setLevel(it->first, it->second.operator int());
            paramsTable->SetParam("log.level." + it->first, it->second.operator string());
        }
    }
}

void COmbMessenger::ResetLoggers(){
    information("Reseting log levels...");
    CLogger::InitializeDefaultLevels();
    Database::CParamsTable* paramsTable =
        Database::CDatabase::GetTable<Database::CParamsTable>();
    paramsTable->DeleteParamsLike("log.level.");
}

void COmbMessenger::ResetLoggers(const Messenger::CRequestMessage::Ptr _pReqMsg, Messenger::CResponseMessage::Ptr _pRsp){
    ResetLoggers();
}

#ifdef USE_TEST_API
void COmbMessenger::GenPayload(const Messenger::CRequestMessage::Ptr _pReqMsg, Messenger::CResponseMessage::Ptr _pRsp) {
    Request::CGenPayload::Ptr pReq = NULL;
    pReq = Request::CGenPayload::Parse(_pReqMsg);
    Poco::DynamicStruct::Ptr params = new Poco::DynamicStruct();
    unsigned int size = pReq->GetSize();
    Poco::Timestamp time;
    string payload(size, 'a');
    params->insert("payload", payload);
    Poco::Timespan::TimeDiff diff = time.elapsed();
    params->insert("diff", diff);
    _pRsp->SetParams(params);
}
#endif

std::string COmbMessenger::DecryptString(const std::string& _szEncoded) {
    return m_pCipher->decryptString(_szEncoded, Poco::Crypto::Cipher::ENC_BASE64);
}

Messenger::CUser::Ptr COmbMessenger::GetUser(const std::string& _szTag) {
    Messenger::CUser::Ptr user = NULL;

    if (m_pOmb) {
        user = m_pOmb->GetUser(_szTag);
    } else {
        throw Poco::IllegalStateException("Messenger is NULL.");
    }

    return user;
}

Poco::DynamicStruct::Ptr COmbMessenger::SendOmbCommand(const CAppModule::ID& _ModuleId, const std::string& _szCmdName, const Poco::DynamicStruct& _Params) {
    return SendCommand(_ModuleId, _szCmdName, _Params);
}

void COmbMessenger::OnTrace(const std::string& _Msg) {
    trace(_Msg);
}

void COmbMessenger::OnDebug(const std::string& _Msg) {
    debug(_Msg);
}

void COmbMessenger::OnInformation(const std::string& _Msg) {
    information(_Msg);
}

void COmbMessenger::OnWarning(const std::string& _Msg) {
    warning(_Msg);
}

void COmbMessenger::OnError(const std::string& _Msg) {
    error(_Msg);
}

const string CConfig::KEY_BASE = "messenger.";

CConfig::CConfig(): SharedValues::CHelper(KEY_BASE) {
}

std::string CConfig::GetTag() const {
    std::string v = m_pConfigValues->getString(m_szKeyBase + "tag", "");

    if (v.empty()) {
        SharedValues::CGateway::Ptr gtw = SharedValues::CFactory::GetHelper<SharedValues::CGateway>();
        v = "gtw/" + gtw->GetId() + "/omb";
    }

    return v;
}

std::string CConfig::GetUsername() const {
    return m_pConfigValues->getString(m_szKeyBase + "username", "app/omb");
}

std::string CConfig::GetPassword() const {
    return m_pConfigValues->getString(m_szKeyBase + "password", "");
}

std::string CConfig::GetAddr() const {
    return m_pConfigValues->getString(m_szKeyBase + "addr", "tcp://localhost:1883");
}

}
}
