/****************************************************************************
 *  Copyright (c) 2016 RT-RK d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.rt-rk.com
 *  -----------------------------------------------------
 *
 * \project Oblo Message Broker - Broker application.
 * \file OmbMessenger.h
 * \brief
 * Messaging controller.
 *
 * @Author Milan Tucic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef OMB_OMB_MESSENGER_H
#define OMB_OMB_MESSENGER_H

#include "core/AppModule.h"
#include "core/omb-messenger/Messages.h"
#include "core/Logger.h"
#include "core/omb-messenger/Operations.h"
#include "core/SharedValues.h"

#include "messenger/Messenger.h"
#include "Poco/Crypto/Cipher.h"

namespace MessageBroker {
namespace OmbMessenger {

class COmbMessenger;

class CRequestHandling {
public:
    typedef void (COmbMessenger::*RespondFunctionPtr)(const Messenger::CRequestMessage::Ptr, Messenger::CResponseMessage::Ptr);
    typedef std::pair<std::string, RespondFunctionPtr> RequestHandler;
    typedef std::map<std::string, RespondFunctionPtr> RequestHandlers;
    typedef std::map<std::string, RequestHandlers> TypeRequestHandlers;
    typedef std::map<std::string, TypeRequestHandlers> FunctTypeReqHandlers;
    CRequestHandling(COmbMessenger* _pObj);
protected:
    virtual void AddRequestHandler(const std::string& _szType, const std::string& _szCommandName,
                                   const RespondFunctionPtr _pFunc);
    virtual void CallHandler(const std::string& _szType, const std::string& _szCommandName,
                             const Messenger::CRequestMessage::Ptr _pReqMsg, Messenger::CResponseMessage::Ptr _pRsp);
    virtual void AddFunctRequestHandler(const std::string& _szFunctionality, const std::string& _szType,
                                        const std::string& _szCommandName, const RespondFunctionPtr _pFunc);
    virtual void CallFunctHandler(const std::string& _szFunctionality, const std::string& _szType,
                                  const std::string& _szCommandName, const Messenger::CRequestMessage::Ptr _pReqMsg,
                                  Messenger::CResponseMessage::Ptr _pRsp);
private:
    virtual void AddTypeRequestHandler(TypeRequestHandlers& _TypeHandlers, const std::string& _szType,
                                       const std::string& _szCommandName, const RespondFunctionPtr _pFunc);
    void CallTypeHandler(TypeRequestHandlers& _TypeHandlers, const std::string& _szType, const std::string& _szCommandName,
                         const Messenger::CRequestMessage::Ptr _pReqMsg, Messenger::CResponseMessage::Ptr _pRsp);

    TypeRequestHandlers m_RequestHandlers;
    FunctTypeReqHandlers m_FunctReqHandlers;
    Poco::Mutex m_Mutex;
    COmbMessenger* m_pObj;
};

class CConfig: public SharedValues::CHelper {
public:
    typedef Poco::SharedPtr<CConfig> Ptr;
    static const string KEY_BASE;

    CConfig();
    string GetTag() const;
    string GetUsername() const;
    string GetPassword() const;
    string GetAddr() const;
};

class IAssignmentHandler{
public:
    static const std::string    GET_ASSIGNMENT;
    static const std::string    ACCOUNT_ID;
    static const int            ASSIGN_TIMEOUT;
    static const int            MAX_RETRANSMISSIONS;

    virtual void GetAssignment() = 0;
    virtual void OnAssignmentSuccess(const std::string &_accountId) = 0;
    virtual void OnAssignmentGet(unsigned int _nRequestId, const Messenger::CResponseMessage& _Response) = 0;
    virtual void OnAssignmentGetFail(unsigned int _nRequestId) = 0;
};

class COmbMessenger:
    public CAppModule,
    public CRequestHandling,
    public Messenger::IEventListener,
    public Messenger::IResponder,
    public Messenger::IConnector,
    public Messenger::ILogListener,
    public Async::CManager,
    public IAssignmentHandler{
public:
    COmbMessenger();
    ~COmbMessenger();

    void Start();
    void Stop();

    virtual void OnEvent(const std::string& _szSenderTag, const Messenger::CEventMessage::Ptr _pEvent);
    virtual void OnLastwill(const std::string& _szSenderTag, const Messenger::CLastwillMessage::Ptr _pLastwill);
    virtual const Messenger::CResponseMessage::Ptr Respond(const Messenger::CRequestMessage::Ptr _pRequest);
    virtual const Messenger::CResponseMessage::Ptr FunctionalityRespond(const std::string& _szFunctionality,
            const Messenger::CRequestMessage::Ptr _pRequest);
    virtual void ProcessEvent(const std::string& _szName, const CEvent* const _pEvt);

    void GetAssignment();
    void OnAssignmentSuccess(const std::string &_accountId);
    void OnAssignmentGet(unsigned int _nRequestId, const Messenger::CResponseMessage& _Response);
    void OnAssignmentGetFail(unsigned int _nRequestId);
private:
    virtual void Connect();
    void ResetLoggers();
    void SetLogLevels();

    void ChangePassword(const Messenger::CRequestMessage::Ptr _pReqMsg, Messenger::CResponseMessage::Ptr _pRsp);
    void AddUser(const Messenger::CRequestMessage::Ptr _pReqMsg, Messenger::CResponseMessage::Ptr _pRsp);
    void RemoveUser(const Messenger::CRequestMessage::Ptr _pReqMsg, Messenger::CResponseMessage::Ptr _pRsp);
    void RegisterUser(const Messenger::CRequestMessage::Ptr _pReqMsg, Messenger::CResponseMessage::Ptr _pRsp);
    void GetProp(const Messenger::CRequestMessage::Ptr _pReqMsg, Messenger::CResponseMessage::Ptr _pRsp);
    void SetProp(const Messenger::CRequestMessage::Ptr _pReqMsg, Messenger::CResponseMessage::Ptr _pRsp);
    void GetInfo(const Messenger::CRequestMessage::Ptr _pReqMsg, Messenger::CResponseMessage::Ptr _pRsp);
    void ServiceCmd(const Messenger::CRequestMessage::Ptr _pReqMsg, Messenger::CResponseMessage::Ptr _pRsp);
    void GetLoggers(const Messenger::CRequestMessage::Ptr _pReqMsg, Messenger::CResponseMessage::Ptr _pRsp);
    void SetLoggers(const Messenger::CRequestMessage::Ptr _pReqMsg, Messenger::CResponseMessage::Ptr _pRsp);
    void ResetLoggers(const Messenger::CRequestMessage::Ptr _pReqMsg, Messenger::CResponseMessage::Ptr _pRsp);
#ifdef USE_TEST_API
    void GenPayload(const Messenger::CRequestMessage::Ptr _pReqMsg, Messenger::CResponseMessage::Ptr _pRsp);
#endif
    virtual std::string DecryptString(const std::string& _szEncoded);
    virtual Messenger::CUser::Ptr GetUser(const std::string& _szTag);
    virtual Poco::DynamicStruct::Ptr SendOmbCommand(const CAppModule::ID& _ModuleId, const std::string& _szCmdName,
            const Poco::DynamicStruct& _Params);

    virtual void OnTrace(const std::string& _Msg);
    virtual void OnDebug(const std::string& _Msg);
    virtual void OnInformation(const std::string& _Msg);
    virtual void OnWarning(const std::string& _Msg);
    virtual void OnError(const std::string& _Msg);

    Messenger::CUser::Ptr m_pUdm;
    Messenger::CMessenger* m_pOmb;
    Messenger::CResurector m_Resurector;
    Poco::Crypto::Cipher::Ptr m_pCipher;
    int m_iRetransmissionCount;
};

}
}
#endif
