#include "core/omb-messenger/Operations.h"
#include "core/omb-messenger/Messages.h"
#include "core/Globals.h"

namespace MessageBroker {
namespace OmbMessenger {
namespace Async {

Poco::Random COperation::m_RandomGenerator;

void COperation::SetContext(CManager* _pMsgr) {
    m_pMsgr = _pMsgr;
}

void COperation::SetLogListener(Messenger::ILogListener* _pLog) {
    m_pLog = _pLog;
}

void COperation::OnTrace(const std::string& _Msg) {
    if (m_pLog) {
        m_pLog->OnTrace(_Msg);
    }
}

void COperation::OnDebug(const std::string& _Msg) {
    if (m_pLog) {
        m_pLog->OnDebug(_Msg);
    }
}

void COperation::OnInformation(const std::string& _Msg) {
    if (m_pLog) {
        m_pLog->OnInformation(_Msg);
    }
}

void COperation::OnWarning(const std::string& _Msg) {
    if (m_pLog) {
        m_pLog->OnWarning(_Msg);
    }
}

void COperation::OnError(const std::string& _Msg) {
    if (m_pLog) {
        m_pLog->OnError(_Msg);
    }
}

CManager::CManager(Messenger::ILogListener* _pLog):
    m_AsyncOperator(*this), m_pLog(_pLog) {
    m_Thread.start(m_AsyncOperator);
}

CManager::~CManager() {
    m_AsyncOperator.Stop();
    m_Thread.join();
}

void CManager::AddOperation(COperation* _pOperation) {
    _pOperation->SetContext(this);
    _pOperation->SetLogListener(m_pLog);
    m_AsyncOperator.Add(_pOperation);
}

void CManager::OnProcess(COperation* _pOperation) {
    _pOperation->Process();
}

CChangePassword::CChangePassword(const std::string& _szSender, const std::string& _szUsername, const std::string& _szName, const std::string& _szAppversion, const std::string& _szOsType, int _nExpiration):
    m_szSender(_szSender), m_szUsername(_szUsername), m_szName(_szName), m_szAppVersion(_szAppversion), m_szOsType(_szOsType), m_nExpiration(_nExpiration) {
}

void CChangePassword::Process() {
    m_szPassword = GenerateHash(USER_PASSWORD_LENGTH);
    Messenger::CRequestMessage::Ptr req =
        Request::CMsgFactory::MakeChangeCredentials(m_szUsername, m_szPassword, m_nExpiration);

    try {
        Messenger::CUser::Ptr sender = m_pMsgr->GetUser(m_szSender);
        sender->Request(*req, MAKE_AUTO_DEL_REQ_DELEGATE(CChangePassword, OnSuccess, OnFailure, this));
    } catch (const Poco::Exception& e) {
        OnError("Failed to generate a password for user: <"
                + m_szUsername + ">. Reason: " + e.message());
    }
}

void CChangePassword::OnSuccess(unsigned int _nRequestId, const Messenger::CResponseMessage& _Response) {
    OnTrace("Response for credentials update arrived.");

    if (_Response.GetCode() == 0) {
        OnDebug("Adding credentials: username <" + m_szUsername +
                ">, password <" + m_szPassword + ">" + " Name: " + m_szName + " Ostype: " + m_szOsType + " Appversion: " + m_szAppVersion);
        Poco::DynamicStruct params;
        params.insert(Mosquitto::PARAM_USERNAME, m_szUsername);
        params.insert(Mosquitto::PARAM_PASSWORD, m_szPassword);

        if (m_nExpiration > 0) {
            params.insert(Mosquitto::PARAM_EXPIRATION, m_nExpiration);
        }

        params.insert(Mosquitto::PARAM_NAME, m_szName);
        params.insert(Mosquitto::PARAM_APPVERSION, m_szAppVersion);
        params.insert(Mosquitto::PARAM_OSTYPE, m_szOsType);
        m_pMsgr->SendOmbCommand(Mosquitto::MODULE_ID, Mosquitto::CMD_ADD_CREDENTIALS, params);
    } else {
        OnInformation("Failed to generate credentials for <" + m_szUsername + ">");
    }
}

void CChangePassword::OnFailure(unsigned int _nRequestId) {
    std::stringstream ss;
    ss << "Failed to generate credentials for request <" << _nRequestId << ">.";
    OnError(ss.str());
}

CRegUser::CRegUser(const std::string& _szSender, const std::string& _szId, const std::string& _szGroup, const std::string& _szusername, const std::string& _szappversion, const std::string& _szostype):
    m_szSender(_szSender), m_szId(_szId), m_szGroup(_szGroup), m_szUserName(_szusername), m_szAppVersion(_szappversion), m_szOsType(_szostype) {
    m_nRegNum = m_RandomGenerator.next();
}

int CRegUser::GetRegNumber() const {
    return m_nRegNum;
}

void CRegUser::Process() {
    m_szChallenge = GenerateHash(CHALLENGE_CODE_LENGTH); // 10 byte code. Do not use lenght that is multiplication of 16, because of padding problem.
    Messenger::CRequestMessage::Ptr req = Request::CMsgFactory::MakeRegChallenge(m_szChallenge);

    try {
        Messenger::CUser::Ptr user = m_pMsgr->GetUser(m_szSender);
        user->Request(*req, MAKE_AUTO_DEL_REQ_DELEGATE(CRegUser, OnSuccess, OnFailure, this));
    } catch (const Poco::Exception& e) {
        OnError("Failed to authorize mobile user: <"
                + m_szId + ">. Reason: " + e.message());
    }
}

void CRegUser::OnSuccess(unsigned int _nRequestId, const Messenger::CResponseMessage& _Response) {
    OnTrace("Response from mobile user arrived. Authorization in progress.");
    Response::CRegChallenge::Ptr rspChallenge;
    std::string recvCode;

    try {
        rspChallenge = Response::CRegChallenge::Parse(_Response);
        recvCode = rspChallenge->GetCode();
    } catch (const Poco::Exception& e) {
        OnError("Failed to extract code from challenge response.");
    }

    if (!recvCode.empty()) {
        std::string decryptedCode;

        try {
            decryptedCode = m_pMsgr->DecryptString(recvCode);
        } catch (const Poco::Exception& e) {
            OnError("Failed on decrypt <" + recvCode + ">: " + e.message());
        }

        if (m_szChallenge == decryptedCode) {
            std::string username;
            Poco::format(username, "%s/%s", m_szGroup, m_szId);
            Async::CFinalizeUserReg* opr =
                new Async::CFinalizeUserReg(m_szSender, username, USER_EXPIRATION_DAYS, m_nRegNum, m_szUserName, m_szAppVersion, m_szOsType);
            m_pMsgr->AddOperation(opr);
            OnInformation("User <" + username + "> succesfully queued for credentials.");
        } else {
            OnInformation("User registration failed on challenge => " + m_szChallenge + "!=" + decryptedCode);
        }
    }
}

void CRegUser::OnFailure(unsigned int _nRequestId) {
    std::stringstream ss;
    ss << "Failed to register user with request ID <" << _nRequestId << ">.";
    OnError(ss.str());
}

CFinalizeUserReg::CFinalizeUserReg(const std::string& _szSender,
                                   const std::string& _szUsername,
                                   int _nExpiration,
                                   int _nRegNum,
                                   const std::string& _szName,
                                   const std::string& _szAppVersion,
                                   const std::string& _szOsType): CChangePassword(_szSender, _szUsername, _szName, _szAppVersion, _szOsType, _nExpiration) {
    m_nRegNum = _nRegNum;
}

void CFinalizeUserReg::OnSuccess(unsigned int _nRequestId, const Messenger::CResponseMessage& _Response) {
    OnTrace("Response for credentials update arrived.");

    if (_Response.GetCode() == 0) {
        // It is part of registration procedure.
        const Poco::DynamicStruct::Ptr params = _Response.GetParams();

        if (params.isNull()) {
            OnError("No params received in a change credentials response. Breaking procedure for <" + m_szUsername + ">.");
            return;
        }

        try {
            int receivedRegNum = (*params)["reg_number"].convert<int>();

            if (receivedRegNum != m_nRegNum) {
                OnError("User missed registration number in a response message. Breaking procedure for <" + m_szUsername + ">.");
                return;
            }
        } catch (const Poco::Exception& e) {
            OnError("Failed to extract registration number from change credentials response. Breaking procedure for <" + m_szUsername + ">.");
            return;
        }

        CChangePassword::OnSuccess(_nRequestId, _Response);
    } else {
        OnInformation("Failed to generate credentials for <" + m_szUsername + ">");
    }
}


}
}
}
