/****************************************************************************
 *  Copyright (c) 2018 OBLO Living LLC
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.obloliving.com
 *  -----------------------------------------------------
 *
 * \project Oblo Message Broker - Broker application.
 * \file bridge/AsyncActions.h
 * \brief
 * OBLO bridge async actions.
 *
 * @Author Milan Tucic
 *
 * \notes
 *
 * \history
 * 28.8.2018 moved from Bridge.h
 *****************************************************************************/
#include "AsyncActions.h"
#include "Config.h"
#include "Bridge.h"
#include <core/Clients.h>
#include <core/Globals.h>
#include <sstream>

namespace MessageBroker {
namespace Bridge {

namespace Async {

CAsyncFunc::CAsyncFunc(const AsyncFunctionPtr _pFunc, void* _pParams):
    m_pFunc(_pFunc), m_pParams(_pParams) {
}

CAsyncFunc::~CAsyncFunc() {
}

CAsyncFuncHandler::CAsyncFuncHandler(CBridge* _pObj) :
    m_AsyncProcessor(*this),
    m_pObj(_pObj) {
    m_AsyncThread.start(m_AsyncProcessor);
}

CAsyncFuncHandler::~CAsyncFuncHandler() {
    m_AsyncProcessor.Stop();
    m_AsyncThread.join();
}

void CAsyncFuncHandler::Run(CAsyncFunc* _pFunc) {
    m_AsyncProcessor.Add(_pFunc);
}

void CAsyncFuncHandler::OnProcess(CAsyncFunc* _pFunc) {
    CAsyncFunc::AsyncFunctionPtr func = _pFunc->m_pFunc;
    (m_pObj->*func)(_pFunc->m_pParams);
    delete _pFunc;
}

void CSessionCleaner::Run(const std::string& _szCloudAddr,
                          Poco::Timestamp::TimeDiff _nTimeout) {
    m_bConnected = false;
    // Configure client.
    CConfigCloud::Ptr cloudCnfg = SharedValues::CFactory::GetHelper<CConfigCloud>();
    Mqtt::CAsyncClient cli(cloudCnfg->GetId(), _szCloudAddr, NULL);
    cli.SetUsername(cloudCnfg->GetUsername());
    cli.SetPassword(cloudCnfg->GetPassword());
    cli.SetCleanSession(true); // Clean session
    cli.SetConnectionTimeoutInterval(cloudCnfg->GetTimeoutInterval());
#ifdef USE_SECURE_CONNECTION

    // Check if broker address requires secure connection.
    if (_szCloudAddr.compare(0, 6, "ssl://") == 0) {
        std::string stratusCACrt = cloudCnfg->GetCaFile();
        cli.SetSecureConfiguration(stratusCACrt);
    }

#endif
    Poco::Timestamp start;

    while (m_bRunning && !m_bConnected && !(_nTimeout &&
                                            start.isElapsed(_nTimeout))) {
        try {
            cli.Connect();
            m_bConnected = true;
        } catch (const Poco::Exception& e) {
            // Failed to connect.
            std::stringstream ss;
            ss << "(CS) Failed to connect to the cloud broker (c=" << e.displayText() <<
               ")";
            m_pObj->debug(ss.str());
            Poco::Thread::sleep(5000);
        }
    }

    cli.Disconnect();
}

void CSessionCleaner::Cancel() {
    m_bRunning = false;
}

bool CSessionCleaner::IsCleaned() const {
    return m_bConnected;
}

bool CSessionCleaner::IsRunning() const {
    return m_bRunning;
}

CUpdateCloudAddress::CUpdateCloudAddress(CBridge* _pObj,
        const std::string& _szAddr):
    m_pObj(_pObj),
    m_szAddr(_szAddr),
    m_Cleaner(m_pObj) {
}

void CUpdateCloudAddress::run() {
    m_Cleaner.Run(m_szAddr, ADDR_CHECK_DURATION);
    m_pObj->OnAddrUpdateResult(m_szAddr, m_Cleaner.IsCleaned() &&
                               m_Cleaner.IsRunning());
}

void CUpdateCloudAddress::Cancel() {
    if (!m_Cleaner.IsCleaned()) {
        m_Cleaner.Cancel();
    }
}

bool CUpdateCloudAddress::IsCanceled() const {
    return !m_Cleaner.IsRunning();
}

} // Async
} // Bridge
} // MessageBroker
