/****************************************************************************
 *  Copyright (c) 2018 OBLO Living LLC
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.obloliving.com
 *  -----------------------------------------------------
 *
 * \project Oblo Message Broker - Broker application.
 * \file Bridge.h
 * \brief
 * OBLO bridge classes.
 *
 * @Author Milan Tucic
 *
 * \notes
 *
 * \history
 * 2016 Initial version
 *****************************************************************************/
#include "feature/bridge/Bridge.h"
#include "core/Globals.h"
#include "Poco/SHA1Engine.h"
#include "Config.h"
#include "Events.h"
#include <sstream>

namespace MessageBroker {
namespace Bridge {

using namespace std;

#define MAKE_MSG_DELEGATE(m) Mqtt::CMsgDelegate::CreateDelegate<CBridge, &CBridge::m>(this)
#define ASYNC_CALL(f) Run(new Async::CAsyncFunc(&CBridge::f))
#define ASYNC_CALL_P(f, p) Run(new Async::CAsyncFunc(&CBridge::f, p))

CBridge::CBridge():
    Async::CAsyncFuncHandler(this),
    m_pLocalClient(NULL),
    m_pCloudClient(NULL),
    m_szAccount(DEFAULT_ACCOUNT),
    m_nMsgQueueMax(0),
    m_pCloudAddrUpdater(NULL),
    m_pCloudAddrUpdaterThread(NULL) {
    setLoggerName("Bridge");
    AddCommandExecutor(CMD_SET_ACCOUNT, MAKE_CMD_EXEC_DELEGATE(CBridge, CmdSetGlobalAccount));
    AddCommandExecutor(CMD_UPDATE_PROP, MAKE_CMD_EXEC_DELEGATE(CBridge, CmdUpdateProperty));
    AddCommandExecutor(CMD_GET_PROP, MAKE_CMD_EXEC_DELEGATE(CBridge, CmdGetProperty));
    AddApiCommandExecutor(CMD_SUBSCRIBE, MAKE_CMD_EXEC_DELEGATE(CBridge, CmdSubscribe));
    AddApiCommandExecutor(CMD_UNSUBSCRIBE, MAKE_CMD_EXEC_DELEGATE(CBridge, CmdUnsubscribe));
}

CBridge::~CBridge() {
    Stop();
}

void CBridge::Start() {
    trace("Starting...");
    Poco::Mutex::ScopedLock lock(m_ModuleMutex);

    if (!IsRunning()) {
        m_szGatewayId = SharedValues::CFactory::GetHelper<SharedValues::CGateway>()->GetId();
        m_nMsgQueueMax = SharedValues::CFactory::GetHelper<CConfigCloud>()->GetMsgQueueLimit();
        m_szAccount = DEFAULT_ACCOUNT;
        // Configure
        ConstructRoutes(DEFAULT_ACCOUNT);
        ConfigureCloudBridge(DEFAULT_ACCOUNT, true);
        ConfigureLocalBridge();
        // Run
        CAppModule::Start();
        m_pCloudClient->SetConnectionListener(this);
        m_pLocalClient->Start();
        m_pCloudClient->Start();
        information("Started.");
    }
}

void CBridge::Stop() {
    trace("Stoping...");
    Poco::Mutex::ScopedLock lock(m_ModuleMutex);

    if (IsRunning()) {
        CAppModule::Stop();
        // Terminate clients

        if (m_pCloudClient)  {
            m_pCloudClient->Stop();
            SAFE_DELETE(m_pCloudClient);
        }

        if (m_pLocalClient) {
            m_pLocalClient->Stop();
            SAFE_DELETE(m_pLocalClient);
        }

        // Clear routes
        m_Routes.clear();
        // Cancel cloud address updater.
        ClearCloudUpdater(NULL);
        information("Stoped.");
    }
}

void CBridge::RestartCloudConnection(const string& _szAccount, bool _bCleanSession) {
    Poco::Mutex::ScopedLock lock(m_ModuleMutex);

    if (IsRunning()) {
        // Stop
        if (m_pCloudClient)  {
            m_pCloudClient->Stop();
            SAFE_DELETE(m_pCloudClient);
        }

        // Clear routes
        m_Routes.clear();
        // Assign new account, cloud client must be stopped at that time.
        m_szAccount = _szAccount;
        // Re-start
        ConstructRoutes(m_szAccount);
        ConfigureCloudBridge(m_szAccount, _bCleanSession);
        m_pCloudClient->SetConnectionListener(this);
        m_pCloudClient->Start();
        SendMessageQueue();
        information("Cloud connection restarted (a=" + m_szAccount + ")");
    }
}

bool CBridge::IsAssigned() const {
    return m_szAccount != DEFAULT_ACCOUNT;
}

bool CBridge::IsConnectionUsable() const {
    // Connection is valid if account is set, client is initialized and clean session is set to FALSE
    return IsAssigned() && (m_pCloudClient ? !m_pCloudClient->IsCleanSessionSet() : false);
}

bool CBridge::IsUdmDefaultRequest(const std::string& _szTopic) {
    stringstream udmTopic;
    udmTopic << "svc/udm/gtw/" << m_szGatewayId << "/req";
    return m_pCloudClient && (_szTopic == udmTopic.str());
}

void CBridge::ConfigureLocalBridge() {
    // Configure client
    CConfigLocal::Ptr localCnfg = SharedValues::CFactory::GetHelper<CConfigLocal>();
    // Local client
    m_pLocalClient = new CClient(localCnfg->GetId(), localCnfg->GetAddr(), MODULE_ID, "loc");
    m_pLocalClient->SetReconnectionInterval(localCnfg->GetReconnectPeriod());
    m_pLocalClient->SetUsername(localCnfg->GetUsername());
    m_pLocalClient->SetPassword(localCnfg->GetPassword());
    m_pLocalClient->SetCleanSession(false);
    m_pLocalClient->SetKeepAliveInterval(localCnfg->GetKeepAliveInterval());
    m_pLocalClient->RegisterObserver(MAKE_MSG_DELEGATE(RedirectToCloud));
    m_pLocalClient->SetConnectionTimeoutInterval(localCnfg->GetTimeoutInterval());
    // Add subscriptions
    CClient::SubscriptionCollection subs;
    GetLocalSubscriptions(subs);
    m_pLocalClient->Subscribe(subs);
}

void CBridge::ConfigureCloudBridge(const std::string& _szAccount, bool _bCleanSession) {
    // Configure client
    CConfigCloud::Ptr cloudCnfg = SharedValues::CFactory::GetHelper<CConfigCloud>();
    std::string cloudAddr = cloudCnfg->GetAddr();
    m_pCloudClient = new CClient(cloudCnfg->GetId(), cloudAddr, MODULE_ID, "cld");
    m_pCloudClient->SetReconnectionInterval(cloudCnfg->GetReconnectPeriodMinimum(), cloudCnfg->GetReconnectPeriodMaximum());
    m_pCloudClient->SetUsername(cloudCnfg->GetUsername());
    m_pCloudClient->SetPassword(cloudCnfg->GetPassword());
    m_pCloudClient->SetCleanSession(_bCleanSession);
    m_pCloudClient->SetConnectionTimeoutInterval(cloudCnfg->GetTimeoutInterval());
    m_pCloudClient->SetKeepAliveInterval(cloudCnfg->GetKeepAliveInterval());
    m_pCloudClient->RegisterObserver(MAKE_MSG_DELEGATE(RedirectToLocal));
    m_pCloudClient->IgnoreSessionPresent(true); // subscribe even if the session is present.
    m_pCloudClient->SetDnsCacheLimit(cloudCnfg->GetDnsCacheLimit());
    // Add subscriptions
    CClient::SubscriptionCollection subs;
    GetCloudSubscriptions(subs, _szAccount);
    m_pCloudClient->Subscribe(subs);
    std::string statusTopic = "oblo/" + _szAccount +  "/gtw/" + m_szGatewayId;

    if (_szAccount != DEFAULT_ACCOUNT) {
        if(!_bCleanSession) {
            // Activate global LWM only after session is cleared.
            m_pCloudClient->ActivateStatusMessage(statusTopic);
        }
    } else {
        // Force subscription on connect. By default, it will wait for clean session = false to do subscriptions.
        m_pCloudClient->SetSubscribeOnConnect(true);

        if (cloudCnfg->IsStatusForced()) {
            // Status (lwm:online) message is configured to be sent even when account is not set.
            m_pCloudClient->ActivateStatusMessage(statusTopic);
        }
    }

    // Check if broker address requires secure connection.
    if (cloudAddr.compare(0, 6, "ssl://") == 0) {
        std::string stratusCACrt = cloudCnfg->GetCaFile();
        m_pCloudClient->SetSecureConfiguration(stratusCACrt);
    }
}

void CBridge::ConstructRoutes(const std::string& _szAccount) {
    if (_szAccount != DEFAULT_ACCOUNT) {
        // Gateways
        m_Routes.push_back(RouteElement("gtw/", "oblo/" + _szAccount + "/gtw/"));
        // Mobile clients
        m_Routes.push_back(RouteElement("cli/", "oblo/" + _szAccount + "/cli/"));
    } else {
        m_Routes.push_back(RouteElement("gtw/" + m_szGatewayId + "/omb/rsp", "oblo/" + _szAccount + "/gtw/" + m_szGatewayId + "/omb/rsp"));
    }

    // Stratus services
    m_Routes.push_back(RouteElement("svc/", "oblo/" + _szAccount + "/svc/"));
    m_Routes.push_back(RouteElement("svc/", "oblo/svc/"));
}

void CBridge::GetLocalSubscriptions(CClient::SubscriptionCollection& _Subscriptions) const {
    // Local modules
    _Subscriptions.push_back(CClient::Subscription("gtw/" + m_szGatewayId + "/+/evt", Mqtt::CMessage::QOS1));
    _Subscriptions.push_back(CClient::Subscription("gtw/" + m_szGatewayId + "/+/lwm", Mqtt::CMessage::QOS1));
    //gtw sys functionality
    _Subscriptions.push_back(CClient::Subscription("gtw/" + m_szGatewayId + "/+/sys/evt", Mqtt::CMessage::QOS1));
    _Subscriptions.push_back(CClient::Subscription("gtw/" + m_szGatewayId + "/ohm/azc/evt", Mqtt::CMessage::QOS1));
    // Mobile clients
    _Subscriptions.push_back(CClient::Subscription("cli/+/req", Mqtt::CMessage::QOS0));
    _Subscriptions.push_back(CClient::Subscription("cli/+/rsp", Mqtt::CMessage::QOS0));
    _Subscriptions.push_back(CClient::Subscription("cli/+/evt", Mqtt::CMessage::QOS1));
    _Subscriptions.push_back(CClient::Subscription("cli/+/lwm", Mqtt::CMessage::QOS1));
    // Stratus services
    _Subscriptions.push_back(CClient::Subscription("svc/+/gtw/" + m_szGatewayId + "/req", Mqtt::CMessage::QOS0));
    _Subscriptions.push_back(CClient::Subscription("svc/+/gtw/" + m_szGatewayId + "/rsp", Mqtt::CMessage::QOS0));
}

void CBridge::GetCloudSubscriptions(CClient::SubscriptionCollection& _Subscriptions, const std::string& _szAccount) const {
    if (_szAccount != DEFAULT_ACCOUNT) {
        // Local modules
        _Subscriptions.push_back(CClient::Subscription("oblo/" + _szAccount + "/gtw/" + m_szGatewayId + "/+/req", Mqtt::CMessage::QOS0));
        _Subscriptions.push_back(CClient::Subscription("oblo/" + _szAccount + "/gtw/" + m_szGatewayId + "/+/+/req", Mqtt::CMessage::QOS0));
        _Subscriptions.push_back(CClient::Subscription("oblo/" + _szAccount + "/gtw/" + m_szGatewayId + "/+/rsp", Mqtt::CMessage::QOS0));
        // Mobile clients
        _Subscriptions.push_back(CClient::Subscription("oblo/" + _szAccount + "/cli/+/evt", Mqtt::CMessage::QOS1));
        _Subscriptions.push_back(CClient::Subscription("oblo/" + _szAccount + "/cli/+/lwm", Mqtt::CMessage::QOS1));
        // Stratus services
        _Subscriptions.push_back(CClient::Subscription("oblo/" + _szAccount + "/svc/+/gtw/" + m_szGatewayId + "/evt", Mqtt::CMessage::QOS1));
        _Subscriptions.push_back(CClient::Subscription("oblo/" + _szAccount + "/svc/+/gtw/" + m_szGatewayId + "/lwm", Mqtt::CMessage::QOS1));
        _Subscriptions.push_back(CClient::Subscription("oblo/svc/+/evt", Mqtt::CMessage::QOS1));
        /* Bridge only publishes on following topics
        _Topics.push_back(CClient::Subscription("oblo/" + _szAccount + "/cli/+/req", Mqtt::CMessage::QOS0));
        _Topics.push_back(CClient::Subscription("oblo/" + _szAccount + "/cli/+/rsp", Mqtt::CMessage::QOS0));*/
    } else {
        // Stratus services
        _Subscriptions.push_back(CClient::Subscription("oblo/" + _szAccount + "/svc/udm/gtw/" + m_szGatewayId + "/evt", Mqtt::CMessage::QOS1));
        _Subscriptions.push_back(CClient::Subscription("oblo/" + _szAccount + "/svc/udm/gtw/" + m_szGatewayId + "/lwm", Mqtt::CMessage::QOS1));
        _Subscriptions.push_back(CClient::Subscription("oblo/" + _szAccount + "/gtw/" + m_szGatewayId + "/omb/rsp", Mqtt::CMessage::QOS0));
    }
}

Poco::DynamicStruct::Ptr CBridge::CmdSetGlobalAccount(const std::string& _szCmdName, const Poco::DynamicStruct& _Params) {
    Poco::Mutex::ScopedLock lock(m_ModuleMutex);

    if (IsRunning()) {
        std::string account = _Params[PARAM_ACCOUNT].convert<std::string>();

        if (account != m_szAccount) {
            RestartCloudConnection(account, true);
            information("Account changed (a=" + account + ")");
        } else {
            information("Already assigned (a=" + account + ").");
        }
    }

    return NULL;
}

Poco::DynamicStruct::Ptr CBridge::CmdUpdateProperty(const std::string& _szCmdName, const Poco::DynamicStruct& _Params) {
    std::string property = _Params[PARAM_PROPERTY].extract<std::string>();
    Poco::Dynamic::Var value = _Params[PARAM_VALUE];

    if (property == PROP_CLOUD_ADDR) {
        string* pAddr = new string(value.extract<std::string>());
        ASYNC_CALL_P(AsyncUpdateCloudAddr, pAddr);
    } else {
        throw Poco::InvalidArgumentException("Unknown property name.", -1);
    }

    return NULL;
}

Poco::DynamicStruct::Ptr CBridge::CmdGetProperty(const std::string& _szCmdName, const Poco::DynamicStruct& _Params) {
    Poco::DynamicStruct::Ptr ret = new Poco::DynamicStruct();
    std::string property = _Params[PARAM_PROPERTY].extract<std::string>();

    if (property == PROP_CLOUD_ADDR) {
        string addr = SharedValues::CFactory::GetHelper<CConfigCloud>()->GetAddr();
        ret->insert(PARAM_VALUE, addr);
    } else  if (property == PROP_CLOUD_STATUS) {
        int status = GetCloudStatus();
        ret->insert(PARAM_VALUE, status);
    } else {
        throw Poco::InvalidArgumentException("Unknown property name.", -1);
    }

    return ret;
}

Poco::DynamicStruct::Ptr CBridge::CmdSubscribe(const std::string& _szCmdName, const Poco::DynamicStruct& _Params) {
    Poco::Mutex::ScopedLock lock(m_ModuleMutex);

    if (IsRunning()) {
        string svc = _Params[PARAM_SVC].convert<string>();
        string functionality = _Params[PARAM_FUNCTIONALITY].convert<string>();
        string topic = "oblo/svc/" + svc;

        if (!functionality.empty()) {
            topic += "/" + functionality;
        }

        topic += "/evt";
        m_pCloudClient->Subscribe(topic, Mqtt::CMessage::QOS1);
        information("Subscribed (t=" + topic + ", s=)");
        return new Poco::DynamicStruct();
    } else {
        throw Poco::IllegalStateException("Cloud client not initialized");
    }
}

Poco::DynamicStruct::Ptr CBridge::CmdUnsubscribe(const std::string& _szCmdName, const Poco::DynamicStruct& _Params) {
    Poco::Mutex::ScopedLock lock(m_ModuleMutex);

    if (IsRunning()) {
        string svc = _Params[PARAM_SVC].convert<string>();
        string functionality = _Params[PARAM_FUNCTIONALITY].convert<string>();
        string topic = "oblo/svc/" + svc;

        if (!functionality.empty()) {
            topic += "/" + functionality;
        }

        topic += "/evt";
        m_pCloudClient->Unsubscribe(topic);
        information("Unsubscribed (t=" + topic + ", s=)");
        return new Poco::DynamicStruct();
    } else {
        throw Poco::IllegalStateException("Cloud client not initialized");
    }
}

CloudStatus CBridge::GetCloudStatus() const {
    bool conn = IsRunning() && m_pCloudClient && m_pCloudClient->IsConnected();
    bool asgn = IsAssigned();
    CloudStatus status;

    if (conn && asgn) {
        status = STATUS_CONNECTED_ASSIGNED;
    } else if (conn) {
        status = STATUS_CONNECTED_NOT_ASSIGNED;
    } else {
        status = STATUS_DISCONNECTED;
    }

    return status;
}

void CBridge::AsyncUpdateCloudAddr(void* _pParams) {
    string* pAddr = static_cast<string*>(_pParams);
    string addr = *pAddr;
    SAFE_DELETE(pAddr);
    CConfigCloud::Ptr cloudCnfg = SharedValues::CFactory::GetHelper<CConfigCloud>();
    std::string currAddr = cloudCnfg->GetAddr();

    if (addr != currAddr) {
        Poco::Mutex::ScopedLock lock(m_CloudUpdaterMutex);
        ClearCloudUpdater(NULL);
        m_pCloudAddrUpdater = new Async::CUpdateCloudAddress(this, addr);
        m_pCloudAddrUpdaterThread = new Poco::Thread();
        m_pCloudAddrUpdaterThread->start(*m_pCloudAddrUpdater);
        std::stringstream ss;
        ss << "Cloud address update to " << addr << " is in progress.";
        information(ss.str());
    } else {
        std::stringstream ss;
        ss << "Request for cloud address update is dismissed (address " << addr <<
           " is already in use).";
        information(ss.str());
    }
}

void CBridge::OnAddrUpdateResult(const std::string& _szAddr, bool _bConnected) {
    if (_bConnected) {
        CConfigCloud::Ptr cloudCnfg = SharedValues::CFactory::GetHelper<CConfigCloud>();
        cloudCnfg->SetAddr(_szAddr);
        std::stringstream ss;
        ss << "Cloud address successfully UPDATED to: " << _szAddr;
        information(ss.str());
        ASYNC_CALL(ResetAssignment);
    } else {
        std::stringstream ss;
        ss << "Cloud address update FAILED for: " << _szAddr;
        information(ss.str());
    }

    if (!m_pCloudAddrUpdater->IsCanceled()) {
        ASYNC_CALL(ClearCloudUpdater);
    }
}

void CBridge::ClearCloudUpdater(void* _pParams) {
    Poco::Mutex::ScopedLock lock(m_CloudUpdaterMutex);

    if (m_pCloudAddrUpdater && m_pCloudAddrUpdaterThread) {
        debug("Cleaning cloud address updater...");

        if (!m_pCloudAddrUpdater->IsCanceled()) {
            m_pCloudAddrUpdater->Cancel();
            m_pCloudAddrUpdaterThread->join();
        }

        SAFE_DELETE(m_pCloudAddrUpdater);
        SAFE_DELETE(m_pCloudAddrUpdaterThread);
        debug("Cloud address updater canceled.");
    }
}

void CBridge::ResetAssignment(void* _pParams) {
    RestartCloudConnection(DEFAULT_ACCOUNT, true);
}

void CBridge::ResetCleanSession(void* _pParams) {
    RestartCloudConnection(m_szAccount, false);
}

void CBridge::SwitchToCleanSession(void* _pParams) {
    Poco::Mutex::ScopedLock lock(m_ModuleMutex);

    if (m_pCloudClient && (m_szAccount != DEFAULT_ACCOUNT || !m_pCloudClient->IsCleanSessionSet())) {
        int minutes = SharedValues::CFactory::GetHelper<CConfigCloud>()->GetSessionExpiration();
        stringstream ss;
        ss << "Disconnected for more than " << minutes << " min, switching to 'default' with clean session";
        information(ss.str());
        RestartCloudConnection(DEFAULT_ACCOUNT, true);
    }
}

void CBridge::RedirectToCloud(const std::string& _szTopic, const Mqtt::CMessage& _Msg) {
    bool locked = false;

    while (IsRunning() && !locked) {
        locked = m_ModuleMutex.tryLock(1000);
    }

    if (locked) {
        if (IsConnectionUsable() || IsUdmDefaultRequest(_szTopic)) {
            std::string topicCopy(_szTopic);
            Routes::const_iterator itr;

            for (itr = m_Routes.begin(); itr != m_Routes.end(); ++itr) {
                std::string local_prefix = (*itr).first;

                if (topicCopy.compare(0, local_prefix.length(), local_prefix) == 0) {
                    std::string remote_prefix = (*itr).second;
                    topicCopy.replace(0, local_prefix.length(), remote_prefix);

                    try {
                        m_pCloudClient->PublishMsg(topicCopy, _Msg);
                    } catch (const Poco::RuntimeException& e) {
                        error("Unable to process a message from: " + topicCopy);
                    }

                    break;
                }
            }
        } else {
            Queue(_szTopic, _Msg);
        }

        m_ModuleMutex.unlock();
    }
}

void CBridge::RedirectToLocal(const std::string& _szTopic, const Mqtt::CMessage& _Msg) {
    // No need for mutex and m_pLocalClient pointer check. Local client starts first and ends last;
    // it is alive between start/stop and that is time range in which this method can be called.
    std::string topicCopy(_szTopic);
    Routes::const_iterator itr;

    for (itr = m_Routes.begin(); itr != m_Routes.end(); ++itr) {
        std::string remote_prefix = (*itr).second;

        if (topicCopy.compare(0, remote_prefix.length(), remote_prefix) == 0) {
            std::string local_prefix = (*itr).first;
            topicCopy.replace(0, remote_prefix.length(), local_prefix);

            try {
                m_pLocalClient->PublishMsg(topicCopy, _Msg);
            } catch (const Poco::RuntimeException& e) {
                error("Unable to process a message from: " + topicCopy);
            }

            break;
        }
    }
}

void CBridge::Queue(const std::string& _szTopic, const Mqtt::CMessage& _Msg) {
    // protected with module lock
    if (_Msg.GetQos() >= Mqtt::CMessage::QOS1) {
        int size = m_MsgQueue.size();

        if (m_nMsgQueueMax > 0 && size >= m_nMsgQueueMax) {  // if max is less then 1, queue size is not checked.
            m_MsgQueue.pop();
        }

        m_MsgQueue.push(new Mqtt::CRecvHolder(_szTopic, new Mqtt::CMessage(_Msg)));
        std::stringstream ss;
        ss << "Message is queued and waits on assignment (t=" << _szTopic << ", s=" << size << ")";
        debug(ss.str());
    } else {
        std::stringstream ss;
        ss << "Message is dropped (t=" << _szTopic << ")";
        trace(ss.str());
    }
}

void CBridge::SendMessageQueue() {
    // Module mutex is locked before entry
    // Do not redirect to cloud if not assigned
    if (IsConnectionUsable()) {
        // Send messages waiting for the account.
        size_t qsize = m_MsgQueue.size();

        // Cloud client pointer is initialized before this metnod is run, check only if connection is active
        while (m_pCloudClient->IsConnected() && !m_MsgQueue.empty()) {
            Mqtt::CRecvHolder::Ptr msg = m_MsgQueue.front();
            m_MsgQueue.pop();
            // If redirect fails, message will be placed back to the queue
            RedirectToCloud(msg->GetTopic(), *msg->GetMessage());
        }

        size_t remaining = m_MsgQueue.size();

        if (qsize > remaining) {
            // Messages published sucessfully
            stringstream ss;
            ss << "Queued messages published (count=" << qsize << ", qsize=" << remaining << ")";
            information(ss.str());
        }
    }
}

void CBridge::OnConnected(const std::string& _szTag) {
    // Module mutex must NOT be locked in client's callbacks!
    // don't clean next connect
    if (m_pCloudClient->IsCleanSessionSet() && IsAssigned()) {
        information("Cloud session is cleaned, reconnecting with clean session=false");
        ASYNC_CALL(ResetCleanSession);
    } else {
        SendEvent(new Event::CCloudStatusChanged(GetCloudStatus()));

        if (!m_pCloudClient->IsSessionPressent()) {
            SendEvent(new Event::CCloudSessionCleaned());
            stringstream ss;
            ss << "Cloud session is not present, clean session started (a=" << m_szAccount << ")";
            information(ss.str());
        }
    }
}

void CBridge::OnConnectFailed(const std::string& _szTag, int16_t _nCode, const std::string& _szReason) {
    // Check if session should be cleaned
    int minutes = SharedValues::CFactory::GetHelper<CConfigCloud>()->GetSessionExpiration();
    Poco::Timestamp::TimeDiff diff = minutes * 60000000;

    if (m_tDisconnectedSince.isElapsed(diff)) {
        ASYNC_CALL(SwitchToCleanSession);
    }
}

void CBridge::OnConnectionLost(const std::string& _szTag, const std::string& _szReason) {
    m_tDisconnectedSince.update();
    SendEvent(new Event::CCloudStatusChanged(STATUS_DISCONNECTED));
}

} // Bridge
} // MessageBroker
