/****************************************************************************
 *  Copyright (c) 2018 OBLO Living LLC
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.obloliving.com
 *  -----------------------------------------------------
 *
 * \project Oblo Message Broker - Broker application.
 * \file bridge/Config.cpp
 * \brief
 * OBLO bridge configuration.
 *
 * @Author Milan Tucic
 *
 * \notes
 *
 * \history
 * 28.8.2018 moved from Bridge.cpp
 *****************************************************************************/
#include "Config.h"
#include <core/Globals.h>
#include <core/Database.h>
#include <Poco/SHA1Engine.h>

namespace MessageBroker {
namespace Bridge {

using namespace std;

const string CConfigLocal::KEY_BASE = "bridge.local.";

CConfigLocal::CConfigLocal(): SharedValues::CHelper(KEY_BASE) {
}

std::string CConfigLocal::GetId() const {
    return m_pConfigValues->getString(m_szKeyBase + "id", "stratus");
}

std::string CConfigLocal::GetUsername() const {
    return m_pConfigValues->getString(m_szKeyBase + "username", "stratus");
}

std::string CConfigLocal::GetPassword() const {
    return m_pConfigValues->getString(m_szKeyBase + "password", "");
}

std::string CConfigLocal::GetAddr() const {
    return m_pConfigValues->getString(m_szKeyBase + "addr", "tcp://localhost:1883");
}

int CConfigLocal::GetReconnectPeriod() const {
    return m_pConfigValues->getInt(m_szKeyBase + "reconnection_period", 5);
}

int CConfigLocal::GetKeepAliveInterval() const {
    return m_pConfigValues->getInt(m_szKeyBase + "keepalive", 60);
}

int CConfigLocal::GetTimeoutInterval() const {
    return m_pConfigValues->getInt(m_szKeyBase + "timeout_interval", 5);
}

const string CConfigCloud::KEY_BASE = "bridge.cloud.";

CConfigCloud::CConfigCloud(): SharedValues::CHelper(KEY_BASE) {
}

std::string CConfigCloud::GetId() const {
    SharedValues::CGateway::Ptr gtw =
        SharedValues::CFactory::GetHelper<SharedValues::CGateway>();
    return std::string("g-") + gtw->GetId();
}

std::string CConfigCloud::GetUsername() const {
    SharedValues::CGateway::Ptr gtw =
        SharedValues::CFactory::GetHelper<SharedValues::CGateway>();
    return gtw->GetId();
}

std::string CConfigCloud::GetPassword() const {
    SharedValues::CGateway::Ptr gtw =
        SharedValues::CFactory::GetHelper<SharedValues::CGateway>();
    std::string serial = gtw->GetId();
    std::string security = gtw->GetSecurity();
    Poco::SHA1Engine engine;
    engine.update(serial + ":" + security);
    std::string password = std::string(Poco::DigestEngine::digestToHex(
                                           engine.digest()));
    return password;
}

std::string CConfigCloud::GetAddr() const {
    string v = Database::CDatabase::GetTable<Database::CParamsTable>()->GetParam(
                   m_szKeyBase + "addr", "");

    if (v.empty()) {
        v = m_pConfigValues->getString(m_szKeyBase +
                                       "addr"); // throws exception if not found.
    }

    return v;
}

void CConfigCloud::SetAddr(const std::string& _szAddr) {
    Database::CDatabase::GetTable<Database::CParamsTable>()->SetParam(
                m_szKeyBase + "addr", _szAddr);
}

int CConfigCloud::GetReconnectPeriodMinimum() const {
    return m_pConfigValues->getInt(m_szKeyBase + "reconnection_period_min", 15);
}

int CConfigCloud::GetReconnectPeriodMaximum() const {
    return m_pConfigValues->getInt(m_szKeyBase + "reconnection_period_max", 180);
}

int CConfigCloud::GetKeepAliveInterval() const {
    return m_pConfigValues->getInt(m_szKeyBase + "keepalive", 60);
}

int CConfigCloud::GetTimeoutInterval() const {
    return m_pConfigValues->getInt(m_szKeyBase + "timeout_interval", 10);
}

std::string CConfigCloud::GetCaFile() const {
    return m_pConfigValues->getString(m_szKeyBase + "ca_file", "");
}

int CConfigCloud::GetMsgQueueLimit() const {
    return m_pConfigValues->getInt(m_szKeyBase + "queue_limit", 0);
}

int CConfigCloud::GetSessionExpiration() const {
    return m_pConfigValues->getInt(m_szKeyBase + "session_expiration", 30);
}

int CConfigCloud::GetDnsCacheLimit() const {
    return m_pConfigValues->getInt(m_szKeyBase + "dns_cache_limit", 0);
}

bool CConfigCloud::IsStatusForced() const {
    return m_pConfigValues->getBool(m_szKeyBase + "is_status_forced", false);
}

} // Bridge
} // MessageBroker
