/****************************************************************************
 *  Copyright (c) 2016 RT-RK d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.rt-rk.com
 *  -----------------------------------------------------
 *
 * \project Oblo Message Broker - Broker application.
 * \file Link.h
 * \brief
 * OMB linker configuration helper.
 *
 * @Author Milan Tucic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#include "feature/linker/Discovery.h"
#include "core/Globals.h"

#include "Poco/Net/Socket.h"
#include "Poco/Net/MulticastSocket.h"
#include "Poco/Net/HTTPClientSession.h"
#include "Poco/Net/HTTPRequest.h"
#include "Poco/Net/HTTPResponse.h"
#include "Poco/Net/HTTPMessage.h"
#include "Poco/NumberParser.h"
#include "Poco/StreamCopier.h"
#include "Poco/RegularExpression.h"
#include "Poco/DOM/DOMParser.h"
#include "Poco/DOM/Document.h"
#include "Poco/DOM/NodeIterator.h"
#include "Poco/DOM/NodeFilter.h"
#include "Poco/DOM/AutoPtr.h"

#include <sstream>
#include <iostream>

namespace MessageBroker {
namespace Linker {
namespace Discovery {

#define MAX_SSDP_BUFFER_LENGHT 2048
#define SSDP_PAUSE 100
#define SSDP_TIME_TO_LIVE 4
#define SSDP_SEND_TIMEOUT 3
#define SSDP_RECV_TIMEOUT 3

void CSsdp::Search(const string& _szST, CGtwInfo::Devices& _Devices) {
    Poco::Net::SocketAddress sa(Poco::Net::IPAddress(), SSDP_LISTENER_PORT);
    Poco::Net::SocketAddress group(SSDP_IP, SSDP_PORT);
    Poco::Net::MulticastSocket socket(sa, true);
    socket.setLoopback(true);
    socket.setTimeToLive(SSDP_TIME_TO_LIVE);
    socket.setReceiveTimeout(Poco::Timespan(SSDP_RECV_TIMEOUT, 0));
    socket.setSendTimeout(Poco::Timespan(SSDP_SEND_TIMEOUT, 0));
    // Prepare M-SEARCH message
    std::stringstream ss;
    ss << "M-SEARCH * HTTP/1.1" << std::endl
       << "Host: " << SSDP_IP << ":" << SSDP_PORT << std::endl
       << "Man: \"ssdp:discover\"" << std::endl
       << "MX: 3" << std::endl
       << "ST: " << _szST << std::endl;
    string msearch = ss.str();

    // Send M-SEARCH multicast
    try {
        socket.sendTo(msearch.c_str(), msearch.length(), group);
        //Poco::Thread::sleep(SSDP_PAUSE);
        //socket.sendTo(msearch.c_str(), msearch.length(), group);
    } catch (const Poco::Exception& e) {
    }

    // Get results
    try {
        char buffer[MAX_SSDP_BUFFER_LENGHT] = {};
        Poco::Net::SocketAddress sender;

        while ( socket.receiveFrom(buffer, MAX_SSDP_BUFFER_LENGHT-1, sender) ) {
            CGtwInfo::Ptr p = NULL;
            string location = GetLocation(buffer);
            string data = GetDeviceDesc(location);
            p = CGtwInfo::Parse(data, sender.host().toString());
            _Devices.push_back(p);
            memset(buffer, 0, MAX_SSDP_BUFFER_LENGHT*sizeof(char));
        }
    } catch (const Poco::Exception& e) {
    }

    socket.close();
}

string CSsdp::GetLocation(const string& _szSsdpMsg) {
    std::istringstream ss(_szSsdpMsg);
    string line, location;

    while ( std::getline(ss, line) ) {
        if ( line.compare(0, 10, "LOCATION: ")==0 ) {
            location = line.substr(10, line.length());

            if ( !location.empty() && location[location.length()-1]=='\r' ) {
                location.erase(location.length()-1);
            }

            break;
        }
    }

    return location;
}

string CSsdp::GetDeviceDesc(const string& _szUrl) {
    Poco::RegularExpression regex("https?:\\/\\/(.*):(\\d\\d\\d?\\d?\\d?)(.*)");
    string szHost = _szUrl;
    regex.subst(szHost, "$1");
    string szPort = _szUrl;
    regex.subst(szPort, "$2");
    string szUri = _szUrl;
    regex.subst(szUri, "$3");
    uint16_t nPort = Poco::NumberParser::parse(szPort);
    Poco::Net::HTTPClientSession session(szHost, nPort);
    Poco::Net::HTTPRequest request(Poco::Net::HTTPRequest::HTTP_GET, szUri,
                                   Poco::Net::HTTPMessage::HTTP_1_1);
    Poco::Net::HTTPResponse response;
    session.sendRequest(request);
    std::istream& rs = session.receiveResponse(response);
    string data;

    if ( response.getStatus()==Poco::Net::HTTPResponse::HTTP_OK ) {
        stringstream ss;
        Poco::StreamCopier::copyStream(rs, ss);
        data = ss.str();
    } else {
        throw Poco::Exception("HTTP status different then OK");
    }

    return data;
}

CGtwInfo::Ptr CGtwInfo::Parse(const string& _szXmlData, const string& _szSender) {
    CGtwInfo::Ptr p = new CGtwInfo();
    Poco::XML::DOMParser parser;
    parser.setFeature(Poco::XML::DOMParser::FEATURE_FILTER_WHITESPACE, true);
    Poco::AutoPtr<Poco::XML::Document> pDoc = parser.parseString(_szXmlData);
    Poco::XML::NodeIterator it(pDoc, Poco::XML::NodeFilter::SHOW_ALL);
    Poco::XML::Node* pNode = it.nextNode();
    string prevNodeName;

    while (pNode) {
        if ( prevNodeName=="deviceType" ) {
            p->m_szDeviceType = pNode->nodeValue();
        } else if ( prevNodeName=="manufacturer" ) {
            p->m_szManufacturer = pNode->nodeValue();
        } else if ( prevNodeName=="modelName" ) {
            p->m_szModelName = pNode->nodeValue();
        } else if ( prevNodeName=="modelNumber" ) {
            p->m_szModelNumber = pNode->nodeValue();
        } else if ( prevNodeName=="serialNumber" ) {
            p->m_szSerialNumber = pNode->nodeValue();
        }

        prevNodeName = pNode->nodeName();
        pNode = it.nextNode();
    }

    p->m_szIp = _szSender;
    return p;
}

string CGtwInfo::GetDeviceType() const {
    return m_szDeviceType;
}

string CGtwInfo::GetManufacturer() const {
    return m_szManufacturer;
}

string CGtwInfo::GetModelName() const {
    return m_szModelName;
}

string CGtwInfo::GetModelNumber() const {
    return m_szModelNumber;
}

string CGtwInfo::GetSerialNumber() const {
    return m_szSerialNumber;
}

string CGtwInfo::GetIp() const {
    return m_szIp;
}

string CGtwInfo::ToString() const {
    stringstream ss;
    ss << "id=" << m_szSerialNumber
       << ", ip=" << m_szIp
       << ", man=" << m_szManufacturer
       << ", model=" << m_szModelName
       << ", mnum=" << m_szModelNumber;
    return ss.str();
}

CRunner::CRunner(IDeviceObserver* _pObserver, const uint32_t _nPeriod):
    m_pObserver(_pObserver), m_nPeriod(_nPeriod), m_bRunning(true) {
    setLoggerName(Linker::MODULE_ID);
}

void CRunner::run() {
    if ( m_bRunning ) {
        m_Timer.setStartInterval(m_nPeriod);
        m_Timer.setPeriodicInterval(m_nPeriod);
        m_Timer.start(Poco::TimerCallback<CRunner>(*this, &CRunner::OnTimer));
    }

    while ( m_bRunning ) {
        try {
            CGtwInfo::Devices devs;
            CSsdp::Search(SSDP_ST, devs);
            CGtwInfo::Devices::const_iterator itr;

            for ( itr=devs.begin(); m_bRunning && itr!=devs.end(); ++itr ) {
                m_pObserver->OnDeviceLocated(*itr);
            }
        } catch (const Poco::Exception& e) {
            error("Search failed" + e.message());
        }

        m_SyncObj.wait();
    }
}

void CRunner::Trigger() {
    m_SyncObj.notify();
}

void CRunner::Stop() {
    m_bRunning = false;
    m_Timer.stop();
    m_SyncObj.notify();
}

void CRunner::OnTimer(Poco::Timer& timer) {
    m_SyncObj.notify();
}


CManager::CManager(IDeviceObserver* _pDevObserver):
    m_pDevObserver(_pDevObserver),
    m_pDiscoveryRunnable(NULL),
    m_pDiscoveryThread(NULL),
    m_nPeriod(0),
    m_bRunning(false) {
}

CManager::~CManager() {
    Stop();
}

void CManager::SetPeriod(const uint32_t _nPeriod) {
    Poco::Mutex::ScopedLock lock(m_DiscoveryMutex);
    m_nPeriod = _nPeriod*1000;
}

void CManager::Start() {
    Poco::Mutex::ScopedLock lock(m_DiscoveryMutex);

    if ( !m_bRunning ) { // Discovery is not running
        m_pDiscoveryRunnable = new CRunner(m_pDevObserver, m_nPeriod);
        m_pDiscoveryThread = new Poco::Thread();
        m_pDiscoveryThread->start(*m_pDiscoveryRunnable);
        m_bRunning = true;
    }
}

void CManager::Stop() {
    Poco::Mutex::ScopedLock lock(m_DiscoveryMutex);

    if ( m_bRunning ) {
        m_bRunning = false;
        m_pDiscoveryRunnable->Stop();
        m_pDiscoveryThread->join(); // Wait thread to finish.
        SAFE_DELETE(m_pDiscoveryRunnable);
        SAFE_DELETE(m_pDiscoveryThread);
    }
}

bool CManager::IsRunning() const {
    return m_bRunning;
}

CCleaner::CCleaner():
    m_CleaningTimer(1000, 1000) { // 1 s period
    m_CleaningTimer.start(Poco::TimerCallback<CCleaner>(*this, &CCleaner::OnCleaningTime));
}

CCleaner::CCleaner(const CCleaner& _Cleaner):
    m_CleaningList(_Cleaner.m_CleaningList),
    m_CleaningTimer(1000, 1000) { // 1 s period
    m_CleaningTimer.start(Poco::TimerCallback<CCleaner>(*this, &CCleaner::OnCleaningTime));
}

CCleaner::~CCleaner() {
    StopCleaning();
}


void CCleaner::StopCleaning() {
    m_CleaningTimer.stop();
    CleanAll();
}

void CCleaner::ScheduleCleaning(const string& _szGtwId, int16_t _nTimeout) {
    Poco::Mutex::ScopedLock lock(m_CleaningMutex);
    List::iterator itr;
    int16_t diff = _nTimeout;
    int16_t prevDiff = diff;

    for ( itr=m_CleaningList.begin(); itr!=m_CleaningList.end(); ++itr) {
        ListElm& tmpElm = *itr;
        diff -= tmpElm.second;

        if ( diff <= 0 ) {
            tmpElm.second = abs(diff);
            m_CleaningList.insert(itr, ListElm(_szGtwId, prevDiff));
            break;
        }

        prevDiff = diff;
    }

    // If list is empty, or element is not insert inside the loop it will be processed here.
    if ( itr==m_CleaningList.end() ) {
        m_CleaningList.push_back(ListElm(_szGtwId, prevDiff));
    }
}

void CCleaner::OnCleaningTime(Poco::Timer& timer) {
    Poco::Mutex::ScopedLock lock(m_CleaningMutex);
    List::iterator itr;
    int16_t diff = 1 + timer.skipped();

    for ( itr=m_CleaningList.begin(); itr!=m_CleaningList.end(); ++itr) {
        ListElm& elm = *itr;
        elm.second = elm.second-diff;

        if ( elm.second < 0 ) {
            OnCleanDiscovery(elm.first);
        } else {
            break;
        }
    }

    m_CleaningList.erase(m_CleaningList.begin(), itr);
}

void CCleaner::CleanAll() {
    Poco::Mutex::ScopedLock lock(m_CleaningMutex);
    List::iterator itr;

    for ( itr=m_CleaningList.begin(); itr!=m_CleaningList.end(); ++itr) {
        ListElm& elm = *itr;
        OnCleanDiscovery(elm.first);
    }
}

} // Discovery
} // Linker
} // MessageBroker
