/****************************************************************************
 *  Copyright (c) 2016 RT-RK d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.rt-rk.com
 *  -----------------------------------------------------
 *
 * \project Oblo Message Broker - Broker application.
 * \file Liker/Discovery.h
 * \brief
 * OMB linker discovery.
 *
 * @Author Milan Tucic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef OMB_LINKER_DISCOVERY_H
#define OMB_LINKER_DISCOVERY_H

#include "core/SharedValues.h"
#include "core/Logger.h"
#include "Poco/SharedPtr.h"
#include "Poco/Timer.h"
#include "Poco/SynchronizedObject.h"
#include <list>

namespace MessageBroker {
namespace Linker {
class CLinker;
namespace Discovery {

using namespace std;

class CGtwInfo {
public:
    typedef Poco::SharedPtr<CGtwInfo> Ptr;
    typedef std::list<CGtwInfo::Ptr> Devices;

    CGtwInfo() {}

    static CGtwInfo::Ptr Parse(const string& _szXmlData, const string& _szSender);

    string GetDeviceType() const;
    string GetManufacturer() const;
    string GetModelName() const;
    string GetModelNumber() const;
    string GetSerialNumber() const;
    string GetIp() const;
    string ToString() const;
private:
    string m_szDeviceType;
    string m_szManufacturer;
    string m_szModelName;
    string m_szModelNumber;
    string m_szSerialNumber;
    string m_szIp;
};

class CSsdp {
public:
    static void Search(const string& _szST, CGtwInfo::Devices& _Devices);
private:
    static string GetLocation(const string& _szSsdpMsg);
    static string GetDeviceDesc(const string& _szUrl);
};

class IDeviceObserver {
public:
    virtual void OnDeviceLocated(CGtwInfo::Ptr _pDevice)=0;
};

class CRunner : public Poco::Runnable, private Logger::CLogger {
public:
    CRunner(IDeviceObserver* _pObserver, const uint32_t _nPeriod);
    void run();
    void Trigger();
    void Stop();
    void OnTimer(Poco::Timer& timer);

protected:
    IDeviceObserver* m_pObserver;
    Poco::Timer m_Timer;
    unsigned int m_nPeriod;
    Poco::SynchronizedObject m_SyncObj;
    bool m_bRunning;
};

class CManager {
public:
    CManager(IDeviceObserver* _pDevObserver);
    virtual ~CManager();

    void SetPeriod(const uint32_t _nPeriod);
    void Start();
    void Stop();
    bool IsRunning() const;

private:
    Poco::Mutex m_DiscoveryMutex;
    IDeviceObserver* m_pDevObserver;
    CRunner* m_pDiscoveryRunnable;
    Poco::Thread* m_pDiscoveryThread;
    uint32_t m_nPeriod;
    bool m_bRunning;
};

class CResultHandler {
public:
    typedef void (CLinker::*DiscCompletedFuncPtr)(const std::string& _szGtwId, CGtwInfo::Ptr _pDevInfo, void* _pData);

    CResultHandler(const std::string& _szGtwId, CLinker* _pObj, DiscCompletedFuncPtr _pFunc, void* _pData):
        m_szGtwId(_szGtwId), m_pObj(_pObj), m_pFunc(_pFunc), m_pData(_pData), m_bCompleted(false) {}

    std::string GetGwtId() const {
        return m_szGtwId;
    }
    bool IsCompleted() const {
        return m_bCompleted;
    }
    void Complete(CGtwInfo::Ptr _pDevInfo) {
        m_bCompleted = true;
        (m_pObj->*m_pFunc)(m_szGtwId, _pDevInfo, m_pData);
    }

private:
    std::string m_szGtwId;
    CLinker* m_pObj;
    DiscCompletedFuncPtr m_pFunc;
    void* m_pData;
    bool m_bCompleted;
};

class CCleaner {
protected:
    CCleaner();
    CCleaner(const CCleaner& _Cleaner);
    virtual ~CCleaner();
    void ScheduleCleaning(const std::string& _szGtwId, int16_t _nTimeout);
    virtual void OnCleanDiscovery(const std::string& _szGtwId)=0;
private:
    void StopCleaning();
    void CleanAll();
    void OnCleaningTime(Poco::Timer& timer);

    typedef std::pair<std::string, int16_t> ListElm;
    typedef std::list<ListElm> List;
    List m_CleaningList;
    Poco::Mutex m_CleaningMutex;
    Poco::Timer m_CleaningTimer;
};

} // Discovery
} // Linker
} // MessageBroker

#endif
