/****************************************************************************
 *  Copyright (c) 2016 RT-RK d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.rt-rk.com
 *  -----------------------------------------------------
 *
 * \project Oblo Message Broker - Broker application.
 * \file Link.cpp
 * \brief
 * OMB link object implementation.
 *
 * @Author Milan Tucic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#include "feature/linker/Link.h"
#include "core/Globals.h"
#include "feature/linker/Config.h"

#define MAKE_MSG_DELEGATE(m) \
    Mqtt::CMsgDelegate::CreateDelegate<CLink, &CLink::m>(this)

namespace MessageBroker {
namespace Linker {

const string CLink::KEY_ID = "id";
const string CLink::KEY_STATE = "state";

CLink::CLink(const string& _szId, const string& _szSecId):
    m_szId(_szId),
    m_szSecId(_szSecId),
    m_bRunning(false),
    m_State(LINK_IDLE),
    m_pLocalClient(NULL),
    m_pRemoteClient(NULL),
    m_pListener(NULL) {
    setLoggerName(MODULE_ID);
}

CLink::~CLink() {
    Stop();
}

void CLink::Start() {
    trace(m_szId + " starting..");
    Poco::Mutex::ScopedLock lock(m_Mutex);

    if ( !m_bRunning ) {
        SetState(LINK_IDLE);
        StartRemoteClient();
        StartLocalClient();

        if ( m_State==LINK_IDLE ) { // if connect not failed.
            SetState(LINK_RUNNING);
        }

        m_bRunning = true;
        information(m_szId + " started.");
    }
}

void CLink::Stop() {
    trace(m_szId + " stopping..");
    Poco::Mutex::ScopedLock lock(m_Mutex);

    if ( m_bRunning ) {
        m_bRunning = false;
        m_pLocalClient->Stop();
        SAFE_DELETE(m_pLocalClient);
        m_pRemoteClient->Stop();
        SAFE_DELETE(m_pRemoteClient);
        SetState(LINK_IDLE);
        information(m_szId + " stopped.");
    }
}

string CLink::GetId() const {
    return m_szId;
}

string CLink::GetSecId() const {
    return m_szSecId;
}

string CLink::GetUrl() const {
    return m_szUrl;
}

void CLink::SetUrl(const string& _szUrl) {
    m_szUrl = _szUrl;
}

void CLink::SetUsername(const string& _szUsername) {
    m_szUsername = _szUsername;
}

void CLink::SetPassword(const string& _szPassword) {
    m_szPassword = _szPassword;
}

uint16_t CLink::GetRegFailureCnt() const {
    return m_nRegFailCnt;
}

void CLink::SetListener(ILinkListener* _pL) {
    m_pListener = _pL;
}

void CLink::SetState(CLink::STATE _State) {
    m_State = _State;

    if ( m_State==LINK_NOT_REGISTERED ) {
        ++m_nRegFailCnt;
    }

    m_pListener ? m_pListener->OnStateChanged(m_szId, m_State) : void();
}

Poco::DynamicStruct::Ptr CLink::ToStruct() const {
    Poco::DynamicStruct::Ptr p = new Poco::DynamicStruct();
    p->insert(KEY_ID, m_szId);
    p->insert(KEY_STATE, static_cast<int>(m_State));
    return p;
}

void CLink::StartLocalClient() {
    trace(m_szId + " starting local client...");
    // Configure local client
    CConfig::Ptr conf = SharedValues::CFactory::GetHelper<CConfig>();
    string cliId = "lnk/" + m_szId;
    m_pLocalClient = new CClient(cliId, conf->GetLocalAddr(), MODULE_ID);
    m_pLocalClient->SetUsername(cliId);
    m_pLocalClient->SetPassword("");
    m_pLocalClient->RegisterObserver(MAKE_MSG_DELEGATE(RedirectToRemote));
    // Subscription
    Mqtt::CClientImpl::TopicCollection topics;
    Mqtt::CClientImpl::QosCollection qoss;
    topics.push_back("gtw/" + m_szId + "/+/req");
    qoss.push_back(Mqtt::CMessage::QOS1);
    topics.push_back("gtw/" + m_szId + "/+/rsp");
    qoss.push_back(Mqtt::CMessage::QOS1);
    // TODO add for clients and services
    m_pLocalClient->AddSubscriptions(topics, qoss);
    m_pLocalClient->Start();
    trace(m_szId + " local client started...");
}

void CLink::StartRemoteClient() {
    trace(m_szId + " starting remote client...");
    // Configure remote client
    CConfig::Ptr conf = SharedValues::CFactory::GetHelper<CConfig>();
    string gtwId = SharedValues::CFactory::GetHelper<SharedValues::CGateway>()->GetId();
    m_pRemoteClient = new CClient(m_szUsername, m_szUrl, MODULE_ID);
    m_pRemoteClient->SetUsername(m_szUsername);
    m_pRemoteClient->SetPassword(m_szPassword);
#ifdef USE_SECURE_CONNECTION
    m_pRemoteClient->SetSecureConfiguration(conf->GetRemoteCaPath());
#endif
    m_pRemoteClient->RegisterObserver(MAKE_MSG_DELEGATE(RedirectToLocal));
    // Subscription
    Mqtt::CClientImpl::TopicCollection topics;
    Mqtt::CClientImpl::QosCollection qoss;
    topics.push_back("gtw/" + m_szId + "/+/evt");
    qoss.push_back(Mqtt::CMessage::QOS1);
    topics.push_back("gtw/" + m_szId + "/+/lwm");
    qoss.push_back(Mqtt::CMessage::QOS1);
    topics.push_back("gtw/" + gtwId + "/+/rsp");
    qoss.push_back(Mqtt::CMessage::QOS1);
    // TODO add for clients and services
    m_pRemoteClient->AddSubscriptions(topics, qoss);
    m_pRemoteClient->DeactivateResurector();
    m_pRemoteClient->SetConnectionListener(this); // receive connection lost on remote
    m_pRemoteClient->Start();
    trace(m_szId + " remote client started...");
}

void CLink::OnConnected(const string& _szCliId) {
    m_nRegFailCnt = 0; // Reset counter.
}

#define MQTT_NOT_AUTHORIZED 5

void CLink::OnConnectFailed(const string& _szCliId, int16_t _nCode, const string& _szReason) { // for remote client
    switch (_nCode) {
        case MQTT_NOT_AUTHORIZED:
            SetState(LINK_NOT_REGISTERED);
            break;

        default:
            SetState(LINK_OFFLINE);
            break;
    }
}

void CLink::OnConnectionLost(const string& _szCliId, const string& _szCause) { // for remote client
    SetState(LINK_OFFLINE);
}

void CLink::RedirectToRemote(const string& _szTopic, const Mqtt::CMessage& _Msg) {
    std::stringstream ss;
    ss << "(" << m_szId << ") " << "Redirecting message (d=L, t=" << _szTopic << ", p=" << _Msg.GetPayload();
    debug(ss.str());
    Poco::Mutex::ScopedLock lock(m_Mutex);
    m_pRemoteClient->PublishMsg(_szTopic, _Msg);
}

void CLink::RedirectToLocal(const string& _szTopic, const Mqtt::CMessage& _Msg) {
    std::stringstream ss;
    ss << "(" << m_szId << ") "  << "Redirecting message (d=R, t=" << _szTopic << ", p=" << _Msg.GetPayload();
    debug(ss.str());
    Poco::Mutex::ScopedLock lock(m_Mutex);
    m_pLocalClient->PublishMsg(_szTopic, _Msg);
}

} // Linker
} // MessageBroker
