/****************************************************************************
 *  Copyright (c) 2016 RT-RK d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.rt-rk.com
 *  -----------------------------------------------------
 *
 * \project Oblo Message Broker - Broker application.
 * \file Linker.h
 * \brief
 * OMB module used to manage links to other OMB's.
 *
 * @Author Milan Tucic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#include "feature/linker/Linker.h"
#include "feature/linker/Discovery.h"
#include "core/Globals.h"

namespace MessageBroker {
namespace Linker {

#define ASYNC_CALL(f) Run(new Async::CAsyncFunc(&CLinker::f))
#define ASYNC_CALL_P(f, p) Run(new Async::CAsyncFunc(&CLinker::f, p))

CLinker::CLinker():
    Async::CAsyncFuncHandler(this),
    m_DiscoveryMngr(this),
    m_pLinksTable(NULL) {
    setLoggerName(Linker::MODULE_ID);
    AddApiCommandExecutor(Commands::CAddLink::NAME, MAKE_CMD_EXEC_DELEGATE(CLinker, AddLinkHandler));
    AddApiCommandExecutor(Commands::CRemoveLink::NAME, MAKE_CMD_EXEC_DELEGATE(CLinker, RemoveLinkHandler));
    AddCommandExecutor(CMD_GET_PROP, MAKE_CMD_EXEC_DELEGATE(CLinker, GetProp));
}

CLinker::~CLinker() {
    Stop();
}

void CLinker::Start() {
    trace("Starting...");
    Poco::Mutex::ScopedLock lock(m_ModuleMutex);

    if ( !IsRunning() ) {
        m_pConf = SharedValues::CFactory::GetHelper<CConfig>();
        m_pLinksTable = Database::CDatabase::GetTable<CLinksTable>();
        uint32_t discPeriod = m_pConf->GetDiscoveryPeriod();
        m_DiscoveryMngr.SetPeriod(discPeriod);
        SharedValues::CGateway::Ptr gtw = SharedValues::CFactory::GetHelper<SharedValues::CGateway>();
        m_szGtwId = gtw->GetId();
        LoadLinks();
        StartLinks();
        CAppModule::Start();
        information("Started.");
    }
}

void CLinker::Stop() {
    trace("Stoping...");
    Poco::Mutex::ScopedLock lock(m_ModuleMutex);

    if ( IsRunning() ) {
        StopLinks();
        ClearLinks();
        CAppModule::Stop();
        information("Stoped.");
    }
}

void CLinker::GetDbTables(Database::CDatabase::Tables& _Tables) const {
    _Tables.push_back(new CLinksTable());
}

void CLinker::LoadLinks() {
    std::list<CLinkRow> links;
    m_pLinksTable->LoadLinks(links);
    Poco::Mutex::ScopedLock lock(m_ModuleMutex);
    std::list<CLinkRow>::const_iterator itr;

    for ( itr=links.begin(); itr!=links.end(); ++itr ) {
        CLink* p = new CLink(itr->GetGtwId(), itr->GetSecGtwId());
        p->SetUrl(itr->GetUrl());
        p->SetUsername(itr->GetUsername());
        p->SetPassword(itr->GetPassword());
        p->SetListener(this);
        m_Links.insert(LinkElm(p->GetId(), p));
    }
}

void CLinker::StartLinks() {
    Poco::Mutex::ScopedLock lock(m_ModuleMutex);
    Links::iterator itr;

    for ( itr=m_Links.begin(); itr!=m_Links.end(); ++itr ) {
        CLink* p = itr->second;
        p->Start();
    }
}

void CLinker::ClearLinks() {
    Poco::Mutex::ScopedLock lock(m_ModuleMutex);
    // Clear links
    Links::iterator itr;

    for ( itr=m_Links.begin(); itr!=m_Links.end(); ++itr) {
        CLink* p = itr->second;
        SAFE_DELETE(p);
    }

    m_Links.clear();
}

void CLinker::StopLinks() {
    Poco::Mutex::ScopedLock lock(m_ModuleMutex);
    Links::iterator itr;

    for ( itr=m_Links.begin(); itr!=m_Links.end(); ++itr ) {
        CLink* p = itr->second;
        p->Stop();
    }
}

Poco::DynamicStruct::Ptr CLinker::AddLinkHandler(const std::string& _szCmdName, const Poco::DynamicStruct& _Params) {
    trace("Add link requested.");
    Commands::CAddLink* cmd = Commands::CAddLink::Parse(_Params);
    string id = cmd->GetGtwId();
    Poco::Mutex::ScopedLock lock(m_ModuleMutex);

    if ( id==m_szGtwId ) {
        throw Poco::InvalidArgumentException("Can not add link to itself.", -1);
    }

    if ( m_Links.size()>=m_pConf->GetMaxLinks() ) {
        throw Poco::RangeException("Maximum number of links exceeded.", -1);
    }

    if ( m_Links.find(id)!=m_Links.end() ) {
        throw Poco::ExistsException("Link already exists", -1);
    }

    RunDiscovery(id, &CLinker::RegistrationDiscoveryCompleted, true, new string(cmd->GetGtwSec()));
    return new Poco::DynamicStruct();
}

void CLinker::ChangeLinkState(const string& _szId, CLink::STATE _State) {
    Poco::Mutex::ScopedLock lock(m_ModuleMutex);
    Links::iterator itr = m_Links.find(_szId);

    if ( itr!=m_Links.end() ) {
        CLink* p = itr->second;
        p->SetState(_State);
    } else {
        SendEvent(new Events::CLinkStateChanged(_szId, _State));
    }
}

void CLinker::RunDiscovery(const string& _szGtwId, Discovery::CResultHandler::DiscCompletedFuncPtr _pFunc, bool _bWithTimeout, void* _pData) {
    Poco::Mutex::ScopedLock lock(m_DiscMutex);

    if ( m_DiscReqs.size()==0 ) {
        m_DiscoveryMngr.Start();
        debug("Discovery started.");
    }

    DiscReqs::const_iterator itr = m_DiscReqs.find(_szGtwId);

    if ( itr==m_DiscReqs.end() ) {
        Discovery::CResultHandler* p = new Discovery::CResultHandler(_szGtwId, this, _pFunc, _pData);
        m_DiscReqs.insert(DiscReq(_szGtwId, p));

        if ( _bWithTimeout ) {
            ScheduleCleaning(_szGtwId, m_pConf->GetDiscoveryTimeout());
        }

        ChangeLinkState(_szGtwId, CLink::LINK_DISCOVERY_STARTED);
    }
}

void CLinker::OnDeviceLocated(Discovery::CGtwInfo::Ptr _pDevice) {
    debug(_pDevice->ToString());

    if ( !m_DiscMutex.tryLock(500) && !m_DiscoveryMngr.IsRunning() ) {
        return; // discovery is stopped.
    }

    string id = _pDevice->GetSerialNumber();
    DiscReqs::iterator itr = m_DiscReqs.find(id);

    if ( itr!=m_DiscReqs.end() ) {
        // Device is located
        Discovery::CResultHandler* p = itr->second;

        if ( !p->IsCompleted() ) {
            p->Complete(_pDevice);
            ASYNC_CALL_P(CleanDiscovery, new string(p->GetGwtId()));
        }
    }

    m_DiscMutex.unlock();
}

void CLinker::CleanDiscovery(void* _pGtwId) {
    trace("Cleaning discovery...");
    Poco::Mutex::ScopedLock lock(m_DiscMutex);
    Poco::SharedPtr<string> gtwId = static_cast<string*>(_pGtwId);
    DiscReqs::iterator itr = m_DiscReqs.find(*gtwId);

    if ( itr!=m_DiscReqs.end() ) {
        Discovery::CResultHandler* p = itr->second;
        SAFE_DELETE(p);
        m_DiscReqs.erase(itr);
    }

    if ( m_DiscReqs.size()==0 ) {
        m_DiscoveryMngr.Stop();
        debug("Discovery stopped.");
    }
}

void CLinker::OnCleanDiscovery(const string& _szGtwId) {
    Poco::Mutex::ScopedLock lock(m_DiscMutex);
    DiscReqs::iterator itr = m_DiscReqs.find(_szGtwId);

    if ( itr!=m_DiscReqs.end() ) {
        CleanDiscovery(new string(_szGtwId));
        ChangeLinkState(_szGtwId, CLink::LINK_DISCOVERY_CANCELED);
    }
}

void CLinker::RegistrationDiscoveryCompleted(const string& _szGtwId, Discovery::CGtwInfo::Ptr _pDevInfo, void* _pData) {
    ChangeLinkState(_szGtwId, CLink::LINK_DISCOVERY_COMPLETED);
    Poco::SharedPtr<string> secId = static_cast<string*>(_pData);
    string url;
    Poco::format(url, GTW_URL_FORMAT, _pDevInfo->GetIp());
    Register(_szGtwId, *secId, url);
}

void CLinker::Register(const string& _szGtwId, const string& _szGtwSec, const string& _szUrl) {
    CRegistration* p = new CRegistration(m_szGtwId, _szGtwId, _szGtwSec, _szUrl, this, REG_TIMEOUT);
    string path = m_pConf->GetRemoteCaPath();
    p->SetSecureConnection(path);
    QueueRegistration(p);
    ChangeLinkState(_szGtwId, CLink::LINK_REG_STARTED);
    std::stringstream ss;
    ss << "Queued for registration (g=" << _szGtwId << ")";
    debug(ss.str());
}

void CLinker::OnRegCompleted(const string& _szRemoteGtwId, const string& _szRemoteGtwSecId, const string& _szUsername, const string& _szPassword, const string& _szUrl) {
    Poco::Mutex::ScopedLock lock(m_ModuleMutex);
    ChangeLinkState(_szRemoteGtwId, CLink::LINK_REG_COMPLETED);
    Links::const_iterator itr = m_Links.find(_szRemoteGtwId);

    if ( itr==m_Links.end() && m_Links.size()<m_pConf->GetMaxLinks() ) { // new link
        m_pLinksTable->StoreLink(_szRemoteGtwId, _szRemoteGtwSecId, _szUsername, _szPassword, _szUrl);
        CLink* p = new CLink(_szRemoteGtwId, _szRemoteGtwSecId);
        p->SetUrl(_szUrl);
        p->SetUsername(_szUsername);
        p->SetPassword(_szPassword);
        p->SetListener(this);
        m_Links.insert(LinkElm(p->GetId(), p));
        SendEvent(new Events::CLinkAdded(p->GetId()));
        std::stringstream ss;
        ss << "New link added (g=" << _szRemoteGtwId << ")";
        information(ss.str());
        p->Start();
    } else if ( itr==m_Links.end() && m_Links.size()>=m_pConf->GetMaxLinks() ) { // new link but already on maximum
        SendEvent(new Events::CLinkAddFailed(_szRemoteGtwId, -1));
    } else if ( itr!=m_Links.end() ) { // link re-registration completed.
        CLink* p = itr->second;
        p->SetUsername(_szUsername);
        p->SetPassword(_szPassword);
        m_pLinksTable->UpdateLinkCredentials(p->GetId(), _szUsername, _szPassword);
        p->Start();
    }
}

void CLinker::OnRegFailed(const string& _szRemoteGtwId, int16_t _nCode, const string& _szCause) {
    std::stringstream ss;
    ss << "Registration failed (g=" << _szRemoteGtwId << ", c=" << _nCode << ", m= " << _szCause << ")";
    warning(ss.str());
    ChangeLinkState(_szRemoteGtwId, CLink::LINK_NOT_REGISTERED);
}

void CLinker::OnStateChanged(const string& _szId, CLink::STATE _State) {
    switch (_State) {
        case CLink::LINK_OFFLINE: {
            ASYNC_CALL_P(ResurectLink, new string(_szId));
            break;
        }

        case CLink::LINK_NOT_REGISTERED: {
            ASYNC_CALL_P(ReregisterLink, new string(_szId));
            break;
        }

        default:
            break;
    }

    SendEvent(new Events::CLinkStateChanged(_szId, static_cast<int>(_State)));
}

void CLinker::ResurectLink(void* _pParams) {
    Poco::Mutex::ScopedLock lock(m_ModuleMutex);
    Poco::SharedPtr<string> id = static_cast<string*>(_pParams);
    Links::iterator itr = m_Links.find(*id);

    if ( itr!=m_Links.end() ) {
        CLink* p = itr->second;
        p->Stop();
    }

    RunDiscovery(*id, &CLinker::ResurectionDiscoveryCompleted, false);
}

void CLinker::ResurectionDiscoveryCompleted(const string& _szGtwId, Discovery::CGtwInfo::Ptr _pDevInfo, void* _pData) {
    Poco::Mutex::ScopedLock lock(m_ModuleMutex);
    Links::iterator itr = m_Links.find(_szGtwId);

    if ( itr!=m_Links.end() ) {
        CLink* l = itr->second;
        string url;
        Poco::format(url, GTW_URL_FORMAT, _pDevInfo->GetIp());
        l->SetUrl(url);
        l->Start();
        m_pLinksTable->UpdateLinkUrl(_szGtwId, url);
    }
}

void CLinker::ReregisterLink(void* _pParams) {
    Poco::Mutex::ScopedLock lock(m_ModuleMutex);
    Poco::SharedPtr<string> id = static_cast<string*>(_pParams);
    Links::iterator itr = m_Links.find(*id);

    if ( itr!=m_Links.end() ) {
        CLink* p = itr->second;
        p->Stop();

        if ( p->GetRegFailureCnt()<m_pConf->GetMaxRegFailures() ) {
            Register(p->GetId(), p->GetSecId(), p->GetUrl());
        } else {
            p->SetState(CLink::LINK_ERROR);
            stringstream ss;
            ss << "Link (" << p->GetId() << ") exceeded registration run-time count limit.";
            error(ss.str());
        }
    }
}

Poco::DynamicStruct::Ptr CLinker::RemoveLinkHandler(const string& _szCmdName, const Poco::DynamicStruct& _Params) {
    trace("Link removement requested...");
    Commands::CRemoveLink* cmd = Commands::CRemoveLink::Parse(_Params);
    string id = cmd->GetGtwId();
    Poco::Mutex::ScopedLock lock(m_ModuleMutex);
    Links::iterator itr = m_Links.find(id);

    if ( itr==m_Links.end() ) {
        throw Poco::ExistsException("Link not registered", -1);
    }

    CLink* p = itr->second;
    m_Links.erase(itr);
    p->Stop();
    SAFE_DELETE(p);
    m_pLinksTable->DeleteLink(id);
    ASYNC_CALL_P(CleanDiscovery, new string(id));
    SendEvent(new Events::CLinkRemoved(id));
    return new Poco::DynamicStruct();
}

Poco::DynamicStruct::Ptr CLinker::GetProp(const string& _szCmdName, const Poco::DynamicStruct& _Params) {
    Poco::DynamicStruct::Ptr ret = new Poco::DynamicStruct();
    std::string property = _Params[PARAM_PROPERTY].extract<std::string>();

    if ( property==PROP_MAX_LINKS ) {
        ret->insert(PARAM_VALUE, m_pConf->GetMaxLinks());
    } else if ( property==PROP_LINKS ) {
        Poco::Mutex::ScopedLock lock(m_ModuleMutex);
        Poco::Dynamic::List list;
        Links::const_iterator itr;

        for ( itr=m_Links.begin(); itr!=m_Links.end(); ++itr ) {
            CLink* p = itr->second;
            list.push_back(*p->ToStruct());
        }

        ret->insert(PARAM_VALUE, list);
    }

    return ret;
}

namespace Async {

CAsyncFunc::CAsyncFunc(const AsyncFunctionPtr _pFunc, void* _pParams):
    m_pFunc(_pFunc), m_pParams(_pParams) {
}

CAsyncFunc::~CAsyncFunc() {
}

CAsyncFuncHandler::CAsyncFuncHandler(CLinker* _pObj) :
    m_AsyncProcessor(*this),
    m_pObj(_pObj) {
    m_AsyncThread.start(m_AsyncProcessor);
}

CAsyncFuncHandler::~CAsyncFuncHandler() {
    m_AsyncProcessor.Stop();
    m_AsyncThread.join();
}

void CAsyncFuncHandler::Run(CAsyncFunc* _pFunc) {
    m_AsyncProcessor.Add(_pFunc);
}

void CAsyncFuncHandler::OnProcess(CAsyncFunc* _pFunc) {
    CAsyncFunc::AsyncFunctionPtr func = _pFunc->m_pFunc;
    (m_pObj->*func)(_pFunc->m_pParams);
    delete _pFunc;
}

} // Async

} // Linker
} // MessageBroker
