/****************************************************************************
 *  Copyright (c) 2016 RT-RK d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.rt-rk.com
 *  -----------------------------------------------------
 *
 * \project Oblo Message Broker - Broker application.
 * \file UserRegistration.cpp
 * \brief
 * OMB linker user registration.
 *
 * @Author Milan Tucic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#include "feature/linker/UserRegistration.h"
#include "core/Globals.h"
#include <sstream>

namespace MessageBroker {
namespace Linker {

using namespace Messenger;

const int8_t CRegistration::ERR_FAILURE = -1;
const int8_t CRegistration::ERR_PARSING = -2;
const int8_t CRegistration::ERR_PROTOCOL = -3;
const int8_t CRegistration::ERR_TIMEOUT = -4;

CRegistration::CRegistration(const std::string& _szGtwId,
                             const std::string& _szRemoteGtwId,
                             const std::string& _szRemoteGtwSecId,
                             const std::string& _szRemoteUrl,
                             IRegistrationListener* _pRegL,
                             uint8_t _nTimeout):
    m_szGtwId(_szGtwId),
    m_szRemoteGtwId(_szRemoteGtwId),
    m_szRemoteGtwSecId(_szRemoteGtwSecId),
    m_szRemoteUrl(_szRemoteUrl),
    m_pRegL(_pRegL),
    m_nTimeout(_nTimeout),
    m_Msgr("user", "user", "user", m_szRemoteUrl) {
    setLoggerName(MODULE_ID);
    m_nErr = ERR_FAILURE;
}

CRegistration::~CRegistration() {
}

void CRegistration::SetSecureConnection(const std::string& _szCA) {
    m_Msgr.SetSecureConfiguration(_szCA, "", "", "");
}

void CRegistration::Execute() {
    Poco::SHA1Engine engine;
    engine.update(m_szRemoteGtwId);
    engine.update(m_szRemoteGtwSecId);
    // Get hash
    std::string hash(Poco::DigestEngine::digestToHex(engine.digest()));
    std::string hashKey = hash.substr(0, REGISTRATION_CIPHER_KEY_LENGTH);
    const Poco::Crypto::CipherKey::ByteVec key(hashKey.begin(), hashKey.end());
    Poco::Crypto::CipherKey::ByteVec vector;
    m_pCipher = Poco::Crypto::CipherFactory::defaultFactory().createCipher(Poco::Crypto::CipherKey(ENCRYPTION_TYPE, key, vector));

    try {
        Credentials creds = GetCredentials(m_nTimeout);
        m_pRegL->OnRegCompleted(m_szRemoteGtwId, m_szRemoteGtwSecId, creds.first, creds.second, m_szRemoteUrl);
    } catch (const Poco::TimeoutException& e) {
        m_pRegL->OnRegFailed(m_szRemoteGtwId, ERR_TIMEOUT, e.message());
    } catch (const Poco::ProtocolException& e) {
        m_pRegL->OnRegFailed(m_szRemoteGtwId, ERR_PROTOCOL, e.message());
    } catch (const Poco::NullValueException& e) {
        m_pRegL->OnRegFailed(m_szRemoteGtwId, e.code(), e.message());
    } catch (const Poco::Exception& e) {
        m_pRegL->OnRegFailed(m_szRemoteGtwId, ERR_FAILURE, e.message());
    }
}

CRegistration::Credentials CRegistration::GetCredentials(uint8_t _nTimeout) {
    Poco::Timestamp time;
    RequestCredentials();
    int64_t timeout = _nTimeout*1000 - time.elapsed()/1000;
    m_SyncObj.wait(timeout);

    if ( m_szUsername.empty() || m_szPassword.empty() ) {
        throw Poco::NullValueException("Failed to get credentials", m_nErr);
    }

    return Credentials(m_szUsername, m_szPassword);
}

void CRegistration::RequestCredentials() {
    CUser::Ptr omb = m_Msgr.GetUser("gtw/" + m_szRemoteGtwId + "/omb");
    CFunctionality::Ptr auth = omb->GetFunctionality("auth");
    CRequestMessage req;
    req.SetType("command");
    req.SetName("register_user");
    // set args
    Poco::DynamicStruct::Ptr args = new Poco::DynamicStruct();
    args->insert("id", m_szGtwId);
    args->insert("group", "elnk");
    req.SetParams(args);
    m_Msgr.Connect();
    m_Msgr.RegisterResponder(this);
    CResponseMessage::Ptr rsp = auth->Request(req);

    if ( rsp->GetCode()==0 ) {
        Poco::DynamicStruct::Ptr params = rsp->GetParams();
        m_nRegNumber = (*params)["reg_number"].convert<int>();
        std::stringstream ss;
        ss << "Registration number arrived <" << m_nRegNumber  << ">, waiting for challenge.";
        trace(ss.str());
    } else {
        throw Poco::ProtocolException("Response for user registration failed: " + rsp->GetMessage(), rsp->GetCode());
    }
}

const CResponseMessage::Ptr CRegistration::Respond(const CRequestMessage::Ptr _pRequest) {
    CResponseMessage::Ptr rsp(NULL);

    if ( _pRequest->GetName()=="reg_challenge" ) {
        rsp = RespondOnChallenge(_pRequest);
    } else if ( _pRequest->GetName()=="change_credentials" ) {
        rsp = RespondOnChangeCreds(_pRequest);
    }

    return rsp;
}

const CResponseMessage::Ptr CRegistration::FunctionalityRespond(const std::string& _szFunctionality, const CRequestMessage::Ptr _pRequest) {
    return CResponseMessage::Ptr(NULL);
}

const CResponseMessage::Ptr CRegistration::RespondOnChallenge(const CRequestMessage::Ptr _pRequest) {
    CResponseMessage::Ptr rsp(NULL);
    std::string code;

    try {
        Poco::DynamicStruct args = *_pRequest->GetParams();
        code = args["code"].extract<std::string>();
    } catch (const Poco::Exception& e) {
        error("Error while parsing challenge request");
        m_nErr = ERR_PARSING;
        m_SyncObj.notify();
        return rsp;
    }

    std::string encryptedCode = m_pCipher->encryptString(code, Poco::Crypto::Cipher::ENC_BASE64);
    Poco::DynamicStruct::Ptr retArgs = new Poco::DynamicStruct();
    retArgs->insert("code", encryptedCode);
    rsp = new CResponseMessage();
    rsp->SetParams(retArgs);
    debug("Returning encrypted code: " + encryptedCode);
    return rsp;
}

const CResponseMessage::Ptr CRegistration::RespondOnChangeCreds(const CRequestMessage::Ptr _pRequest) {
    CResponseMessage::Ptr rsp(NULL);

    try {
        Poco::DynamicStruct args = *_pRequest->GetParams();
        m_szUsername = args["username"].extract<std::string>();
        m_szPassword = args["password"].extract<std::string>();
    } catch (const Poco::Exception& e) {
        error("Error while parsing change credentials request");
        m_nErr = ERR_PARSING;
        m_SyncObj.notify();
        return rsp;
    }

    rsp = new CResponseMessage();
    Poco::DynamicStruct::Ptr params = new Poco::DynamicStruct();
    params->insert("reg_number", m_nRegNumber);
    rsp->SetParams(params);
    debug("Password received: " + m_szPassword);
    m_SyncObj.notify(); // password arrived.
    return rsp;
}

CRegistrator::CRegistrator(): m_Registrator(*this) {
    m_Thread.start(m_Registrator);
}

CRegistrator::~CRegistrator() {
    m_Registrator.Stop();
    m_Thread.join();
}

void CRegistrator::QueueRegistration(CRegistration* _pReg) {
    m_Registrator.Add(_pReg);
}

void CRegistrator::OnProcess(CRegistration* _pReg) {
    _pReg->Execute();
    SAFE_DELETE(_pReg);
}


} // Linker
} // MessageBroker
