#include <iostream>

#include "Poco/NumberFormatter.h"
#include "Poco/Crypto/Cipher.h"
#include "Poco/Crypto/CipherFactory.h"
#include "Poco/Crypto/CipherKey.h"
#include "Poco/SHA1Engine.h"

#include "messenger/Messenger.h"
using namespace Messenger;

#define GTW_ID "150000000015"
#define GTW_SEC_ID GTW_ID "0000"

#define PRINTOUT(m) std::cout << m << std::endl;

class CLogPrinter: public ILogListener {
public:
    virtual void OnTrace(const std::string& _Msg) {
        //PRINTOUT("[T]" + _Msg);
    }
    virtual void OnDebug(const std::string& _Msg) {
        //PRINTOUT("[D]" + _Msg);
    }
    virtual void OnInformation(const std::string& _Msg) {
        //PRINTOUT("[I]" + _Msg);
    }
    virtual void OnWarning(const std::string& _Msg) {
        PRINTOUT("[W]" + _Msg);
    }
    virtual void OnError(const std::string& _Msg) {
        PRINTOUT("[E]" + _Msg);
    }
};

Poco::Crypto::Cipher::Ptr Cipher;
CLogPrinter Logger;

class CResponder: public IResponder {
public:

    virtual const CResponseMessage::Ptr Respond(const CRequestMessage::Ptr _Request) {
        CResponseMessage::Ptr rsp(NULL);

        if ( _Request->GetName()=="reg_challenge" ) {
            std::string code;

            try {
                Poco::DynamicStruct args = *_Request->GetParams();
                code = args["code"].extract<std::string>();
            } catch (Poco::Exception e) {
                PRINTOUT("Error while extracting challenge request");
                return rsp;
            }

            std::string encryptedCode = Cipher->encryptString(code, Poco::Crypto::Cipher::ENC_BASE64);
            Poco::DynamicStruct::Ptr retArgs = new Poco::DynamicStruct();
            retArgs->insert("code", encryptedCode);
            rsp = new CResponseMessage();
            rsp->SetParams(retArgs);
            PRINTOUT("Returning encrypted code: " + encryptedCode);
        } else if ( _Request->GetName()=="change_credentials" ) {
            try {
                Poco::DynamicStruct args = *_Request->GetParams();
                m_szUsername = args["username"].extract<std::string>();
                m_szPassword = args["password"].extract<std::string>();
                m_SyncObj.notify();
            } catch (Poco::Exception e) {
                PRINTOUT("Error while extracting change credentials request");
                m_SyncObj.notify();
                return rsp;
            }

            rsp = new CResponseMessage();
            Poco::DynamicStruct::Ptr params = new Poco::DynamicStruct();
            params->insert("reg_number", m_nRegNum);
            rsp->SetParams(params);
            PRINTOUT("Password received: " << m_szPassword);
        }

        return rsp;
    }

    virtual const CResponseMessage::Ptr FunctionalityRespond(const std::string& _szFunctionality, const CRequestMessage::Ptr _Request) {
        return CResponseMessage::Ptr(NULL);
    }

    void WaitForPassword() {
        m_SyncObj.wait(3000);
    }

    std::string GetUsername() const {
        return m_szUsername;
    }

    std::string GetPassword() const {
        return m_szPassword;
    }

    void SetRegistrationNumber(int _nRegNum) {
        m_nRegNum = _nRegNum;
    }

private:
    Poco::SynchronizedObject m_SyncObj;
    std::string m_szUsername;
    std::string m_szPassword;
    int m_nRegNum;
};

#define USER_PREFIX "cli/"

bool testConnection(const std::string& _szTag, const std::string& _szPassword) {
    bool passed = false;
    PRINTOUT("Testing connection for <"+ _szTag + "> with password <" + _szPassword + ">");
    CMessenger msgr(_szTag, _szTag, _szPassword);

    try {
        msgr.Connect();
        PRINTOUT("+++++Succesfully connected.");
        msgr.Disconnect();
        passed = true;
    } catch (Poco::Exception e) {
        PRINTOUT("-----Failed to connect. " + e.message());
    }

    return passed;
}

bool getCredentials(const std::string& _szId, std::string& _szUsername, std::string& _szPassword) {
    bool passed = false;
    CResponder responder;
    CMessenger msgr("user", "user", "user");
    CUser::Ptr omb = msgr.GetUser("gtw/" GTW_ID "/omb");
    CFunctionality::Ptr auth = omb->GetFunctionality("auth");
    CRequestMessage req;
    req.SetType("command");
    req.SetName("register_user");
    // set args
    Poco::DynamicStruct::Ptr args = new Poco::DynamicStruct();
    args->insert("id", _szId);
    req.SetParams(args);

    try {
        msgr.Connect();
        msgr.RegisterResponder(&responder);
        CResponseMessage::Ptr rsp = auth->Request(req);

        if ( rsp->GetCode()==0 ) {
            Poco::DynamicStruct::Ptr params = rsp->GetParams();
            int regnum = (*params)["reg_number"].convert<int>();
            responder.SetRegistrationNumber(regnum);
            PRINTOUT("Registration number arrived <" << regnum  << ">, waiting for challenge.");
            responder.WaitForPassword();
            _szUsername = responder.GetUsername();
            _szPassword = responder.GetPassword();
            // Wait for response to be delivered.
            Poco::Thread::sleep(100);
            passed = true;
        } else {
            PRINTOUT("Response for user registration failed: " + rsp->GetMessage());
        }
    } catch (Poco::Exception e) {
        PRINTOUT(e.message());
    }

    return passed;
}

#define REG_NUM 20


#define OUTPUT_DEFAULT "\x1b[0m"
#define OUTPUT_SUCCEEDED "\x1b[32;1mSUCCEEDED" OUTPUT_DEFAULT
#define OUTPUT_FAILED "\x1b[31;1mFAILED" OUTPUT_DEFAULT
#define OUTPUT_INFO_START "\x1b[34m"

void testUserRegistration() {
    // Create engine
    Poco::SHA1Engine engine;
    engine.update(GTW_ID);
    engine.update(GTW_SEC_ID);
    // Get hash
    std::string hash(Poco::DigestEngine::digestToHex(engine.digest()));
    std::string hashKey = hash.substr(0, 16);
    const Poco::Crypto::CipherKey::ByteVec key(hashKey.begin(), hashKey.end());
    Poco::Crypto::CipherKey::ByteVec vector;
    Cipher = Poco::Crypto::CipherFactory::defaultFactory().createCipher(Poco::Crypto::CipherKey("aes-128-ecb", key, vector));

    for (int i=0; i<REG_NUM; ++i) {
        std::string id = Poco::NumberFormatter::format0(i, 5);
        std::string username = std::string(USER_PREFIX) + id;
        std::string password;
        PRINTOUT(OUTPUT_INFO_START << "Test for user <" << username << ">" << OUTPUT_DEFAULT);
        std::cout << "Test with empty password:" << std::endl;
        std::cout << ( !testConnection(username, password) ? OUTPUT_SUCCEEDED : OUTPUT_FAILED) << std::endl;
        std::cout << "Test for getting credentials:" << std::endl;
        std::cout << ( getCredentials(id, username, password) ? OUTPUT_SUCCEEDED : OUTPUT_FAILED) << std::endl;
        std::cout << "Test with new credentials:" << std::endl;
        std::cout << ( testConnection(username, password) ? OUTPUT_SUCCEEDED : OUTPUT_FAILED) << std::endl;
        std::cout << "Test with bad credentials:" << std::endl;
        std::cout << ( !testConnection(username, "bad") ? OUTPUT_SUCCEEDED : OUTPUT_FAILED) << std::endl;
    }
}

CMessenger SafeInitContext("test", "test", "test");

int main() {
    testUserRegistration();
    return 0;
}
