
package org.apache.jmeter.protocol.oblomsgr.client;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.lang.System;

import org.apache.jorphan.logging.LoggingManager;
import org.eclipse.paho.client.mqttv3.IMqttDeliveryToken;
import org.eclipse.paho.client.mqttv3.MqttClient;
import org.eclipse.paho.client.mqttv3.MqttConnectOptions;
import org.eclipse.paho.client.mqttv3.MqttException;
import org.eclipse.paho.client.mqttv3.MqttMessage;

public class ObloClient extends IObloClient {

	private static final org.apache.log.Logger log = LoggingManager.getLoggerForClass();

	private class MsgDetails {
		public static final long EXPIRATION_TIME = 1000; // ms

		public MsgDetails(String topic, MqttMessage message) {
			this.topic = topic;
			this.message = message;
			expirationTime = System.currentTimeMillis() + EXPIRATION_TIME;
		}

		public String getTopic() {
			return topic;
		}

		public MqttMessage getMessage() {
			return message;
		}

		public boolean isExpired() {
			long currentTime = System.currentTimeMillis();
			return expirationTime <= currentTime;
		}

		private String topic;
		private MqttMessage message;
		private long expirationTime;
	}

	private MqttClient client;
	private String tag;
	private String brokerUrl;
	private MqttConnectOptions conOpt;
	private List<MsgDetails> messages = new LinkedList<MsgDetails>();
	List<String> subs = new LinkedList<String>();
	int protocol;
	String prefix;

	public ObloClient(String tag, String cliId, String url, int protocol) throws MqttException {
		this.tag = tag;
		this.brokerUrl = url;
		this.protocol = protocol;
		conOpt = new MqttConnectOptions();
		conOpt.setCleanSession(true);
		conOpt.setKeepAliveInterval(60);
		conOpt.setAutomaticReconnect(true);
		if (url.startsWith("ssl://")) {
			conOpt.setSocketFactory(SslUtil.getSocketFactory());
		}
		client = new MqttClient(this.brokerUrl, cliId, null);
		client.setCallback(this);
	}
	
	public void setCredentials(String username, String password) {
		conOpt.setUserName(username);
		conOpt.setPassword(password.toCharArray());
	}
	
	public void setCleanSession(boolean cleanSession) {
		conOpt.setCleanSession(cleanSession);
	}
	
	public void setAutoReconnect(boolean value) {
		conOpt.setAutomaticReconnect(value);
	}
	
	public void setKeepalive(int keepalive) {
		conOpt.setKeepAliveInterval(keepalive);
	}
	
	public void setTopicPrefix(String prefix) {
		this.prefix = prefix;
	}
	
	@Override
	public int getProtocol() {
		return protocol;
	}

	@Override
	public void connect() throws MqttException {
		client.connect(conOpt);
	}

	@Override
	public void disconnect() throws MqttException {
		client.disconnect();
	}

	@Override
	public void publish(String topicName, int qos, String payload) throws MqttException {
		MqttMessage message = new MqttMessage(payload.getBytes());
		message.setRetained(false);
		message.setQos(qos);
		String topic = prefix + topicName;
		client.publish(topic, message);
	}

	@Override
	public void subscribe(String topicName, int qos) throws MqttException {
		String topic = prefix + topicName;

		if( !subs.contains(topic) ) {
			log.info("Subscribed: " + topic);
			client.subscribe(topic, qos);
			subs.add(topic);
		}
	}

	@Override
	public void unsubscribe(String topicName) throws MqttException {
		String topic = prefix + topicName;
		client.unsubscribe(topic);
		subs.remove(topic);
		log.info("Unsubscribed: " + topic);
	}

	@Override
	public void close() throws MqttException {
		messages.clear();
		client.disconnect();
	}

	@Override
	public boolean isConnected() {
		return client.isConnected();
	}

	@Override
	public String getBrokerUrl() {
		return brokerUrl;
	}

	@Override
	public String getTag() {
		return tag;
	}

	@Override
	public void connectionLost(Throwable cause) {
		log.info("Connection to " + brokerUrl + " lost!" + cause);
	}

	@Override
	public void deliveryComplete(IMqttDeliveryToken token) {
	}

	@Override
	public synchronized void messageArrived(String topic, MqttMessage message) throws MqttException {
		boolean processed = false;
		
		if( !prefix.isEmpty() ) {
			topic = topic.replace(prefix, "");
		}

		log.debug("Arrived: " + topic + ", m=" + message);
		for (IMessageListener listener : listeners) {
			processed = listener.OnMessageArrived(topic, message);
			if (processed) {
				// Break the loop. Only one listener will get a message.
				break;
			}
		}
		if (!processed) {
			Insert(new MsgDetails(topic, message));
		}
	}

	@Override
	public synchronized void registerListener(IMessageListener listener) {
		super.registerListener(listener);
		removeExpiredMessages();
		deliverPendingMessages(listener);
	}

	private synchronized void Insert(MsgDetails msg) {
		if (messages.size()>=100) {
			messages.remove(0);
		}

		messages.add(msg);
	}

	private void removeExpiredMessages() {
		int index = -2;

		for (int i=0; i<messages.size(); ++i) {
			MsgDetails e = messages.get(i);

			if (!e.isExpired()) { // find the first valid element
				index = i-1;
				break; // all expired elements are removed; break
			}
		}

		if (index == -2) { // all expired
			messages.clear();
			log.debug("All messages expired, queue dropped.");
		} else if (index >= 0) { // check if index is greater than 0
			messages.subList(0, index).clear(); // remove elements
			log.debug("Expired messages removed: " + (index + 1) + ", remained: " + messages.size());
		}
	}

	private void deliverPendingMessages(IMessageListener listener) {
		List<Integer> messagesToRemove = new LinkedList<Integer>();

		// Report messages that arrived before the registration
		for (int i=0; i<messages.size(); ++i) {
			MsgDetails e = messages.get(i);
			String topic = e.getTopic();
			MqttMessage msg = e.getMessage();
			boolean processed = listener.OnMessageArrived(topic, msg);

			if (processed) {
				messagesToRemove.add(i);
			}
		}

		if(messagesToRemove.size() != 0) {
			log.debug("Messages to remove: " + messagesToRemove.size() + " of " + messages.size());	
		}

		// Remove pending messages.
		for (Integer index : messagesToRemove) {
			int i = index;
			messages.remove(i);
		}
	}
}