
package org.apache.jmeter.protocol.oblomsgr.sampler;

import java.util.concurrent.TimeoutException;
import org.json.JSONObject;
import java.util.Iterator;

import org.apache.jmeter.protocol.oblomsgr.client.ClientStore;
import org.apache.jmeter.protocol.oblomsgr.client.IMessageListener;
import org.apache.jmeter.protocol.oblomsgr.client.IObloClient;
import org.apache.jmeter.protocol.oblomsgr.data.EventMessage;
import org.apache.jmeter.samplers.AbstractSampler;
import org.apache.jmeter.samplers.Entry;
import org.apache.jmeter.samplers.SampleResult;
import org.apache.jorphan.logging.LoggingManager;
import org.apache.log.Logger;
import org.eclipse.paho.client.mqttv3.MqttMessage;

public class ObloMsgEventListener extends AbstractSampler implements IMessageListener {

	private static final long serialVersionUID = 233L;
	private static final Logger log = LoggingManager.getLoggerForClass();

	public static final String TIMEOUT = "oblomsgr.timeout";
	public static final String EVT_UID = "oblomsgr.evt.uid";
	public static final String EVT_SENDER = "oblomsgr.evt.sender";
	public static final String EVT_SENDER_FUNCTIONALITY = "oblomsgr.evt.sender.funct";
	public static final String EVT_NAME = "oblomsgr.evt.name";
	public static final String EVT_PARAMS = "oblomsgr.evt.params";
	public static final String EVT_EXPECTED_PARAMS = "oblomsgr.evt.expected_params";

	private boolean arrived = false;
	private JSONObject expectedParams = null;
	private int protocol = 0;
	private String topic = "";

	public ObloMsgEventListener() {
	}

	public String getTimeout() {
		return getPropertyAsString(TIMEOUT);
	}

	public int getEvtUid() {
		return getPropertyAsInt(EVT_UID);
	}

	public String getEvtSender() {
		return getPropertyAsString(EVT_SENDER);
	}

	public String getEvtSenderFunctionality() {
		return getPropertyAsString(EVT_SENDER_FUNCTIONALITY);
	}

	public String getEvtName() {
		return getPropertyAsString(EVT_NAME);
	}

	public String getEvtParams() {
		return getPropertyAsString(EVT_PARAMS);
	}

	public String getEvtExpectedParams() {
		return getPropertyAsString(EVT_EXPECTED_PARAMS);
	}

	public void setEvtUid(int uid) {
		setProperty(EVT_UID, uid);
	}

	public void setTimeout(String timeout) {
		setProperty(TIMEOUT, timeout);
	}

	public void setEvtSender(String sender) {
		setProperty(EVT_SENDER, sender.trim());
	}

	public void setEvtSenderFunctionality(String funct) {
		setProperty(EVT_SENDER_FUNCTIONALITY, funct.trim());
	}

	public void setEvtName(String name) {
		setProperty(EVT_NAME, name.trim());
	}

	public void setEvtParams(String params) {
		setProperty(EVT_PARAMS, params.trim());
	}

	public void setEvtExpectedParams(String params) {
		setProperty(EVT_EXPECTED_PARAMS, params.trim());
	}

	@Override
	public SampleResult sample(Entry entry) {
		SampleResult result = new SampleResult();
		result.setSampleLabel(getName());
		result.sampleStart();

		String sender = getEvtSender();
		arrived = false;
		String evtExpectedParams = getEvtExpectedParams();
		if( !evtExpectedParams.isEmpty() ) {
			expectedParams = new JSONObject(evtExpectedParams);
		} else {
			expectedParams = null;
		}

		ClientStore cliStore = ClientStore.getInstance();
		IObloClient client = cliStore.getClientByThread();

		if (client == null) {
			result.setResponseMessage("Client not active. Do login first.");
			result.setDataType(org.apache.jmeter.samplers.SampleResult.TEXT);
			result.setResponseCode("-1");
			return result;
		}

		protocol = client.getProtocol();
		client.registerListener(this);
		
		String funct = getEvtSenderFunctionality();

		if (funct.isEmpty()) {
			topic = sender + "/evt";
		} else {
			topic = sender + "/" + funct + "/evt";
		}
		// log.info("Waiting for an event...");

		try {
			client.subscribe(topic, 1);
			waitForEvent(Integer.parseInt(getTimeout())); // wait for a response
			result.setSuccessful(true);
			result.sampleEnd();
			result.setResponseCode("0");
			result.setResponseMessage("Event received.");
			result.setDataType(org.apache.jmeter.samplers.SampleResult.TEXT);
			result.setResponseData(getEvtParams(), null);
		} catch (Exception e) {
			result.sampleEnd();
			result.setSuccessful(false);
			result.setResponseMessage("Event failed: " + e.toString());
			result.setDataType(org.apache.jmeter.samplers.SampleResult.TEXT);
			result.setResponseCode("-1");
		}

		client.unregisterListener(this);
		return result;
	}

	@Override
	public synchronized boolean OnMessageArrived(String topic, MqttMessage message) {
		String payload = new String(message.getPayload());

		if (this.topic.equals(topic)) {
			try {
				EventMessage event = EventMessage.parseJson(payload, protocol);
				if (event.getName().equals(getEvtName()) &&
					paramsMatch(event.getParams(), expectedParams)) {
					// log.info("Request arrived.");
					setEvtUid(event.getUid());
					setEvtParams(event.getParamsString());
					arrived = true;
					notify();
					log.debug("Event matches: " + getEvtName());
					return true;
				} else {
					String p = expectedParams==null ? "" : expectedParams.toString(protocol);
					log.info("Event doesn't match (evt=" + event.toString(protocol) +
						", expName=" + getEvtName() + ", expParams=" + p + "). Still waiting.");
				}
			} catch (Exception e) {
				log.error("Error while event message: " + e.toString());
			}
		}

		return false;
	}

	private boolean paramsMatch(JSONObject received, JSONObject expected) {
		boolean match = true;

		if( expected != null ) {
			Iterator<?> keys = expected.keys();

			while( keys.hasNext() ) {
			    String key = (String)keys.next();

			    if( !received.has(key) ||
			    	!expected.get(key).equals(received.get(key)) ) {
			    	match = false;
			    	break;
			    }
			}
		}

		return match;
	}

	private synchronized void waitForEvent(int timeout) throws Exception {
		if (!arrived) {
			wait(timeout);
			if (!arrived) {
				throw new TimeoutException("");
			}
		}
	}
}
