
package org.apache.jmeter.protocol.oblomsgr.sampler;

import org.apache.jmeter.protocol.oblomsgr.client.ClientStore;
import org.apache.jmeter.protocol.oblomsgr.client.IObloClient;
import org.apache.jmeter.protocol.oblomsgr.client.ObloClient;
import org.apache.jmeter.protocol.oblomsgr.utilities.Constants;
import org.apache.jmeter.samplers.AbstractSampler;
import org.apache.jmeter.samplers.Entry;
import org.apache.jmeter.samplers.SampleResult;
import org.apache.jmeter.testelement.TestStateListener;
import org.eclipse.paho.client.mqttv3.MqttException;

public class ObloMsgInitializer extends AbstractSampler implements TestStateListener {
	private static final long serialVersionUID = 233L;

	public static final String TAG = "oblomsgr.tag";
	public static final String TOPIC_PREFIX = "oblomsgr.prefix";
	public static final String CLIENT_ID = "oblomsgr.clientId";
	public static final String USERNAME = "oblomsgr.username";
	public static final String PASSWORD = "oblomsgr.password";
	public static final String PROTOCOL = "oblomsgr.protocol";
	public static final String BROKER_URL = "oblomsgr.url";
	public static final String BROKER_CLEANSESSION = "oblomsgr.cleansession";
	public static final String BROKER_KEEPALIVE = "oblomsgr.keepalive";
	public static final String BROKER_AUTORECONNECT = "oblomsgr.autoreconnect";

	public ObloMsgInitializer() {
		setBrokerUrl(Constants.OBLOMSGR_OMB_DEFAULT_URL);
		setCleanSession(Constants.OBLOMSGR_DEFAULT_CLEANSESSION);
		setKeepalive(Constants.OBLOMSGR_DEFAULT_KEEPALIVE);
		setAutoReconnect(Constants.OBLOMSGR_DEFAULT_AUTORECONNECT);
	}

	public String getTag() {
		return getPropertyAsString(TAG);
	}

	public String getTopicPrefix() {
		return getPropertyAsString(TOPIC_PREFIX);
	}

	public String getClientId() {
		return getPropertyAsString(CLIENT_ID);
	}

	public String getUsername() {
		return getPropertyAsString(USERNAME);
	}

	public String getPassword() {
		return getPropertyAsString(PASSWORD);
	}

	public String getBrokerUrl() {
		return getPropertyAsString(BROKER_URL);
	}

	public boolean IsCleanSession() {
		return getPropertyAsBoolean(BROKER_CLEANSESSION);
	}

	public int getKeepalive() {
		return getPropertyAsInt(BROKER_KEEPALIVE);
	}

	public int getProtocol() {
		return getPropertyAsInt(PROTOCOL);
	}

	public boolean IsAutoReconnect() {
		return getPropertyAsBoolean(BROKER_AUTORECONNECT);
	}

	public void setTopicPrefix(String prefix) {
		setProperty(TOPIC_PREFIX, prefix.trim());
	}

	public void setTag(String tag) {
		setProperty(TAG, tag.trim());
	}

	public void setClientId(String cliId) {
		setProperty(CLIENT_ID, cliId.trim());
	}

	public void setUsername(String username) {
		setProperty(USERNAME, username.trim());
	}

	public void setPassword(String password) {
		setProperty(PASSWORD, password.trim());
	}

	public void setBrokerUrl(String brokerURL) {
		setProperty(BROKER_URL, brokerURL.trim());
	}

	public void setCleanSession(boolean value) {
		setProperty(BROKER_CLEANSESSION, value);
	}

	public void setKeepalive(String value) {
		setProperty(BROKER_KEEPALIVE, value);
	}

	public void setAutoReconnect(boolean value) {
		setProperty(BROKER_AUTORECONNECT, value);
	}

	public void setProtocol(int value) {
		setProperty(PROTOCOL, value);
	}

	@Override
	public void testEnded() {
		String tag = getTag();
		ClientStore cliStore = ClientStore.getInstance();
		IObloClient client = cliStore.getClientByTag(tag);

		if (client != null) {
			try {
				client.close();
				cliStore.removeClientByTag(tag);
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
	}

	@Override
	public void testEnded(String arg0) {
		testEnded();
	}

	@Override
	public void testStarted() {
	}

	@Override
	public void testStarted(String arg0) {
	}

	@Override
	public SampleResult sample(Entry entry) {
		SampleResult result = new SampleResult();
		result.setSampleLabel(getName());
		result.sampleStart();

		String tag = getTag();
		String prefix = getTopicPrefix();
		String cliId = getClientId();
		String username = getUsername();
		String password = getPassword();
		String brokerURL = getBrokerUrl();
		int protocol = getProtocol();
		ClientStore cliStore = ClientStore.getInstance();

		if (cliId.isEmpty()) {
			cliId = tag;
		}
		if (username.isEmpty()) {
			username = tag;
		}

		try {
			ObloClient client = new ObloClient(tag, cliId, brokerURL, protocol);
			client.setTopicPrefix(prefix);
			client.setCredentials(username, password);
			client.setCleanSession(IsCleanSession());
			client.setAutoReconnect(IsAutoReconnect());
			client.setKeepalive(getKeepalive());
			cliStore.addClient(client);
			client.connect();
			client.subscribe(client.getTag() + "/req", 1);
			client.subscribe(client.getTag() + "/rsp", 1);
			result.setSuccessful(true);
			result.sampleEnd();
			result.setResponseMessage(
					"Client: " + cliId + ", tag: " + tag + ", protocol: " + protocol + " is initialized and connected to " + client.getBrokerUrl() + ".");
			result.setResponseCode("0");
		} catch (Exception e) {
			result.sampleEnd();
			result.setSuccessful(false);
			result.setResponseMessage("User: " + tag + " is not able to connect with <" + username + ", " + password
					+ ">: " + e.toString());
			result.setDataType(org.apache.jmeter.samplers.SampleResult.TEXT);
			result.setResponseCode("-1");
		}

		return result;
	}
}
