
package org.apache.jmeter.protocol.oblomsgr.sampler;

import java.util.concurrent.TimeoutException;

import org.apache.jmeter.protocol.oblomsgr.client.ClientStore;
import org.apache.jmeter.protocol.oblomsgr.client.IMessageListener;
import org.apache.jmeter.protocol.oblomsgr.client.IObloClient;
import org.apache.jmeter.protocol.oblomsgr.data.RequestMessage;
import org.apache.jmeter.protocol.oblomsgr.data.ResponseMessage;
import org.apache.jmeter.samplers.AbstractSampler;
import org.apache.jmeter.samplers.Entry;
import org.apache.jmeter.samplers.SampleResult;
import org.apache.jorphan.logging.LoggingManager;
import org.apache.log.Logger;
import org.eclipse.paho.client.mqttv3.MqttMessage;

public class ObloMsgRequester extends AbstractSampler implements IMessageListener {
	private static final long serialVersionUID = 233L;
	private static final Logger log = LoggingManager.getLoggerForClass();

	public static final String USER = "oblomsgr.user";
	public static final String REQ_TIMEOUT = "oblomsgr.req.timeout";
	public static final String REQ_TYPE = "oblomsgr.req.type";
	public static final String REQ_NAME = "oblomsgr.req.name";
	public static final String REQ_FUNCTIONALITY = "oblomsgr.req.funct";
	public static final String REQ_PARAMS = "oblomsgr.req.params";
	public static final String RSP_PARAMS = "oblomsgr.rsp.params";

	private RequestMessage request = null;
	private ResponseMessage response = null;
	private int protocol = 0;

	public ObloMsgRequester() {
	}

	public String getUser() {
		return getPropertyAsString(USER);
	}

	public String getReqTimeout() {
		return getPropertyAsString(REQ_TIMEOUT);
	}

	public String getReqType() {
		return getPropertyAsString(REQ_TYPE);
	}

	public String getReqName() {
		return getPropertyAsString(REQ_NAME);
	}

	public String getReqParams() {
		return getPropertyAsString(REQ_PARAMS);
	}

	public String getReqFunctionality() {
		return getPropertyAsString(REQ_FUNCTIONALITY);
	}

	public String getRspParams() {
		return getPropertyAsString(RSP_PARAMS);
	}

	public void setUser(String user) {
		setProperty(USER, user.trim());
	}

	public void setReqTimeout(String timeout) {
		setProperty(REQ_TIMEOUT, timeout);
	}

	public void setReqType(String type) {
		setProperty(REQ_TYPE, type.trim());
	}

	public void setReqName(String name) {
		setProperty(REQ_NAME, name.trim());
	}

	public void setReqParams(String params) {
		setProperty(REQ_PARAMS, params.trim());
	}

	public void setReqFunctionality(String funct) {
		setProperty(REQ_FUNCTIONALITY, funct.trim());
	}

	public void setRspParams(String params) {
		setProperty(RSP_PARAMS, params.trim());
	}

	@Override
	public SampleResult sample(Entry entry) {
		SampleResult result = new SampleResult();
		result.setSampleLabel(getName());
		result.sampleStart();

		String user = getUser();
		ClientStore cliStore = ClientStore.getInstance();
		IObloClient client = cliStore.getClientByThread();

		if (client == null) {
			result.setResponseMessage("Client not active. Do login first.");
			result.setDataType(org.apache.jmeter.samplers.SampleResult.TEXT);
			result.setResponseCode("-1");
			return result;
		}

		protocol = client.getProtocol();
		request = new RequestMessage();
		request.setSender(client.getTag());
		request.setType(getReqType());
		request.setName(getReqName());
		request.setParams(getReqParams());

		response = null;

		String payload = request.toString(protocol);
		client.registerListener(this);
		result.setRequestHeaders(payload);
		// log.info("Sending request: " + reqString);

		String topic;
		String funct = getReqFunctionality();

		if (funct.isEmpty()) {
			topic = user + "/req";
		} else {
			topic = user + "/" + funct + "/req";
		}

		try {
			client.publish(topic, 0, payload);
			waitForResponse(Integer.parseInt(getReqTimeout())); // wait for a response
			setRspParams(response.getParamsString());
			result.sampleEnd();
			result.setSuccessful(true);
			result.setResponseCode("" + response.getCode());
			result.setResponseMessage(response.getMessage());
			result.setDataType(org.apache.jmeter.samplers.SampleResult.TEXT);
			result.setResponseData(response.toString(protocol), null);
		} catch (Exception e) {
			result.sampleEnd();
			result.setSuccessful(false);
			result.setResponseMessage("Unable to send request: " + e.toString());
			result.setDataType(org.apache.jmeter.samplers.SampleResult.TEXT);
			result.setResponseCode("-1");
		}

		client.unregisterListener(this);
		return result;
	}

	@Override
	public synchronized boolean OnMessageArrived(String topic, MqttMessage message) {
		String payload = new String(message.getPayload());
		// log.info("A message arrived to topic: " + topic + ", payload: " +
		// payload);

		if (topic.endsWith("rsp")) {
			try {
				ResponseMessage tmpRsp = ResponseMessage.parseJson(payload, protocol);
				if (tmpRsp.getUid() == request.getUid()) {
					// log.info("Response arrived, calling notify.");
					response = tmpRsp;
					notify();
					return true;
				} else {
					// log.info("Response UID is not valid. Still waiting.");
				}
			} catch (Exception e) {
				log.error("Error while parsing message: " + e.toString());
			}
		}

		return false;
	}

	private synchronized void waitForResponse(int timeout) throws Exception {
		// log.info("Request sent, waiting for a response.");
		if (response == null) {
			wait(timeout);
			if (response == null) {
				throw new TimeoutException("Response timeout.");
			}
		}
	}
}
