
package org.apache.jmeter.protocol.oblomsgr.sampler;

import java.util.concurrent.TimeoutException;
import org.json.JSONObject;
import java.util.Iterator;

import org.apache.jmeter.protocol.oblomsgr.client.ClientStore;
import org.apache.jmeter.protocol.oblomsgr.client.IMessageListener;
import org.apache.jmeter.protocol.oblomsgr.client.IObloClient;
import org.apache.jmeter.protocol.oblomsgr.data.RequestMessage;
import org.apache.jmeter.protocol.oblomsgr.data.ResponseMessage;
import org.apache.jmeter.samplers.AbstractSampler;
import org.apache.jmeter.samplers.Entry;
import org.apache.jmeter.samplers.SampleResult;
import org.apache.jorphan.logging.LoggingManager;
import org.apache.log.Logger;
import org.eclipse.paho.client.mqttv3.MqttMessage;

public class ObloMsgResponder extends AbstractSampler implements IMessageListener {
	private static final long serialVersionUID = 233L;
	private static final Logger log = LoggingManager.getLoggerForClass();

	public static final String TIMEOUT = "oblomsgr.timeout";
	public static final String REQ_UID = "oblomsgr.req.uid";
	public static final String REQ_SENDER = "oblomsgr.req.sender";
	public static final String REQ_TYPE = "oblomsgr.req.type";
	public static final String REQ_NAME = "oblomsgr.req.name";
	public static final String REQ_PARAMS = "oblomsgr.req.params";
	public static final String REQ_FUNCTIONALITY = "oblomsgr.req.funct";
	public static final String RSP_PARAMS = "oblomsgr.rsp.params";

	private boolean arrived = false;
	RequestMessage request;
	private String responseData = new String();
	private int protocol = 0;
	private String topic = "";
	private JSONObject expectedParams = null;

	public ObloMsgResponder() {
	}

	public int getReqUid() {
		return getPropertyAsInt(REQ_UID);
	}

	public String getTimeout() {
		return getPropertyAsString(TIMEOUT);
	}

	public String getReqSender() {
		return getPropertyAsString(REQ_SENDER);
	}

	public String getReqType() {
		return getPropertyAsString(REQ_TYPE);
	}

	public String getReqName() {
		return getPropertyAsString(REQ_NAME);
	}

	public String getReqParams() {
		return getPropertyAsString(REQ_PARAMS);
	}

	public String getReqFunctionality() {
		return getPropertyAsString(REQ_FUNCTIONALITY);
	}

	public String getRspParams() {
		return getPropertyAsString(RSP_PARAMS);
	}

	public void setReqUid(int uid) {
		setProperty(REQ_UID, uid);
	}

	public void setTimeout(String timeout) {
		setProperty(TIMEOUT, timeout);
	}

	public void setReqSender(String sender) {
		setProperty(REQ_SENDER, sender.trim());
	}

	public void setReqType(String type) {
		setProperty(REQ_TYPE, type.trim());
	}

	public void setReqName(String name) {
		setProperty(REQ_NAME, name.trim());
	}

	public void setReqParams(String params) {
		setProperty(REQ_PARAMS, params.trim());
	}

	public void setReqFunctionality(String funct) {
		setProperty(REQ_FUNCTIONALITY, funct.trim());
	}

	public void setRspParams(String params) {
		setProperty(RSP_PARAMS, params.trim());
	}

	@Override
	public SampleResult sample(Entry entry) {
		SampleResult result = new SampleResult();
		result.setSampleLabel(getName());
		result.sampleStart();
		arrived = false;
		
		String reqExpectedParams = getReqParams();
		if( !reqExpectedParams.isEmpty() ) {
			expectedParams = new JSONObject(reqExpectedParams);
		} else {
			expectedParams = null;
		}

		ClientStore cliStore = ClientStore.getInstance();
		IObloClient client = cliStore.getClientByThread();

		if (client == null) {
			result.setResponseMessage("Client not active. Do login first.");
			result.setDataType(org.apache.jmeter.samplers.SampleResult.TEXT);
			result.setResponseCode("-1");
			return result;
		}
		client.registerListener(this);
		protocol = client.getProtocol();
		// log.info("Waiting for a request...");

		String funct = getReqFunctionality();

		if (funct.isEmpty()) {
			topic = client.getTag() + "/req";
		} else {
			topic = client.getTag() + "/" + funct + "/req";
		}

		try {
			client.subscribe(topic, 0);

			waitForRequest(Integer.parseInt(getTimeout())); // wait for a response
			
			result.setResponseData(request.toString(protocol), null);

			sendResponse(client, request);

			result.setSuccessful(true);
			result.sampleEnd();
			result.setResponseCode("0");
			result.setResponseMessage("Request arrived within " + getTimeout() + "ms");
			result.setDataType(org.apache.jmeter.samplers.SampleResult.TEXT);
			result.setRequestHeaders(responseData);
		} catch (Exception e) {
			result.sampleEnd();
			result.setSuccessful(false);
			result.setResponseMessage("Failed to get a request: " + e.toString());
			result.setDataType(org.apache.jmeter.samplers.SampleResult.TEXT);
			result.setResponseCode("-1");
		}

		client.unregisterListener(this);
		return result;
	}

	@Override
	public synchronized boolean OnMessageArrived(String topic, MqttMessage message) {
		String payload = new String(message.getPayload());
		// log.info("A message arrived to topic: " + topic + ", payload: " +
		// payload);

		if (this.topic.equals(topic)) {
			try {
				RequestMessage req = RequestMessage.parseJson(payload, protocol);
				if (req.getType().equals(getReqType()) && req.getName().equals(getReqName())
					&& paramsMatch(req.getParams(), expectedParams)) {
					// log.info("Request arrived.");
					setReqUid(req.getUid());
					setReqSender(req.getSender());
					setReqParams(req.getParamsString());
					request = req;
					arrived = true;
					notify();
					return true;
				} else {
					// log.info("Response doesn't match requested type and name.
					// Still waiting.");
				}
			} catch (Exception e) {
				log.error("Error while parsing message: " + e.toString());
			}
		}

		return false;
	}

	private boolean paramsMatch(JSONObject received, JSONObject expected) {
		boolean match = true;

		if( expected != null ) {
			Iterator<?> keys = expected.keys();

			while( keys.hasNext() ) {
			    String key = (String)keys.next();

			    if( !received.has(key) ||
			    	!expected.get(key).equals(received.get(key)) ) {
			    	match = false;
			    	break;
			    }
			}
		}

		return match;
	}

	private synchronized void waitForRequest(int timeout) throws Exception {
		if (!arrived) {
			wait(timeout);
			if (!arrived) {
				throw new TimeoutException("Request timeout.");
			}
		}
	}

	private void sendResponse(IObloClient client, RequestMessage request) {
		String rspParams = getRspParams();

		if (!rspParams.isEmpty()) {
			ResponseMessage response = new ResponseMessage();
			response.setUid(request.getUid());
			response.setCode(0);
			response.setParams(rspParams);
			responseData = response.toString(protocol);

			try {
				client.publish(request.getSender() + "/rsp", 0, responseData);
			} catch (Exception e) {
				log.error("Failed to send the response: " + e.toString());
			}
		}
	}
}
